package de.mpg.mpimet.yacgui.xml.interpolation;

import de.mpg.mpimet.yacgui.xml.Dom;

/**
 * Copyright (c) 2013 Max-Planck-Institute for Meteorology, Hamburg.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of Oracle or the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * Definite interpolation method radial basis function
 * 
 * @author Rene Redler
 */

public class RadialBasisFunctionInterpolation extends Interpolation {
	
	/**
	 * the rbf kernel method
	 * {@code GAUSS_KERNEL}
	 */
	public enum RBFKernel {
		GAUSS_KERNEL;
		
		public static boolean isValidString(String s) {
			return (s.equals(Dom.Value.GAUSS_KERNEL));
		}
		public static RBFKernel objectFromString(String s) {
			if (s.equals(Dom.Value.GAUSS_KERNEL)) {
				return GAUSS_KERNEL;
			}
			return null; 
		}
		
		@Override
		public String toString() {
			if (this == GAUSS_KERNEL) {
				return Dom.Value.GAUSS_KERNEL;
			}
			return null;
		}
	}

	private int n;
	private double rbfscalefactor;
	private RBFKernel rbfkernel;
	
	public RadialBasisFunctionInterpolation(int n, RBFKernel rbfkernel, double rbfscalefactor) throws IllegalArgumentException {
		if (n < 1) {
			throw new IllegalArgumentException("n must be bigger than 0");
		}
		
		attributesNumber = 3;
		method = InterpolationMethod.RADIAL_BASIS_FUNCTION;
		this.n = n;
		this.rbfscalefactor = rbfscalefactor;
		this.rbfkernel = rbfkernel;
	}
	
	/**
	 * 
	 * @return number of source point values
	 */
	public int getN() {
		return n;
	}
	
	/**
	 * 
	 * @return RBF scale factor
	 */
	public double getRBFScaleFactor() {
		return rbfscalefactor;
	}
	
	/**
	 * 
	 * @return object of {@link RBFKernel} indicating kernel method
	 */
	public RBFKernel getRBFKernel() {
		return rbfkernel;
	}
	
	/**
	 * get attribute name by the order of index
	 * @param index the order of attributes on the panel
	 * @return  the name of attribute
	 * 				if 0 return N,
	 *              if 1 return RBF_KERNEL,
	 *              if 2 return RBF_SCALE
	 */
	public String getAttributeName(int index) {
		if (index == 0) {
			return Dom.Attribute.N;
		} else if (index == 1) {
			return Dom.Attribute.RBF_KERNEL;
		} else if (index == 2) {
			return Dom.Attribute.RBF_SCALE;
		}
		return null;
	}
	
	/**
	 * get attribute value by the order of index
	 * @param index the order of attributes on the panel
	 * @return the value of attribute as String
	 * 				if 0 return N,
	 *              if 1 return RBF_KERNEL,
	 *              if 2 return RBF_SCALE
	 */
	public String getAttributeValue(int index) {
		if (index == 0) {
			return String.valueOf(n);
		} else if (index == 1) {
			return rbfkernel.toString();
		} else if (index == 2) {
			return String.valueOf(rbfscalefactor);
		}
		return null;
	}
}
