#!/usr/bin/env python3

##
# @file netcdf2yac.py
# @brief reading a netcdf file and send its content via yac to the other component
#
# This example reads a NetCDF file "tsurf.nc" and registers the
# defined grid to yac using the `Reg2dGrid` function. The data field is
# registered to yac. In the time loop the field is read and send to
# the other component via the `Field.put` function.
#
# @copyright Copyright  (C)  2022 DKRZ, MPI-M
#
# @author Nils-Arne Dreier <dreier@dkrz.de>
#

#
# Keywords:
# Maintainer: Nils-Arne Dreier <dreier@dkrz.de>
# URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
#
# This file is part of YAC.
#
# Redistribution and use in source and binary forms, with or without
# modification, are  permitted provided that the following conditions are
# met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# Neither the name of the DKRZ GmbH nor the names of its contributors
# may be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from yac import *
import xarray as xr
import numpy as np

yac = YAC("coupling.xml", "coupling.xsd")
comp = yac.def_comp(f"C0")

f = xr.open_dataset("tsurf.nc")
deg2rad = np.pi/180
lon = deg2rad*f["lon"]
lat = deg2rad*f["lat"]

grid = Reg2dGrid(f"grid0", lon, lat)
points = grid.def_points(Location.CORNER, lon, lat)

field = Field.create("example_field", comp, points)

yac.search()
time = f["time"]
for i, t in enumerate(time.values):
    print(t)
    field.put(f["tsurf"][i, :, :].values.reshape((1,-1)))
