/**
 * @file interpolation.c
 *
 * @copyright Copyright  (C)  2019 Moritz Hanke <hanke@dkrz.de>
 *                                 Thomas Jahns <jahns@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Thomas Jahns <jahns@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
// Get the definition of the 'restrict' keyword.
#include "config.h"
#endif

#include <string.h>

#include "utils.h"
#include "interpolation.h"
#include "yac_mpi.h"

enum buffer_state {
  BUFFER_UNUSED, //! contains no data and is currently
                 //! not being used in any communication
  BUFFER_WAIT,   //! contains valid data, but is not yet communicated
  BUFFER_ACTIVE, //! contains valid data, which is being communicated
};

struct t_mvp_at_src {
  Xt_redist halo_redist;
  size_t * halo_sizes;
  double ** halo_buffers;
  size_t tgt_count;
  enum buffer_state result_buffer_state;
  double ** result_buffer;
  size_t * num_src_per_tgt;
  double * weights;
  size_t * src_field_idx;
  size_t * src_idx;
  size_t num_src_fields;
  double ** src_fields_buffer;
  Xt_redist result_redist;
  Xt_request result_request;
  int is_source;
  int is_target;
};

struct t_mvp_at_tgt {
  Xt_redist src_redist;
  Xt_request send_src_request;
  enum buffer_state src_send_buffer_state;
  double ** src_send_buffer;
  double ** src_recv_buffer;
  double ** src_fields_buffer;
  size_t * src_send_buffer_sizes;
  size_t * src_recv_buffer_sizes;
  size_t * tgt_pos;
  size_t tgt_count;
  size_t * num_src_per_tgt;
  double * weights;
  size_t * src_field_idx;
  size_t * src_idx;
  size_t num_src_fields;
  int is_source;
  int is_target;
};

struct t_direct {
  Xt_redist redist;
  Xt_request request_redist;
  enum buffer_state src_send_buffer_state;
  size_t src_send_buffer_size;
  double ** src_send_buffer;
  double ** src_field_buffer;
  double ** dummy_recv_buffer;
  int is_source;
  int is_target;
};

struct t_direct_mf {
  Xt_redist redist;
  Xt_request request_redist;
  enum buffer_state src_send_buffer_state;
  size_t * src_send_buffer_sizes;
  double ** src_send_buffer;
  double ** src_field_buffer;
  double ** dummy_recv_buffer;
  size_t num_src_fields;
  int is_source;
  int is_target;
};

struct t_fixed {
  double value;
  size_t * pos;
  size_t count;
};

struct interpolation {

  // compute target results on source processes
  struct t_mvp_at_src * mvp_at_src; // with weights
  struct t_mvp_at_src * sum_at_src;  // without weights (all weights == 1.0)

  // compute target results on target processes
  struct t_mvp_at_tgt * mvp_at_tgt; // with weights
  struct t_mvp_at_tgt * sum_at_tgt;  // without weights (all weights == 1.0)

  struct t_direct * direct;
  struct t_direct_mf * direct_mf;

  struct t_fixed * fixed;
  size_t fixed_count;

  int is_source, is_target;
  size_t collection_size;

  int ref_count;
};

struct interpolation * yac_interpolation_new(size_t collection_size) {

  struct interpolation * interp = xmalloc(1 * sizeof(*interp));

  interp->mvp_at_src = NULL;
  interp->sum_at_src = NULL;
  interp->mvp_at_tgt = NULL;
  interp->sum_at_tgt = NULL;
  interp->direct = NULL;
  interp->direct_mf = NULL;
  interp->fixed = NULL;
  interp->fixed_count = 0;

  interp->is_source = 0;
  interp->is_target = 0;
  interp->collection_size = collection_size;

  interp->ref_count = 1;

  return interp;
}

static struct t_fixed generate_fixed(
  double value, size_t count, size_t * pos) {

  struct t_fixed fixed;

  fixed.value = value;
  fixed.count = count;
  fixed.pos = xmalloc(count * sizeof(*fixed.pos));
  memcpy(fixed.pos, pos, count * sizeof(*pos));

  return fixed;
}

void yac_interpolation_add_fixed(
  struct interpolation * interp, double value, size_t count, size_t * pos) {

  interp->is_target |= (count > 0);
  interp->fixed =
    xrealloc(
      interp->fixed, (interp->fixed_count + 1) * sizeof(*(interp->fixed)));
  interp->fixed[interp->fixed_count] = generate_fixed(value, count, pos);
  interp->fixed_count++;
}

double ** setup_src_buffer(
  size_t * src_buffer_sizes, size_t num_src_fields, size_t collection_size) {

  size_t src_buffer_size = 0;
  for (int i = 0; i < num_src_fields; ++i)
    src_buffer_size += src_buffer_sizes[i];
  double ** src_buffer =
    xmalloc(collection_size * num_src_fields * sizeof(*src_buffer));
  src_buffer[0] = xmalloc(collection_size * src_buffer_size);
  for (size_t i = 0, offset = 0; i < collection_size; ++i) {
    for (size_t j = 0; j < num_src_fields; offset += src_buffer_sizes[j++]) {
      src_buffer[i * num_src_fields + j] =
        (double*)((char*)(src_buffer[0]) + offset);
    }
  }
  return src_buffer;
}

static struct t_direct * generate_direct(
  Xt_redist redist, size_t src_send_buffer_size,
  size_t collection_size, int is_source, int is_target) {

  struct t_direct * direct = xmalloc(1 * sizeof(*direct));
  direct->redist = (redist == NULL)?NULL:xt_redist_copy(redist);
  direct->request_redist = XT_REQUEST_NULL;
  direct->src_send_buffer_state = BUFFER_UNUSED;
  direct->src_send_buffer_size = src_send_buffer_size;
  direct->src_send_buffer =
    setup_src_buffer(&src_send_buffer_size, 1, collection_size);
  direct->src_field_buffer =
    xmalloc(collection_size * sizeof(*(direct->src_field_buffer)));
  direct->dummy_recv_buffer =
    xmalloc(collection_size * sizeof(*(direct->dummy_recv_buffer)));
  for (size_t i = 0; i < collection_size; ++i) {
    direct->src_field_buffer[i] = NULL;
    direct->dummy_recv_buffer[i] = NULL;
  }
  direct->is_source = is_source;
  direct->is_target = is_target;

  return direct;
}

static struct t_direct_mf * generate_direct_mf(
  Xt_redist redist, size_t * src_send_buffer_sizes,
  size_t collection_size, size_t num_src_fields,
  int is_source, int is_target) {

  struct t_direct_mf * direct_mf = xmalloc(1 * sizeof(*direct_mf));
  direct_mf->redist = (redist == NULL)?NULL:xt_redist_copy(redist);
  direct_mf->request_redist = XT_REQUEST_NULL;
  direct_mf->src_send_buffer_state = BUFFER_UNUSED;
  direct_mf->src_send_buffer_sizes =
    xmalloc(num_src_fields * sizeof(*(direct_mf->src_send_buffer_sizes)));
  memcpy(direct_mf->src_send_buffer_sizes, src_send_buffer_sizes,
         num_src_fields * sizeof(*(direct_mf->src_send_buffer_sizes)));
  direct_mf->src_send_buffer =
    setup_src_buffer(src_send_buffer_sizes, num_src_fields, collection_size);
  direct_mf->src_field_buffer =
    xmalloc(num_src_fields * collection_size *
            sizeof(*(direct_mf->src_field_buffer)));
  direct_mf->dummy_recv_buffer =
    xmalloc(num_src_fields * collection_size *
            sizeof(*(direct_mf->dummy_recv_buffer)));
  for (size_t i = 0; i < num_src_fields * collection_size; ++i) {
    direct_mf->src_field_buffer[i] = NULL;
    direct_mf->dummy_recv_buffer[i] = NULL;
  }
  direct_mf->num_src_fields = num_src_fields;
  direct_mf->is_source = is_source;
  direct_mf->is_target = is_target;

  return direct_mf;
}

size_t * get_buffer_sizes(
  Xt_redist * redists, size_t num_src_fields,
  size_t (*xt_redist_get_buffer_size)(Xt_redist)) {

  size_t * buffer_sizes;
  if (redists == NULL) {
    buffer_sizes = xcalloc(num_src_fields, sizeof(*buffer_sizes));
  } else {
    buffer_sizes = xmalloc(num_src_fields * sizeof(*buffer_sizes));
    for (size_t i = 0; i < num_src_fields; ++i)
      buffer_sizes[i] = xt_redist_get_buffer_size(redists[i]);
  }
  return buffer_sizes;
}

size_t xt_redist_get_buffer_size(
  Xt_redist redist, MPI_Datatype(*xt_redist_get_MPI_Datatype)(Xt_redist, int)) {

  if (redist == NULL) return 0;

  MPI_Comm comm = xt_redist_get_MPI_Comm(redist);
  int comm_size;
  yac_mpi_call(MPI_Comm_size(comm, &comm_size), comm);

  size_t max_size = 0;

  for (int i = 0; i < comm_size; ++i) {

    MPI_Datatype dt = xt_redist_get_MPI_Datatype(redist, i);
    if (dt == MPI_DATATYPE_NULL) continue;
    MPI_Aint lb, extent;
    yac_mpi_call(MPI_Type_get_extent(dt, &lb, &extent), comm);
    size_t curr_size = (size_t)extent + (size_t)lb;
    if (curr_size > max_size) max_size = curr_size;
    yac_mpi_call(MPI_Type_free(&dt), comm);
  }
  return max_size;
}

static size_t xt_redist_get_send_buffer_size(Xt_redist redist) {

  return xt_redist_get_buffer_size(redist, xt_redist_get_send_MPI_Datatype);
}

static size_t xt_redist_get_recv_buffer_size(Xt_redist redist) {

  return xt_redist_get_buffer_size(redist, xt_redist_get_recv_MPI_Datatype);
}

void yac_interpolation_add_direct(
  struct interpolation * interp, Xt_redist redist_) {

  Xt_redist redist;
  size_t collection_size = (size_t)(interp->collection_size);

  size_t src_send_buffer_size = xt_redist_get_send_buffer_size(redist_);

  if ((collection_size > 1) && (redist_ != NULL)) {

    Xt_redist * redist_buffer =
      xmalloc(collection_size  * sizeof(*redist_buffer));

    for (size_t i = 0; i < collection_size; ++i)
      redist_buffer[i] = redist_;
    redist =
      xt_redist_collection_new(
        redist_buffer, (int)(collection_size), -1,
        xt_redist_get_MPI_Comm(redist_));
    free(redist_buffer);

  } else {
    redist = redist_;
  }

  int is_source = (redist != NULL) && (xt_redist_get_num_send_msg(redist) > 0);
  int is_target = (redist != NULL) && (xt_redist_get_num_recv_msg(redist) > 0);

  interp->direct =
    generate_direct(
      redist, src_send_buffer_size, collection_size, is_source, is_target);
  interp->is_source |= is_source;
  interp->is_target |= is_target;

  if (collection_size > 1) {

    if (redist != NULL) xt_redist_delete(redist);
  }
}

static Xt_redist combine_redists(
  Xt_redist * redists, size_t num_redists, size_t collection_size) {

  // check redists
  if (redists != NULL) {
    int redists_are_null = redists[0] == NULL;
    for (size_t i = 1; i < num_redists; ++i)
      YAC_ASSERT(
        redists_are_null == (redists[i] == NULL),
        "ERROR(combine_redists): invalid argument")
  } else {
    return NULL;
  }

  if (collection_size == 1) {
    if (num_redists == 1) return redists[0];
    else
      return
        xt_redist_collection_new(
          redists, (int)(num_redists * collection_size), -1,
          xt_redist_get_MPI_Comm(redists[0]));

  } else {
    Xt_redist * redists_buffer =
      xmalloc(collection_size * num_redists * sizeof(*redists_buffer));
    for (size_t i = 0; i < collection_size; ++i)
      for (size_t j = 0; j < num_redists; ++j)
        redists_buffer[i * num_redists + j] = redists[j];
    Xt_redist redist =
      xt_redist_collection_new(
        redists_buffer, (int)(num_redists * collection_size), -1,
        xt_redist_get_MPI_Comm(redists[0]));
    free(redists_buffer);
    return redist;
  }
}

void yac_interpolation_add_direct_mf(
  struct interpolation * interp, Xt_redist * redists, size_t num_src_fields) {

  size_t collection_size = (size_t)(interp->collection_size);
  Xt_redist redist = combine_redists(redists, num_src_fields, collection_size);

  size_t * src_send_buffer_sizes =
    get_buffer_sizes(redists, num_src_fields, xt_redist_get_send_buffer_size);

  int is_source = (redist != NULL) && (xt_redist_get_num_send_msg(redist) > 0);
  int is_target = (redist != NULL) && (xt_redist_get_num_recv_msg(redist) > 0);

  interp->direct_mf =
    generate_direct_mf(
      redist, src_send_buffer_sizes, collection_size, num_src_fields,
      is_source, is_target);
  interp->is_source |= is_source;
  interp->is_target |= is_target;

  free(src_send_buffer_sizes);

  if ((redist != NULL) && (redists != NULL) && (redist != redists[0]))
    xt_redist_delete(redist);
}

static struct t_mvp_at_src * generate_sum_mvp_at_src(
  Xt_redist halo_redist, size_t * halo_sizes, size_t tgt_count,
  size_t * num_src_per_tgt, double * weights, size_t * src_field_idx,
  size_t * src_idx, size_t num_src_fields, Xt_redist result_redist,
  size_t collection_size, int is_source, int is_target) {

  size_t total_num_src = 0;
  for (size_t i = 0; i < tgt_count; ++i) total_num_src += num_src_per_tgt[i];

  struct t_mvp_at_src * mvp_at_src = xmalloc(1 * sizeof(*(mvp_at_src)));
  mvp_at_src->result_request = XT_REQUEST_NULL;
  mvp_at_src->halo_redist =
    (halo_redist == NULL)?NULL:xt_redist_copy(halo_redist);
  mvp_at_src->halo_sizes =
    xmalloc(num_src_fields * sizeof(*(mvp_at_src->halo_sizes)));
  memcpy(
    mvp_at_src->halo_sizes, halo_sizes,
    num_src_fields * sizeof(*(mvp_at_src->halo_sizes)));
  mvp_at_src->halo_buffers =
    setup_src_buffer(halo_sizes, num_src_fields, collection_size);
  mvp_at_src->tgt_count = tgt_count;
  size_t result_buffer_size = tgt_count * sizeof(*(mvp_at_src->result_buffer));
  mvp_at_src->result_buffer_state = BUFFER_UNUSED;
  mvp_at_src->result_buffer =
    setup_src_buffer(&result_buffer_size, 1, collection_size);
  mvp_at_src->num_src_per_tgt =
    xmalloc(tgt_count * sizeof(*num_src_per_tgt));
  memcpy(mvp_at_src->num_src_per_tgt, num_src_per_tgt,
         tgt_count * sizeof(*num_src_per_tgt));
  if (weights != NULL) {
    mvp_at_src->weights = xmalloc(total_num_src * sizeof(*weights));
    memcpy(mvp_at_src->weights, weights, total_num_src * sizeof(*weights));
  } else {
    mvp_at_src->weights = NULL;
  }
  mvp_at_src->src_field_idx =
    xmalloc(total_num_src * sizeof(*src_field_idx));
  memcpy(mvp_at_src->src_field_idx, src_field_idx,
         total_num_src * sizeof(*src_field_idx));
  mvp_at_src->src_idx =
    xmalloc(total_num_src * sizeof(*src_idx));
  memcpy(mvp_at_src->src_idx, src_idx,
         total_num_src * sizeof(*src_idx));
  mvp_at_src->num_src_fields = num_src_fields;
  mvp_at_src->src_fields_buffer =
    xmalloc(
      collection_size * num_src_fields *
      sizeof(*(mvp_at_src->src_fields_buffer)));
  mvp_at_src->result_redist =
    (result_redist == NULL)?NULL:xt_redist_copy(result_redist);
  mvp_at_src->is_source = is_source;
  mvp_at_src->is_target = is_target;

  return mvp_at_src;
}

static void yac_interpolation_add_weight_sum_mvp_at_src_(
  struct interpolation * interp, Xt_redist * halo_redists,
  size_t tgt_count, size_t * num_src_per_tgt, double * weights,
  size_t * src_field_idx, size_t * src_idx,
  size_t num_src_fields, Xt_redist result_redist_,
  int is_source, int is_target, int with_weights) {

  size_t collection_size = (size_t)(interp->collection_size);
  Xt_redist halo_redist =
    combine_redists(halo_redists, num_src_fields, collection_size);
  Xt_redist result_redist =
    combine_redists(&result_redist_, 1, collection_size);

  size_t * halo_sizes =
    get_buffer_sizes(
      halo_redists, num_src_fields, xt_redist_get_recv_buffer_size);

  if (with_weights) {
    interp->mvp_at_src =
      generate_sum_mvp_at_src(
        halo_redist, halo_sizes, tgt_count, num_src_per_tgt, weights,
        src_field_idx, src_idx, num_src_fields, result_redist, collection_size,
        is_source, is_target);
  } else {
    interp->sum_at_src =
      generate_sum_mvp_at_src(
        halo_redist, halo_sizes, tgt_count, num_src_per_tgt, NULL,
        src_field_idx, src_idx, num_src_fields, result_redist, collection_size,
        is_source, is_target);
  }
  free(halo_sizes);
  interp->is_source |= is_source;
  interp->is_target |= is_target;

  if ((halo_redist != NULL) && (halo_redists != NULL) &&
      (halo_redist != halo_redists[0])) xt_redist_delete(halo_redist);
  if ((result_redist != result_redist_)) xt_redist_delete(result_redist);
}

static struct t_mvp_at_tgt * generate_sum_mvp_at_tgt(
  Xt_redist src_redist, size_t * src_send_buffer_sizes,
  size_t * src_recv_buffer_sizes, size_t * tgt_pos, size_t tgt_count,
  size_t * num_src_per_tgt, double * weights, size_t * src_field_idx,
  size_t * src_idx, size_t num_src_fields, size_t collection_size,
  int is_source, int is_target) {

  size_t total_num_src = 0;
  for (size_t i = 0; i < tgt_count; ++i) total_num_src += num_src_per_tgt[i];

  struct t_mvp_at_tgt * mvp_at_tgt = xmalloc(1 * sizeof(*(mvp_at_tgt)));
  mvp_at_tgt->src_redist = (src_redist == NULL)?NULL:xt_redist_copy(src_redist);
  mvp_at_tgt->send_src_request = XT_REQUEST_NULL;
  mvp_at_tgt->src_send_buffer_state = BUFFER_UNUSED;
  mvp_at_tgt->src_send_buffer =
    setup_src_buffer(src_send_buffer_sizes, num_src_fields, collection_size);
  mvp_at_tgt->src_recv_buffer =
    setup_src_buffer(src_recv_buffer_sizes, num_src_fields, collection_size);
  mvp_at_tgt->src_send_buffer_sizes =
    xmalloc(num_src_fields * sizeof(*src_send_buffer_sizes));
  memcpy(mvp_at_tgt->src_send_buffer_sizes, src_send_buffer_sizes,
         num_src_fields * sizeof(*src_send_buffer_sizes));
  mvp_at_tgt->src_fields_buffer =
    xmalloc(collection_size * num_src_fields *
            sizeof(*(mvp_at_tgt->src_fields_buffer)));
  for (size_t i = 0; i < collection_size * num_src_fields; ++i)
    mvp_at_tgt->src_fields_buffer[i] = NULL;
  mvp_at_tgt->src_recv_buffer_sizes =
    xmalloc(num_src_fields * sizeof(*src_recv_buffer_sizes));
  memcpy(mvp_at_tgt->src_recv_buffer_sizes, src_recv_buffer_sizes,
         num_src_fields * sizeof(*src_recv_buffer_sizes));
  mvp_at_tgt->tgt_pos = xmalloc(tgt_count * sizeof(*tgt_pos));
  memcpy(mvp_at_tgt->tgt_pos, tgt_pos, tgt_count * sizeof(*tgt_pos));
  mvp_at_tgt->tgt_count = tgt_count;
  mvp_at_tgt->num_src_per_tgt =
    xmalloc(tgt_count * sizeof(*num_src_per_tgt));
  memcpy(mvp_at_tgt->num_src_per_tgt, num_src_per_tgt,
         tgt_count * sizeof(*num_src_per_tgt));
  if (weights != NULL) {
    mvp_at_tgt->weights = xmalloc(total_num_src * sizeof(*weights));
    memcpy(mvp_at_tgt->weights, weights, total_num_src * sizeof(*weights));
  } else {
    mvp_at_tgt->weights = NULL;
  }
  mvp_at_tgt->src_field_idx =
    xmalloc(total_num_src * sizeof(*src_field_idx));
  memcpy(mvp_at_tgt->src_field_idx, src_field_idx,
         total_num_src * sizeof(*src_field_idx));
  mvp_at_tgt->src_idx =
    xmalloc(total_num_src * sizeof(*src_idx));
  memcpy(mvp_at_tgt->src_idx, src_idx,
         total_num_src * sizeof(*src_idx));
  mvp_at_tgt->num_src_fields = num_src_fields;
  mvp_at_tgt->is_source = is_source;
  mvp_at_tgt->is_target = is_target;

  return mvp_at_tgt;
}

static void yac_interpolation_add_weight_sum_mvp_at_tgt_(
  struct interpolation * interp, Xt_redist * src_redists, size_t * tgt_pos,
  size_t tgt_count, size_t * num_src_per_tgt, double * weights,
  size_t * src_field_idx, size_t * src_idx, size_t num_src_fields,
  int with_weights) {

  size_t collection_size = (size_t)(interp->collection_size);
  Xt_redist src_redist =
    combine_redists(src_redists, num_src_fields, collection_size);

  int is_source =
    (src_redist != NULL) && (xt_redist_get_num_send_msg(src_redist) > 0);
  int is_target = tgt_count > 0;

  // get send buffer sizes from src_redists
  size_t * src_send_buffer_sizes =
    get_buffer_sizes(
      src_redists, num_src_fields, xt_redist_get_send_buffer_size);
  // extend send buffer sizes for local source data if necessary
  for (size_t i = 0, offset = 0; i < tgt_count; ++i) {
    size_t curr_num_src = num_src_per_tgt[i];
    for (size_t j = 0; j < curr_num_src; ++j, ++offset) {
      size_t curr_src_field_idx = src_field_idx[offset];
      if (curr_src_field_idx == SIZE_MAX) continue;
      size_t curr_src_send_buffer_size =
        src_send_buffer_sizes[curr_src_field_idx];
      size_t curr_src_extent = (src_idx[offset] + 1) * sizeof(double);
      if (curr_src_send_buffer_size < curr_src_extent)
        src_send_buffer_sizes[curr_src_field_idx] = curr_src_extent;
    }
  }
  size_t * src_recv_buffer_sizes =
    get_buffer_sizes(
      src_redists, num_src_fields, xt_redist_get_recv_buffer_size);

  if (with_weights) {
    interp->mvp_at_tgt =
      generate_sum_mvp_at_tgt(
        src_redist, src_send_buffer_sizes, src_recv_buffer_sizes,
        tgt_pos, tgt_count, num_src_per_tgt, weights, src_field_idx, src_idx,
        num_src_fields, collection_size, is_source, is_target);
  } else {
    interp->sum_at_tgt =
      generate_sum_mvp_at_tgt(
        src_redist, src_send_buffer_sizes, src_recv_buffer_sizes,
        tgt_pos, tgt_count, num_src_per_tgt, NULL, src_field_idx, src_idx,
        num_src_fields, collection_size, is_source, is_target);
  }
  interp->is_source |= is_source;
  interp->is_target |= is_target;

  free(src_recv_buffer_sizes);
  free(src_send_buffer_sizes);

  if ((src_redist != NULL) && (src_redists != NULL) &&
      (src_redist != src_redists[0]))
    xt_redist_delete(src_redist);
}

void yac_interpolation_add_sum_at_src(
  struct interpolation * interp, Xt_redist * halo_redists,
  size_t tgt_count, size_t * num_src_per_tgt,
  size_t * src_field_idx, size_t * src_idx,
  size_t num_src_fields, Xt_redist result_redist,
  int is_source, int is_target) {

  yac_interpolation_add_weight_sum_mvp_at_src_(
    interp, halo_redists, tgt_count, num_src_per_tgt, NULL, src_field_idx,
    src_idx, num_src_fields, result_redist, is_source, is_target, 0);
}

void yac_interpolation_add_sum_at_tgt(
  struct interpolation * interp, Xt_redist * src_redists,
  size_t * tgt_pos, size_t tgt_count, size_t * num_src_per_tgt,
  size_t * src_field_idx, size_t * src_idx,
  size_t num_src_fields) {

  yac_interpolation_add_weight_sum_mvp_at_tgt_(
    interp, src_redists, tgt_pos, tgt_count, num_src_per_tgt, NULL,
    src_field_idx, src_idx, num_src_fields, 0);
}

void yac_interpolation_add_weight_sum_mvp_at_src(
  struct interpolation * interp, Xt_redist * halo_redists,
  size_t tgt_count, size_t * num_src_per_tgt, double * weights,
  size_t * src_field_idx, size_t * src_idx,
  size_t num_src_fields, Xt_redist result_redist,
  int is_source, int is_target) {

  yac_interpolation_add_weight_sum_mvp_at_src_(
    interp, halo_redists, tgt_count, num_src_per_tgt, weights, src_field_idx,
    src_idx, num_src_fields, result_redist, is_source, is_target, 1);
};

void yac_interpolation_add_weight_sum_mvp_at_tgt(
  struct interpolation * interp, Xt_redist * src_redists, size_t * tgt_pos,
  size_t tgt_count, size_t * num_src_per_tgt, double * weights,
  size_t * src_field_idx, size_t * src_idx,
  size_t num_src_fields) {

  yac_interpolation_add_weight_sum_mvp_at_tgt_(
    interp, src_redists, tgt_pos, tgt_count, num_src_per_tgt, weights,
    src_field_idx, src_idx, num_src_fields, 1);
}

struct interpolation * yac_interpolation_copy(struct interpolation * interp) {

  struct interpolation * interp_copy =
    yac_interpolation_new(interp->collection_size);

  if (interp->mvp_at_src != NULL) {
    interp_copy->mvp_at_src =
      generate_sum_mvp_at_src(
        interp->mvp_at_src->halo_redist, interp->mvp_at_src->halo_sizes,
        interp->mvp_at_src->tgt_count, interp->mvp_at_src->num_src_per_tgt,
        interp->mvp_at_src->weights, interp->mvp_at_src->src_field_idx,
        interp->mvp_at_src->src_idx, interp->mvp_at_src->num_src_fields,
        interp->mvp_at_src->result_redist, interp->collection_size,
        interp->mvp_at_src->is_source, interp->mvp_at_src->is_target);
    interp_copy->is_source |= interp->mvp_at_src->is_source;
    interp_copy->is_target |= interp->mvp_at_src->is_target;
  }

  if (interp->mvp_at_tgt != NULL) {
    interp_copy->mvp_at_tgt =
      generate_sum_mvp_at_tgt(
        interp->mvp_at_tgt->src_redist, interp->mvp_at_tgt->src_send_buffer_sizes,
        interp->mvp_at_tgt->src_recv_buffer_sizes, interp->mvp_at_tgt->tgt_pos,
        interp->mvp_at_tgt->tgt_count, interp->mvp_at_tgt->num_src_per_tgt,
        interp->mvp_at_tgt->weights, interp->mvp_at_tgt->src_field_idx,
        interp->mvp_at_tgt->src_idx, interp->mvp_at_tgt->num_src_fields,
        interp->collection_size,
        interp->mvp_at_tgt->is_source, interp->mvp_at_tgt->is_target);
    interp_copy->is_source |= interp->mvp_at_tgt->is_source;
    interp_copy->is_target |= interp->mvp_at_tgt->is_target;
  }

  if (interp->sum_at_src != NULL) {
    interp_copy->sum_at_src =
      generate_sum_mvp_at_src(
        interp->sum_at_src->halo_redist, interp->sum_at_src->halo_sizes,
        interp->sum_at_src->tgt_count, interp->sum_at_src->num_src_per_tgt,
        interp->sum_at_src->weights, interp->sum_at_src->src_field_idx,
        interp->sum_at_src->src_idx, interp->sum_at_src->num_src_fields,
        interp->sum_at_src->result_redist, interp->collection_size,
        interp->sum_at_src->is_source, interp->sum_at_src->is_target);
    interp_copy->is_source |= interp->sum_at_src->is_source;
    interp_copy->is_target |= interp->sum_at_src->is_target;
  }

  if (interp->sum_at_tgt != NULL) {
    interp_copy->sum_at_tgt =
      generate_sum_mvp_at_tgt(
        interp->sum_at_tgt->src_redist, interp->sum_at_tgt->src_send_buffer_sizes,
        interp->sum_at_tgt->src_recv_buffer_sizes, interp->sum_at_tgt->tgt_pos,
        interp->sum_at_tgt->tgt_count, interp->sum_at_tgt->num_src_per_tgt,
        interp->sum_at_tgt->weights, interp->sum_at_tgt->src_field_idx,
        interp->sum_at_tgt->src_idx, interp->sum_at_tgt->num_src_fields,
        interp->collection_size,
        interp->sum_at_tgt->is_source, interp->sum_at_tgt->is_target);
    interp_copy->is_source |= interp->sum_at_tgt->is_source;
    interp_copy->is_target |= interp->sum_at_tgt->is_target;
  }

  if (interp->fixed != NULL) {
    interp_copy->fixed =
      xmalloc(interp->fixed_count * sizeof(*(interp_copy->fixed)));
    interp_copy->fixed_count = interp->fixed_count;
    for (size_t i = 0; i < interp->fixed_count; ++i) {
      interp_copy->fixed[i] =
        generate_fixed(
          interp->fixed[i].value, interp->fixed[i].count, interp->fixed[i].pos);
      interp_copy->is_target |= (interp->fixed[i].count > 0);
    }
  }

  if (interp->direct != NULL) {
    interp_copy->direct =
      generate_direct(
        interp->direct->redist, interp->direct->src_send_buffer_size,
        interp->collection_size,
        interp->direct->is_source, interp->direct->is_target);
    interp_copy->is_source |= interp->direct->is_source;
    interp_copy->is_target |= interp->direct->is_target;
  }

  if (interp->direct_mf != NULL) {
    interp_copy->direct_mf =
      generate_direct_mf(
        interp->direct_mf->redist, interp->direct_mf->src_send_buffer_sizes,
        interp->collection_size, interp->direct_mf->num_src_fields,
        interp->direct_mf->is_source, interp->direct_mf->is_target);
    interp_copy->is_source |= interp->direct_mf->is_source;
    interp_copy->is_target |= interp->direct_mf->is_target;
  }

  return interp_copy;
}

static void yac_interpolation_execute_mvp_at_src(
  struct t_mvp_at_src * mvp_at_src,
  double *** src_fields, double ** tgt_field, size_t collection_size) {

  double ** results = mvp_at_src->result_buffer;

  if (mvp_at_src->is_source) {

    size_t num_src_fields = mvp_at_src->num_src_fields;
    double ** temp_src_fields = mvp_at_src->src_fields_buffer;
    double ** halo_buffers = mvp_at_src->halo_buffers;
    for (size_t i = 0; i < collection_size; ++i)
      for (size_t j = 0; j < num_src_fields; ++j)
        temp_src_fields[i * num_src_fields + j] = src_fields[i][j];

    // do halo exchange
    if (mvp_at_src->halo_redist != NULL)
      xt_redist_s_exchange(
        mvp_at_src->halo_redist,
        (int)(collection_size * num_src_fields),
        (void const**)temp_src_fields, (void**)halo_buffers);

    size_t tgt_count = mvp_at_src->tgt_count;
    size_t * num_src_per_tgt = mvp_at_src->num_src_per_tgt;
    double * weights = mvp_at_src->weights;
    size_t * src_field_idx = mvp_at_src->src_field_idx;
    size_t * src_idx = mvp_at_src->src_idx;

    if (weights != NULL) {

      // apply weights
      for (size_t l = 0; l < collection_size; ++l) {
        for (size_t i = 0, k = 0; i < tgt_count; ++i) {
          size_t curr_num_src_per_tgt = num_src_per_tgt[i];
          double result = 0.0;
          for (size_t j = 0; j < curr_num_src_per_tgt; ++j, ++k) {
            if (src_field_idx[k] == SIZE_MAX) {
              result +=
                halo_buffers[l * num_src_fields][src_idx[k]] * weights[k];
            } else {
              result +=
                src_fields[l][src_field_idx[k]][src_idx[k]] * weights[k];
            }
          }
          results[l][i] = result;
        }
      }
    } else {
      // sum source points
      for (size_t l = 0; l < collection_size; ++l) {
        for (size_t i = 0, k = 0; i < tgt_count; ++i) {
          size_t curr_num_src_per_tgt = num_src_per_tgt[i];
          double result = 0.0;
          for (size_t j = 0; j < curr_num_src_per_tgt; ++j, ++k) {
            if (src_field_idx[k] == SIZE_MAX) {
              result += halo_buffers[l * num_src_fields][src_idx[k]];
            } else {
              result += src_fields[l][src_field_idx[k]][src_idx[k]];
            }
          }
          results[l][i] = result;
        }
      }
    }
  }

  double * dummy = NULL;
  if (!mvp_at_src->is_target) tgt_field = &dummy;

  // redistribute results
  if (mvp_at_src->result_redist != NULL)
    xt_redist_s_exchange(
      mvp_at_src->result_redist, collection_size,
      (const void **)results, (void **)tgt_field);
}

static void yac_interpolation_execute_mvp_at_tgt(
  struct t_mvp_at_tgt * mvp_at_tgt, double *** src_fields, double ** tgt_field,
  size_t collection_size) {

  size_t num_src_fields = mvp_at_tgt->num_src_fields;
  double ** temp_src_fields = mvp_at_tgt->src_fields_buffer;
  double ** src_recv_buffer = mvp_at_tgt->src_recv_buffer;
  if (mvp_at_tgt->is_source)
    for (size_t i = 0; i < collection_size; ++i)
      for (size_t j = 0; j < num_src_fields; ++j)
        temp_src_fields[i * num_src_fields + j] = src_fields[i][j];

  // do halo exchange
  if (mvp_at_tgt->src_redist != NULL)
    xt_redist_s_exchange(
      mvp_at_tgt->src_redist, (int)(num_src_fields * collection_size),
      (void const**)temp_src_fields, (void**)src_recv_buffer);

  size_t * tgt_pos = mvp_at_tgt->tgt_pos;
  size_t tgt_count = mvp_at_tgt->tgt_count;
  size_t * num_src_per_tgt = mvp_at_tgt->num_src_per_tgt;
  double * weights = mvp_at_tgt->weights;
  size_t * src_field_idx = mvp_at_tgt->src_field_idx;
  size_t * src_idx = mvp_at_tgt->src_idx;

  if (weights != NULL) {

    // apply weights
    for (size_t l = 0; l < collection_size; ++l) {
      for (size_t i = 0, k = 0; i < tgt_count; ++i) {
        size_t curr_num_src_per_tgt = num_src_per_tgt[i];
        double result = 0.0;
        for (size_t j = 0; j < curr_num_src_per_tgt; ++j, ++k) {
          if (src_field_idx[k] == SIZE_MAX) {
            result +=
              src_recv_buffer[l * num_src_fields][src_idx[k]] * weights[k];
          } else {
            result += src_fields[l][src_field_idx[k]][src_idx[k]] * weights[k];
          }
        }
        tgt_field[l][tgt_pos[i]] = result;
      }
    }
  } else {

    // sum source points
    for (size_t l = 0; l < collection_size; ++l) {
      for (size_t i = 0, k = 0; i < tgt_count; ++i) {
        size_t curr_num_src_per_tgt = num_src_per_tgt[i];
        double result = 0.0;
        for (size_t j = 0; j < curr_num_src_per_tgt; ++j, ++k) {
          if (src_field_idx[k] == SIZE_MAX) {
            result += src_recv_buffer[l * num_src_fields][src_idx[k]];
          } else {
            result += src_fields[l][src_field_idx[k]][src_idx[k]];
          }
        }
        tgt_field[l][tgt_pos[i]] = result;
      }
    }
  }
}

// src_fields dimensions [collection_idx]
//                       [field index]
//                       [local_idx]
// tgt_field dimensions [collection_idx]
//                      [local_idx]
void yac_interpolation_execute(
  struct interpolation * interp, double *** src_fields, double ** tgt_field) {

  size_t collection_size = (size_t)(interp->collection_size);

  if (interp->mvp_at_src != NULL)
    yac_interpolation_execute_mvp_at_src(
      interp->mvp_at_src, src_fields, tgt_field, collection_size);

  if (interp->mvp_at_tgt != NULL)
    yac_interpolation_execute_mvp_at_tgt(
      interp->mvp_at_tgt, src_fields, tgt_field, collection_size);

  if (interp->sum_at_src != NULL)
    yac_interpolation_execute_mvp_at_src(
      interp->sum_at_src, src_fields, tgt_field, collection_size);

  if (interp->sum_at_tgt != NULL)
    yac_interpolation_execute_mvp_at_tgt(
      interp->sum_at_tgt, src_fields, tgt_field, collection_size);

  size_t const fixed_count = interp->fixed_count;
  for (size_t l = 0; l < collection_size; ++l) {
    for (size_t i = 0; i < fixed_count; ++i) {
      size_t count = interp->fixed[i].count;
      size_t * pos = interp->fixed[i].pos;
      double value = interp->fixed[i].value;
      for (size_t j = 0; j < count; ++j)
        tgt_field[l][pos[j]] = value;
    }
  }

  if (interp->direct != NULL) {

    double ** temp_src_fields;
    if (interp->direct->is_source) {
      temp_src_fields = interp->direct->src_field_buffer;
      for (size_t i = 0; i < collection_size; ++i)
        temp_src_fields[i] = src_fields[i][0];
    } else {
      temp_src_fields = interp->direct->dummy_recv_buffer;
    }

    double ** dummy_recv_buffer = interp->direct->dummy_recv_buffer;
    if (interp->direct->is_target)
      for (size_t i = 0; i < collection_size; ++i)
        dummy_recv_buffer[i] = tgt_field[i];

    // send source points to the target processes
    if (interp->direct->is_source || interp->direct->is_target)
      xt_redist_s_exchange(
        interp->direct->redist, collection_size,
        (const void **)temp_src_fields, (void **)dummy_recv_buffer);
  }

  if (interp->direct_mf != NULL) {

    double ** temp_src_fields;
    size_t num_src_fields = interp->direct_mf->num_src_fields;
    if (interp->direct_mf->is_source) {
      temp_src_fields = interp->direct_mf->src_field_buffer;
      for (size_t i = 0; i < collection_size; ++i)
        for (size_t j = 0; j < num_src_fields; ++j)
          temp_src_fields[i * num_src_fields + j] = src_fields[i][j];
    } else {
      temp_src_fields = interp->direct_mf->dummy_recv_buffer;
    }

    double ** dummy_recv_buffer = interp->direct_mf->dummy_recv_buffer;
    if (interp->direct_mf->is_target)
      for (size_t i = 0; i < collection_size; ++i)
        for (size_t j = 0; j < num_src_fields; ++j)
          dummy_recv_buffer[i * num_src_fields + j] = tgt_field[i];

    // send source points to the target processes
    if (interp->direct_mf->is_source || interp->direct_mf->is_target)
      xt_redist_s_exchange(
        interp->direct_mf->redist, collection_size * num_src_fields,
        (const void **)temp_src_fields, (void **)dummy_recv_buffer);
  }
}

static void yac_interpolation_execute_put_mvp_at_src(
  struct t_mvp_at_src * mvp_at_src, double *** src_fields,
  size_t collection_size, int is_target) {

  YAC_ASSERT(mvp_at_src->result_buffer_state != BUFFER_WAIT,
    "ERROR(yac_interpolation_execute_put_mvp_at_src: "
    "result buffer has not yet been communicated to target")

  // wait until previous put is completed
  if (mvp_at_src->result_buffer_state == BUFFER_ACTIVE) {
    xt_request_wait(&(mvp_at_src->result_request));
    mvp_at_src->result_buffer_state = BUFFER_UNUSED;
  }

  size_t num_src_fields = mvp_at_src->num_src_fields;
  double ** temp_src_fields = mvp_at_src->src_fields_buffer;
  double ** halo_buffers = mvp_at_src->halo_buffers;
  for (size_t i = 0; i < collection_size; ++i)
    for (size_t j = 0; j < num_src_fields; ++j)
      temp_src_fields[i * num_src_fields + j] = src_fields[i][j];

  // do halo exchange
  if (mvp_at_src->halo_redist != NULL)
    xt_redist_s_exchange(
      mvp_at_src->halo_redist,
      (int)(collection_size * num_src_fields),
      (void const**)temp_src_fields, (void**)halo_buffers);

  size_t tgt_count = mvp_at_src->tgt_count;
  size_t * num_src_per_tgt = mvp_at_src->num_src_per_tgt;
  double * weights = mvp_at_src->weights;
  size_t * src_field_idx = mvp_at_src->src_field_idx;
  size_t * src_idx = mvp_at_src->src_idx;
  double ** results = mvp_at_src->result_buffer;

  if (weights != NULL) {

    // apply weights
    for (size_t l = 0; l < collection_size; ++l) {
      for (size_t i = 0, k = 0; i < tgt_count; ++i) {
        size_t curr_num_src_per_tgt = num_src_per_tgt[i];
        double result = 0.0;
        for (size_t j = 0; j < curr_num_src_per_tgt; ++j, ++k) {
          if (src_field_idx[k] == SIZE_MAX) {
            result +=
              halo_buffers[l * num_src_fields][src_idx[k]] * weights[k];
          } else {
            result +=
              src_fields[l][src_field_idx[k]][src_idx[k]] * weights[k];
          }
        }
        results[l][i] = result;
      }
    }
  } else {
    // sum source points
    for (size_t l = 0; l < collection_size; ++l) {
      for (size_t i = 0, k = 0; i < tgt_count; ++i) {
        size_t curr_num_src_per_tgt = num_src_per_tgt[i];
        double result = 0.0;
        for (size_t j = 0; j < curr_num_src_per_tgt; ++j, ++k) {
          if (src_field_idx[k] == SIZE_MAX) {
            result += halo_buffers[l * num_src_fields][src_idx[k]];
          } else {
            result += src_fields[l][src_field_idx[k]][src_idx[k]];
          }
        }
        results[l][i] = result;
      }
    }
  }

  // if the local process is only source and not a target
  if (!is_target) {
    // redistribute results
    double ** dummy = xmalloc(collection_size * sizeof(*dummy));
    for (size_t i = 0; i < collection_size; ++i) dummy[i] = NULL;
    xt_redist_a_exchange(
      mvp_at_src->result_redist, collection_size,
      (const void **)results, (void **)dummy,
      &(mvp_at_src->result_request));
    free(dummy);
    mvp_at_src->result_buffer_state = BUFFER_ACTIVE;
  } else {
    mvp_at_src->result_buffer_state = BUFFER_WAIT;
  }
}

static void yac_interpolation_execute_put_mvp_at_tgt(
  struct t_mvp_at_tgt * mvp_at_tgt, double *** src_fields,
  size_t collection_size, int is_target) {

  if (mvp_at_tgt->src_redist != NULL) {

    YAC_ASSERT(
      mvp_at_tgt->src_send_buffer_state != BUFFER_WAIT,
      "ERROR(yac_interpolation_execute_put_mvp_at_tgt: "
      "source field buffer has not yet been communicated to target")

    // wait until previous put is completed
    if (mvp_at_tgt->src_send_buffer_state == BUFFER_ACTIVE) {
      xt_request_wait(&(mvp_at_tgt->send_src_request));
      mvp_at_tgt->src_send_buffer_state = BUFFER_UNUSED;
    }

    size_t num_src_fields = mvp_at_tgt->num_src_fields;
    size_t * src_send_buffer_sizes = mvp_at_tgt->src_send_buffer_sizes;
    double ** src_send_buffer = mvp_at_tgt->src_send_buffer;
    double ** src_recv_buffer = mvp_at_tgt->src_recv_buffer;
    if (mvp_at_tgt->is_source)
      for (size_t i = 0; i < collection_size; ++i)
        for (size_t j = 0; j < num_src_fields; ++j)
          memcpy(src_send_buffer[i * num_src_fields + j], src_fields[i][j],
                 src_send_buffer_sizes[j]);

    // if the local process is only source and not a target
    if (!is_target) {
      xt_redist_a_exchange(
        mvp_at_tgt->src_redist, (int)(num_src_fields * collection_size),
        (const void **)src_send_buffer, (void**)src_recv_buffer,
        &(mvp_at_tgt->send_src_request));
      mvp_at_tgt->src_send_buffer_state = BUFFER_ACTIVE;
    } else {
      mvp_at_tgt->src_send_buffer_state = BUFFER_WAIT;
    }
  }
}

// src_fields dimensions [collection_idx]
//                       [field index]
//                       [local_idx]
void yac_interpolation_execute_put(
  struct interpolation * interp, double *** src_fields) {

  size_t collection_size = (size_t)(interp->collection_size);

  if (!interp->is_source) return;

  if (interp->mvp_at_src != NULL)
    yac_interpolation_execute_put_mvp_at_src(
      interp->mvp_at_src, src_fields, collection_size, interp->is_target);

  if (interp->mvp_at_tgt != NULL)
    yac_interpolation_execute_put_mvp_at_tgt(
      interp->mvp_at_tgt, src_fields, collection_size, interp->is_target);

  if (interp->sum_at_src != NULL)
    yac_interpolation_execute_put_mvp_at_src(
      interp->sum_at_src, src_fields, collection_size, interp->is_target);

  if (interp->sum_at_tgt != NULL)
    yac_interpolation_execute_put_mvp_at_tgt(
      interp->sum_at_tgt, src_fields, collection_size, interp->is_target);

  if ((interp->direct != NULL) && (interp->direct->is_source)){

    if (interp->direct->redist != NULL) {

      YAC_ASSERT(
        interp->direct->src_send_buffer_state != BUFFER_WAIT,
        "ERROR(yac_interpolation_execute_put/direct: "
        "source field buffer has not yet been communicated to target")

      // wait until previous put is completed
      if (interp->direct->src_send_buffer_state == BUFFER_ACTIVE) {
        xt_request_wait(&(interp->direct->request_redist));
        interp->direct->src_send_buffer_state = BUFFER_UNUSED;
      }

      double ** src_send_buffer = interp->direct->src_send_buffer;
      size_t src_send_buffer_size = interp->direct->src_send_buffer_size;
      for (size_t i = 0; i < collection_size; ++i)
        memcpy(src_send_buffer[i], src_fields[i][0], src_send_buffer_size);

      // if the local process is only source and not a target
      if (!interp->is_target) {

        xt_redist_a_exchange(
          interp->direct->redist, collection_size,
          (const void **)src_send_buffer,
          (void **)interp->direct->dummy_recv_buffer,
          &(interp->direct->request_redist));
        interp->direct->src_send_buffer_state = BUFFER_ACTIVE;
      } else {
        interp->direct->src_send_buffer_state = BUFFER_WAIT;
      }
    }
  }

  if ((interp->direct_mf != NULL) && (interp->direct_mf->is_source)){

    if (interp->direct_mf->redist != NULL) {

      YAC_ASSERT(
        interp->direct_mf->src_send_buffer_state != BUFFER_WAIT,
        "ERROR(yac_interpolation_execute_put/direct_mf: "
        "source field buffer has not yet been communicated to target")

      // wait until previous put is completed
      if (interp->direct_mf->src_send_buffer_state == BUFFER_ACTIVE) {
        xt_request_wait(&(interp->direct_mf->request_redist));
        interp->direct_mf->src_send_buffer_state = BUFFER_UNUSED;
      }

      double ** src_send_buffer = interp->direct_mf->src_send_buffer;
      size_t * src_send_buffer_sizes =
        interp->direct_mf->src_send_buffer_sizes;
      size_t num_src_fields = interp->direct_mf->num_src_fields;
      for (size_t i = 0; i < collection_size; ++i)
        for (size_t j = 0; j < num_src_fields; ++j)
          memcpy(src_send_buffer[i * num_src_fields + j],
                 src_fields[i][j], src_send_buffer_sizes[j]);

      // if the local process is only source and not a target
      if (!interp->is_target) {

        xt_redist_a_exchange(
          interp->direct_mf->redist, collection_size * num_src_fields,
          (const void **)src_send_buffer,
          (void **)(interp->direct_mf->dummy_recv_buffer),
          &(interp->direct_mf->request_redist));
        interp->direct_mf->src_send_buffer_state = BUFFER_ACTIVE;
      } else {
        interp->direct_mf->src_send_buffer_state = BUFFER_WAIT;
      }
    }
  }
}

static void yac_interpolation_execute_get_mvp_at_src(
  struct t_mvp_at_src * mvp_at_src, double ** tgt_field,
  size_t collection_size) {

  YAC_ASSERT(mvp_at_src->result_buffer_state != BUFFER_ACTIVE,
    "ERROR(yac_interpolation_execute_get_mvp_at_src): "
    "result buffer already being communicated")

  // redistribute results
  xt_redist_a_exchange(
    mvp_at_src->result_redist, collection_size,
    (const void **)(mvp_at_src->result_buffer),
    (void **)tgt_field, &(mvp_at_src->result_request));
  xt_request_wait(&(mvp_at_src->result_request));
  mvp_at_src->result_buffer_state = BUFFER_UNUSED;
}

static void yac_interpolation_execute_get_mvp_at_tgt(
  struct t_mvp_at_tgt * mvp_at_tgt, double ** tgt_field,
  size_t collection_size) {

  YAC_ASSERT(mvp_at_tgt->src_send_buffer_state != BUFFER_ACTIVE,
    "ERROR(yac_interpolation_execute_get_mvp_at_tgt): "
    "source field buffer already being communicated")

  size_t num_src_fields = mvp_at_tgt->num_src_fields;
  double ** src_send_buffer = mvp_at_tgt->src_send_buffer;
  double ** src_recv_buffer = mvp_at_tgt->src_recv_buffer;

  // receive source field data
  if (mvp_at_tgt->src_redist != NULL) {
    Xt_request request_src_field;
    xt_redist_a_exchange(
      mvp_at_tgt->src_redist, (int)(num_src_fields * collection_size),
      (const void **)src_send_buffer, (void**)src_recv_buffer,
      &request_src_field);
    xt_request_wait(&request_src_field);
    mvp_at_tgt->src_send_buffer_state = BUFFER_UNUSED;
  }

  size_t * tgt_pos = mvp_at_tgt->tgt_pos;
  size_t tgt_count = mvp_at_tgt->tgt_count;
  size_t * num_src_per_tgt = mvp_at_tgt->num_src_per_tgt;
  double * weights = mvp_at_tgt->weights;
  size_t * src_field_idx = mvp_at_tgt->src_field_idx;
  size_t * src_idx = mvp_at_tgt->src_idx;

  if (weights != NULL) {

    // apply weights
    for (size_t l = 0; l < collection_size; ++l) {
      for (size_t i = 0, k = 0; i < tgt_count; ++i) {
        size_t curr_num_src_per_tgt = num_src_per_tgt[i];
        double result = 0.0;
        for (size_t j = 0; j < curr_num_src_per_tgt; ++j, ++k) {
          if (src_field_idx[k] == SIZE_MAX) {
            result +=
              src_recv_buffer[l * num_src_fields][src_idx[k]] * weights[k];
          } else {
            result +=
              src_send_buffer[l * num_src_fields + src_field_idx[k]]
                             [src_idx[k]] * weights[k];
          }
        }
        tgt_field[l][tgt_pos[i]] = result;
      }
    }
  } else {

    // sum source
    for (size_t l = 0; l < collection_size; ++l) {
      for (size_t i = 0, k = 0; i < tgt_count; ++i) {
        size_t curr_num_src_per_tgt = num_src_per_tgt[i];
        double result = 0.0;
        for (size_t j = 0; j < curr_num_src_per_tgt; ++j, ++k) {
          if (src_field_idx[k] == SIZE_MAX) {
            result += src_recv_buffer[l * num_src_fields][src_idx[k]];
          } else {
            result += src_send_buffer[l * num_src_fields + src_field_idx[k]]
                                     [src_idx[k]];
          }
        }
        tgt_field[l][tgt_pos[i]] = result;
      }
    }
  }
}

// tgt_field dimensions [collection_idx]
//                      [local_idx]
void yac_interpolation_execute_get(
  struct interpolation * interp, double ** tgt_field) {
  size_t collection_size = interp->collection_size;

  if (!interp->is_target) return;

  if (interp->mvp_at_src != NULL)
    yac_interpolation_execute_get_mvp_at_src(
      interp->mvp_at_src, tgt_field, collection_size);

  if (interp->mvp_at_tgt != NULL)
    yac_interpolation_execute_get_mvp_at_tgt(
      interp->mvp_at_tgt, tgt_field, collection_size);

  if (interp->sum_at_src != NULL)
    yac_interpolation_execute_get_mvp_at_src(
      interp->sum_at_src, tgt_field, collection_size);

  if (interp->sum_at_tgt != NULL)
    yac_interpolation_execute_get_mvp_at_tgt(
      interp->sum_at_tgt, tgt_field, collection_size);

  size_t const fixed_count = interp->fixed_count;
  for (size_t l = 0; l < collection_size; ++l) {
    for (size_t i = 0; i < fixed_count; ++i) {
      size_t count = interp->fixed[i].count;
      size_t * pos = interp->fixed[i].pos;
      double value = interp->fixed[i].value;
      for (size_t j = 0; j < count; ++j)
        tgt_field[l][pos[j]] = value;
    }
  }

  if ((interp->direct != NULL) && (interp->direct->is_target)) {

    YAC_ASSERT(interp->direct->src_send_buffer_state != BUFFER_ACTIVE,
      "ERROR(yac_interpolation_execute_get/direct): "
      "source field buffer already being communicated")

    // redistribute source data
    xt_redist_a_exchange(
      interp->direct->redist, collection_size,
      (const void **)(interp->direct->src_send_buffer),
      (void **)tgt_field, &(interp->direct->request_redist));
    xt_request_wait(&(interp->direct->request_redist));
    interp->direct->src_send_buffer_state = BUFFER_UNUSED;
  }

  if ((interp->direct_mf != NULL) && (interp->direct_mf->is_target)) {

    YAC_ASSERT(
      interp->direct_mf->src_send_buffer_state != BUFFER_ACTIVE,
      "ERROR(yac_interpolation_execute_get/direct_mf): "
      "source field buffer already being communicated")

    double ** dummy_recv_buffer = interp->direct_mf->dummy_recv_buffer;
    size_t num_src_fields = interp->direct_mf->num_src_fields;
    for (size_t i = 0; i < collection_size; ++i)
      for (size_t j = 0; j < num_src_fields; ++j)
        dummy_recv_buffer[i * num_src_fields + j] = tgt_field[i];

    // redistribute source data
    xt_redist_a_exchange(
      interp->direct_mf->redist,
      collection_size * interp->direct_mf->num_src_fields,
      (const void **)(interp->direct_mf->src_send_buffer),
      (void **)dummy_recv_buffer, &(interp->direct_mf->request_redist));
    xt_request_wait(&(interp->direct_mf->request_redist));
    interp->direct_mf->src_send_buffer_state = BUFFER_UNUSED;
  }
}

void yac_interpolation_inc_ref_count(struct interpolation * interpolation) {

  interpolation->ref_count++;
}

void yac_interpolation_delete(struct interpolation * interp) {

  if (interp == NULL) return;

  if(--(interp->ref_count)) return;

  if (interp->mvp_at_src != NULL) {
    xt_request_wait(&(interp->mvp_at_src->result_request));
    if (interp->mvp_at_src->result_redist != NULL)
      xt_redist_delete(interp->mvp_at_src->result_redist);
    free(interp->mvp_at_src->result_buffer[0]);
    free(interp->mvp_at_src->result_buffer);
    free(interp->mvp_at_src->src_idx);
    free(interp->mvp_at_src->src_field_idx);
    free(interp->mvp_at_src->weights);
    free(interp->mvp_at_src->num_src_per_tgt);
    if (interp->mvp_at_src->halo_redist != NULL)
      xt_redist_delete(interp->mvp_at_src->halo_redist);
    free(interp->mvp_at_src->halo_buffers[0]);
    free(interp->mvp_at_src->halo_buffers);
    free(interp->mvp_at_src->halo_sizes);
    free(interp->mvp_at_src->src_fields_buffer);
    free(interp->mvp_at_src);
  }

  if (interp->sum_at_src != NULL) {
    xt_request_wait(&(interp->sum_at_src->result_request));
    if (interp->sum_at_src->result_redist != NULL)
      xt_redist_delete(interp->sum_at_src->result_redist);
    free(interp->sum_at_src->result_buffer[0]);
    free(interp->sum_at_src->result_buffer);
    free(interp->sum_at_src->src_idx);
    free(interp->sum_at_src->src_field_idx);
    free(interp->sum_at_src->num_src_per_tgt);
    if (interp->sum_at_src->halo_redist != NULL)
      xt_redist_delete(interp->sum_at_src->halo_redist);
    free(interp->sum_at_src->halo_buffers[0]);
    free(interp->sum_at_src->halo_buffers);
    free(interp->sum_at_src->halo_sizes);
    free(interp->sum_at_src->src_fields_buffer);
    free(interp->sum_at_src);
  }

  if (interp->mvp_at_tgt != NULL) {

    if (interp->mvp_at_tgt->src_redist != NULL)
      xt_redist_delete(interp->mvp_at_tgt->src_redist);
    xt_request_wait(&(interp->mvp_at_tgt->send_src_request));
    free(interp->mvp_at_tgt->tgt_pos);
    free(interp->mvp_at_tgt->src_send_buffer[0]);
    free(interp->mvp_at_tgt->src_send_buffer);
    free(interp->mvp_at_tgt->src_recv_buffer[0]);
    free(interp->mvp_at_tgt->src_recv_buffer);
    free(interp->mvp_at_tgt->src_fields_buffer);
    free(interp->mvp_at_tgt->src_send_buffer_sizes);
    free(interp->mvp_at_tgt->src_recv_buffer_sizes);
    free(interp->mvp_at_tgt->src_idx);
    free(interp->mvp_at_tgt->src_field_idx);
    free(interp->mvp_at_tgt->weights);
    free(interp->mvp_at_tgt->num_src_per_tgt);
    free(interp->mvp_at_tgt);
  }

  if (interp->sum_at_tgt != NULL) {

    if (interp->sum_at_tgt->src_redist != NULL)
      xt_redist_delete(interp->sum_at_tgt->src_redist);
    xt_request_wait(&(interp->sum_at_tgt->send_src_request));
    free(interp->sum_at_tgt->tgt_pos);
    free(interp->sum_at_tgt->src_send_buffer[0]);
    free(interp->sum_at_tgt->src_send_buffer);
    free(interp->sum_at_tgt->src_recv_buffer[0]);
    free(interp->sum_at_tgt->src_recv_buffer);
    free(interp->sum_at_tgt->src_fields_buffer);
    free(interp->sum_at_tgt->src_send_buffer_sizes);
    free(interp->sum_at_tgt->src_recv_buffer_sizes);
    free(interp->sum_at_tgt->src_idx);
    free(interp->sum_at_tgt->src_field_idx);
    free(interp->sum_at_tgt->num_src_per_tgt);
    free(interp->sum_at_tgt);
  }

  if (interp->direct != NULL) {

    if (interp->direct->redist != NULL)
      xt_redist_delete(interp->direct->redist);
    xt_request_wait(&(interp->direct->request_redist));
    free(interp->direct->src_send_buffer[0]);
    free(interp->direct->src_send_buffer);
    free(interp->direct->src_field_buffer);
    free(interp->direct->dummy_recv_buffer);
    free(interp->direct);
  }

  if (interp->direct_mf != NULL) {

    if (interp->direct_mf->redist != NULL)
      xt_redist_delete(interp->direct_mf->redist);
    xt_request_wait(&(interp->direct_mf->request_redist));
    free(interp->direct_mf->src_send_buffer[0]);
    free(interp->direct_mf->src_send_buffer);
    free(interp->direct_mf->src_send_buffer_sizes);
    free(interp->direct_mf->src_field_buffer);
    free(interp->direct_mf->dummy_recv_buffer);
    free(interp->direct_mf);
  }

  for (size_t i = 0; i < interp->fixed_count; ++i)
    free(interp->fixed[i].pos);
  free(interp->fixed);
  free(interp);
}
