!>
!! @file yac_finterface.f90
!! @brief Implementation of the YAC Fortran user interface
!!
!! @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
!!                                 Rene Redler <rene.redler@mpimet.mpg.de>
!!
!! @author Moritz Hanke <hanke@dkrz.de>
!!         Rene Redler <rene.redler@mpimet.mpg.de>
!!
!! @par Warranty
!!
!! This code has been tested up to a certain level. Defects and weaknesses,
!! which may be included in the code, do not establish any warranties by the
!! authors. The authors do not make any warranty, express or implied, or assume any
!! liability or responsibility for the use, acquisition or application of this
!! software.
!!
!! @par Fortran User API
!!
!! For a detailed description of the Fortran interface please see
!! \ref finterface. Information on the C interface can be found here
!! \ref cinterface.
!!

!
! Keywords:
! Maintainer: Moritz Hanke <hanke@dkrz.de>
!             Rene Redler <rene.redler@mpimet.mpg.de>
! URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
!
! This file is part of YAC.
!
! Redistribution and use in source and binary forms, with or without
! modification, are  permitted provided that the following conditions are
! met:
!
! Redistributions of source code must retain the above copyright notice,
! this list of conditions and the following disclaimer.
!
! Redistributions in binary form must reproduce the above copyright
! notice, this list of conditions and the following disclaimer in the
! documentation and/or other materials provided with the distribution.
!
! Neither the name of the DKRZ GmbH nor the names of its contributors
! may be used to endorse or promote products derived from this software
! without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
! IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
! TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
! OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
! LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!

!
! -------------------------- init -------------------------------------

subroutine yac_finit_comm ( xml_filename, schema_filename, world_comm )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_finit_comm

  implicit none

  interface

      subroutine yac_cinit_comm ( xml_filename, schema_filename, world_comm ) &
        bind ( c, name='yac_cinit_comm_f2c' )

        use, intrinsic :: iso_c_binding, only : c_char
        use mo_yac_finterface, only : YAC_MPI_FINT_KIND

        character ( kind=c_char), dimension(*)    :: xml_filename
        character ( kind=c_char), dimension(*)    :: schema_filename
        integer ( kind=YAC_MPI_FINT_KIND ), value :: world_comm

      end subroutine yac_cinit_comm

  end interface

  character(len=*), intent(in) :: xml_filename    !< [IN] name of xml configuration file
  character(len=*), intent(in) :: schema_filename !< [IN] corresponding xsd file
  integer, intent(in)          :: world_comm      !< [IN] MPI world communicator (optional)

  call yac_check_strlength ( xml_filename )
  call yac_check_strlength ( schema_filename )

  call yac_cinit_comm ( TRIM(xml_filename)    // c_null_char, &
                        TRIM(schema_filename) // c_null_char, &
                        world_comm )

end subroutine yac_finit_comm

subroutine yac_finit_comm_group ( xml_filename,    &
                                  schema_filename, &
                                  world_comm,      &
                                  group_name )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_finit_comm_group

  implicit none

  interface

      subroutine yac_cinit_comm_group ( &
        xml_filename, schema_filename, world_comm, group_name ) &
        bind ( c, name='yac_cinit_comm_group_f2c' )

        use, intrinsic :: iso_c_binding, only : c_char
        use mo_yac_finterface, only : YAC_MPI_FINT_KIND

        character ( kind=c_char), dimension(*)    :: xml_filename
        character ( kind=c_char), dimension(*)    :: schema_filename
        integer ( kind=YAC_MPI_FINT_KIND ), value :: world_comm
        character ( kind=c_char), dimension(*)    :: group_name

      end subroutine yac_cinit_comm_group

  end interface

  character(len=*), intent(in) :: xml_filename    !< [IN] name of xml configuration file
  character(len=*), intent(in) :: schema_filename !< [IN] corresponding xsd file
  integer, intent(in)          :: world_comm      !< [IN] MPI world communicator (optional)
  character(len=*), intent(in) :: group_name      !< [IN] group name

  call yac_check_strlength ( xml_filename )
  call yac_check_strlength ( schema_filename )
  call yac_check_strlength ( group_name )

  call yac_cinit_comm_group ( TRIM(xml_filename)    // c_null_char, &
                              TRIM(schema_filename) // c_null_char, &
                              world_comm, &
                              TRIM(group_name)      // c_null_char )

end subroutine yac_finit_comm_group

subroutine yac_finit_comm_dummy ( world_comm )

  use mo_yac_finterface, dummy => yac_finit_comm_dummy

  implicit none

  interface

      subroutine yac_cinit_comm_dummy ( world_comm ) &
        bind ( c, name='yac_cinit_comm_dummy_f2c' )

        use mo_yac_finterface, only : YAC_MPI_FINT_KIND

        integer ( kind=YAC_MPI_FINT_KIND ), value :: world_comm

      end subroutine yac_cinit_comm_dummy

  end interface

  integer, intent(in) :: world_comm !< [IN] MPI world communicator (optional)

  call yac_cinit_comm_dummy ( world_comm )

end subroutine yac_finit_comm_dummy

subroutine yac_finit_comm_instance( xml_filename,    &
                                    schema_filename, &
                                    world_comm,      &
                                    yac_instance_id)

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_finit_comm_instance

  implicit none

  interface

      subroutine yac_cinit_comm_instance ( xml_filename,    &
                                           schema_filename, &
                                           world_comm,      &
                                           yac_instance_id) &
        bind ( c, name='yac_cinit_comm_instance_f2c' )

        use, intrinsic :: iso_c_binding, only : c_char, c_int
        use mo_yac_finterface, only : YAC_MPI_FINT_KIND

        character ( kind=c_char), dimension(*)    :: xml_filename
        character ( kind=c_char), dimension(*)    :: schema_filename
        integer ( kind=YAC_MPI_FINT_KIND ), value :: world_comm
        integer (kind=c_int)                      :: yac_instance_id

      end subroutine yac_cinit_comm_instance

  end interface

  character(len=*), intent(in) :: xml_filename    !< [IN] name of xml configuration file
  character(len=*), intent(in) :: schema_filename !< [IN] corresponding xsd file
  integer, intent(in)          :: world_comm      !< [IN] MPI world communicator (optional)
  integer, intent(out)         :: yac_instance_id !< [OUT] returned handle to the YAC instance

  call yac_check_strlength ( xml_filename )
  call yac_check_strlength ( schema_filename )

  call yac_cinit_comm_instance ( TRIM(xml_filename)    // c_null_char, &
                                 TRIM(schema_filename) // c_null_char, &
                                 world_comm,                           &
                                yac_instance_id )

end subroutine yac_finit_comm_instance

subroutine yac_finit_comm_group_instance( xml_filename,    &
                                          schema_filename, &
                                          world_comm,      &
                                          group_name,      &
                                          yac_instance_id)

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_finit_comm_group_instance

  implicit none

  interface

      subroutine yac_cinit_comm_group_instance ( xml_filename,    &
                                                 schema_filename, &
                                                 world_comm,      &
                                                 group_name,      &
                                                 yac_instance_id) &
        bind ( c, name='yac_cinit_comm_group_instance_f2c' )

        use, intrinsic :: iso_c_binding, only : c_char, c_int
        use mo_yac_finterface, only : YAC_MPI_FINT_KIND

        character ( kind=c_char), dimension(*)    :: xml_filename
        character ( kind=c_char), dimension(*)    :: schema_filename
        integer ( kind=YAC_MPI_FINT_KIND ), value :: world_comm
        character ( kind=c_char), dimension(*)    :: group_name
        integer (kind=c_int)                      :: yac_instance_id

      end subroutine yac_cinit_comm_group_instance

  end interface

  character(len=*), intent(in) :: xml_filename    !< [IN] name of xml configuration file
  character(len=*), intent(in) :: schema_filename !< [IN] corresponding xsd file
  integer, intent(in)          :: world_comm      !< [IN] MPI world communicator (optional)
  character(len=*), intent(in) :: group_name      !< [IN] group name
  integer, intent(out)         :: yac_instance_id !< [OUT] returned handle to the YAC instance

  call yac_check_strlength ( xml_filename )
  call yac_check_strlength ( schema_filename )
  call yac_check_strlength ( group_name )

  call yac_cinit_comm_group_instance ( TRIM(xml_filename)    // c_null_char, &
                                       TRIM(schema_filename) // c_null_char, &
                                       world_comm,                           &
                                       TRIM(group_name)      // c_null_char, &
                                       yac_instance_id )

end subroutine yac_finit_comm_group_instance

subroutine yac_finit ( xml_filename, schema_filename )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_finit

  implicit none

  interface

    subroutine yac_cinit ( xml_filename, schema_filename ) &
      bind ( c, name='yac_cinit' )

      use, intrinsic :: iso_c_binding, only : c_char

      character (kind=c_char), dimension(*) :: xml_filename
      character (kind=c_char), dimension(*) :: schema_filename

    end subroutine yac_cinit

  end interface

  character(len=*), intent(in)  :: xml_filename    !< [IN] name of xml configuration file
  character(len=*), intent(in)  :: schema_filename !< [IN] corresponding xsd file

  call yac_check_strlength ( xml_filename )
  call yac_check_strlength ( schema_filename )

  call yac_cinit ( TRIM(xml_filename)    // c_null_char, &
                   TRIM(schema_filename) // c_null_char )

end subroutine yac_finit

subroutine yac_finit_group ( xml_filename, schema_filename, group_name )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_finit_group

  implicit none

  interface

    subroutine yac_cinit_group ( xml_filename, schema_filename, group_name ) &
      bind ( c, name='yac_cinit_group' )

      use, intrinsic :: iso_c_binding, only : c_char

      character (kind=c_char), dimension(*) :: xml_filename
      character (kind=c_char), dimension(*) :: schema_filename
      character (kind=c_char), dimension(*) :: group_name

    end subroutine yac_cinit_group

  end interface

  character(len=*), intent(in) :: xml_filename    !< [IN] name of xml configuration file
  character(len=*), intent(in) :: schema_filename !< [IN] corresponding xsd file
  character(len=*), intent(in) :: group_name      !< [IN] group name

  call yac_check_strlength ( xml_filename )
  call yac_check_strlength ( schema_filename )
  call yac_check_strlength ( group_name )

  call yac_cinit_group ( TRIM(xml_filename)    // c_null_char, &
                         TRIM(schema_filename) // c_null_char, &
                         TRIM(group_name)      // c_null_char )

end subroutine yac_finit_group

subroutine yac_finit_dummy ( )

  use mo_yac_finterface, dummy => yac_finit_dummy

  implicit none

  interface

      subroutine yac_cinit_dummy ( ) &
        bind ( c, name='yac_cinit_dummy' )

      end subroutine yac_cinit_dummy

  end interface

  call yac_cinit_dummy ( )

end subroutine yac_finit_dummy

subroutine yac_finit_instance ( xml_filename, schema_filename, yac_instance_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_finit_instance

  implicit none

  interface

      subroutine yac_cinit_instance ( xml_filename, &
                                      schema_filename, &
                                      yac_instance_id ) &
        bind ( c, name='yac_cinit_instance' )

        use, intrinsic :: iso_c_binding, only : c_char, c_int

        character (kind=c_char), dimension(*) :: xml_filename
        character (kind=c_char), dimension(*) :: schema_filename
        integer (kind=c_int)                  :: yac_instance_id

      end subroutine yac_cinit_instance

  end interface

  character(len=*), intent(in) :: xml_filename    !< [IN] name of xml configuration file
  character(len=*), intent(in) :: schema_filename !< [IN] corresponding xsd file
  integer, intent(out)         :: yac_instance_id !< [OUT] returned handle to the YAC instance

  call yac_check_strlength ( xml_filename )
  call yac_check_strlength ( schema_filename )

  call yac_cinit_instance ( TRIM(xml_filename)    // c_null_char, &
                            TRIM(schema_filename) // c_null_char, &
                            yac_instance_id )

end subroutine yac_finit_instance

subroutine yac_finit_group_instance ( xml_filename,    &
                                      schema_filename, &
                                      group_name,      &
                                      yac_instance_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_finit_group_instance

  implicit none

  interface

    subroutine yac_cinit_group_instance ( xml_filename,     &
                                          schema_filename,  &
                                          group_name,       &
                                          yac_instance_id ) &
      bind ( c, name='yac_cinit_group_instance' )

      use, intrinsic :: iso_c_binding, only : c_char, c_int

      character (kind=c_char), dimension(*) :: xml_filename
      character (kind=c_char), dimension(*) :: schema_filename
      character (kind=c_char), dimension(*) :: group_name
      integer (kind=c_int)                  :: yac_instance_id

    end subroutine yac_cinit_group_instance

  end interface

  character(len=*), intent(in) :: xml_filename    !< [IN] name of xml configuration file
  character(len=*), intent(in) :: schema_filename !< [IN] corresponding xsd file
  character(len=*), intent(in) :: group_name      !< [IN] group name
  integer, intent(out)         :: yac_instance_id !< [OUT] returned handle to the YAC instance

  call yac_check_strlength ( xml_filename )
  call yac_check_strlength ( schema_filename )
  call yac_check_strlength ( group_name )

  call yac_cinit_group_instance ( TRIM(xml_filename)    // c_null_char, &
                                  TRIM(schema_filename) // c_null_char, &
                                  TRIM(group_name)      // c_null_char, &
                                  yac_instance_id )

end subroutine yac_finit_group_instance

! -------------------------- cleanup -----------------------------------

subroutine yac_fcleanup ( )

  use mo_yac_finterface, dummy => yac_fcleanup

  implicit none

  interface

      subroutine yac_ccleanup () bind ( c, name='yac_ccleanup' )
      end subroutine yac_ccleanup

  end interface

  call yac_ccleanup ( )

end subroutine yac_fcleanup

subroutine yac_fcleanup_instance ( yac_instance_id )

  use mo_yac_finterface, dummy => yac_fcleanup_instance

  implicit none

  interface

      subroutine yac_ccleanup_instance ( yac_instance_id ) &
        bind ( c, name='yac_ccleanup_instance' )

        use, intrinsic :: iso_c_binding, only : c_int

        integer (kind=c_int), value :: yac_instance_id

      end subroutine yac_ccleanup_instance

  end interface

  integer, intent(in) :: yac_instance_id !< [IN] YAC instance identifier

  call yac_ccleanup_instance ( yac_instance_id )

end subroutine yac_fcleanup_instance

! -------------------------- final -------------------------------------

subroutine yac_ffinalize ( )

  use mo_yac_finterface, dummy => yac_ffinalize

  implicit none

  interface
     subroutine yac_cfinalize () bind ( c, name='yac_cfinalize' )
     end subroutine yac_cfinalize
  end interface

  call yac_cfinalize ( )

end subroutine yac_ffinalize

subroutine yac_ffinalize_instance ( yac_instance_id )

  use mo_yac_finterface, dummy => yac_ffinalize_instance

  implicit none

  interface
     subroutine yac_cfinalize_instance ( yac_instance_id ) &
        bind ( c, name='yac_cfinalize_instance' )

        use, intrinsic :: iso_c_binding, only : c_int

        integer (kind=c_int), value :: yac_instance_id

     end subroutine yac_cfinalize_instance
  end interface

  integer, intent(in) :: yac_instance_id !< [IN] YAC instance identifier

  call yac_cfinalize_instance ( yac_instance_id )

end subroutine yac_ffinalize_instance

! -------------------------- version ----------------------------------

function yac_fget_version () result (version_string)

  use, intrinsic :: iso_c_binding, only : c_ptr, c_char, c_null_char, c_f_pointer
  use mo_yac_finterface, dummy => yac_fget_version

  implicit none

  interface
     function yac_cget_version () bind ( c, name='yac_cget_version' )

     use, intrinsic :: iso_c_binding, only : c_ptr
     type(c_ptr) :: yac_cget_version

     end function yac_cget_version
  end interface

  type (c_ptr)                                           :: c_string_ptr
  character (len=1, kind=c_char), dimension (:), pointer :: version_chars => null ()
  character (len=YAC_MAX_CHARLEN)                        :: version_string
  integer :: i

  c_string_ptr = yac_cget_version()

  CALL c_f_pointer ( c_string_ptr, version_chars, [ YAC_MAX_CHARLEN ])

  version_string = " "

  do i=1, YAC_MAX_CHARLEN
    if (version_chars(i) == c_null_char) exit
    version_string(i:i) = version_chars(i)
  end do

end function yac_fget_version

! -------------------------- dates ------------------------------------

subroutine yac_fdef_datetime ( start_datetime, end_datetime )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_datetime

  implicit none

  interface

     subroutine yac_cdef_datetime ( start_datetime, end_datetime ) &
    &      bind ( c, name='yac_cdef_datetime' )

       use, intrinsic :: iso_c_binding, only : c_char

       character ( kind=c_char), dimension(*) :: start_datetime
       character ( kind=c_char), dimension(*) :: end_datetime

     end subroutine yac_cdef_datetime

  end interface

  character(len=*), intent(in), optional :: start_datetime !< [IN] start datetime of job
  character(len=*), intent(in), optional :: end_datetime   !< [IN] end datetime of job

  integer :: index

  index = 0


  if (present(start_datetime)) then
    call yac_check_strlength ( start_datetime )
    index = index + 1
  end if

  if (present(end_datetime)) then
    call yac_check_strlength ( end_datetime )
    index = index + 2
  end if

  select case ( index )

    case ( 3 )
      call yac_cdef_datetime ( TRIM(start_datetime) // c_null_char, &
                               TRIM(end_datetime)   // c_null_char )
    case ( 2 )
      call yac_cdef_datetime ( c_null_char, &
                               TRIM(end_datetime)   // c_null_char )
    case ( 1 )
      call yac_cdef_datetime ( TRIM(start_datetime) // c_null_char, &
                               c_null_char )
    end select

end subroutine yac_fdef_datetime

subroutine yac_fdef_datetime_instance ( &
  yac_instance_id, start_datetime, end_datetime )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_datetime_instance

  implicit none

  interface

     subroutine yac_cdef_datetime_instance ( yac_instance_id, &
                                             start_datetime,  &
                                             end_datetime )   &
       bind ( c, name='yac_cdef_datetime_instance' )

       use, intrinsic :: iso_c_binding, only : c_char, c_int

       integer (kind=c_int), value            :: yac_instance_id
       character ( kind=c_char), dimension(*) :: start_datetime
       character ( kind=c_char), dimension(*) :: end_datetime

     end subroutine yac_cdef_datetime_instance

  end interface

  integer, intent(in)                    :: yac_instance_id !< [IN] YAC instance identifier
  character(len=*), intent(in), optional :: start_datetime  !< [IN] start datetime of job
  character(len=*), intent(in), optional :: end_datetime    !< [IN] end datetime of job

  integer :: index

  index = 0


  if (present(start_datetime)) then
    call yac_check_strlength ( start_datetime )
    index = index + 1
  end if

  if (present(end_datetime)) then
    call yac_check_strlength ( end_datetime )
    index = index + 2
  end if

  select case ( index )

    case ( 3 )
      call yac_cdef_datetime_instance ( yac_instance_id, &
                                        TRIM(start_datetime) // c_null_char, &
                                        TRIM(end_datetime)   // c_null_char )
    case ( 2 )
      call yac_cdef_datetime_instance ( yac_instance_id, &
                                        c_null_char, &
                                        TRIM(end_datetime)   // c_null_char )
    case ( 1 )
      call yac_cdef_datetime_instance ( yac_instance_id, &
                                        TRIM(start_datetime) // c_null_char, &
                                         c_null_char )
    end select

end subroutine yac_fdef_datetime_instance

! ------------------------ get_groupcomm -------------------------------

subroutine yac_fget_groupcomm ( group_communicator )

   use mo_yac_finterface, dummy => yac_fget_groupcomm

   implicit none

   interface

     subroutine yac_get_groupcomm ( group_communicator ) &
       bind ( c, name='yac_cget_groupcomm_c2f' )

       use mo_yac_finterface, only : YAC_MPI_FINT_KIND

       integer ( kind=YAC_MPI_FINT_KIND ) :: group_communicator

     end subroutine yac_get_groupcomm

   end interface

   integer, intent (out) :: group_communicator !< [OUT] group MPI communicator


   call yac_get_groupcomm ( group_communicator )

end subroutine yac_fget_groupcomm

subroutine yac_fget_groupcomm_instance ( yac_instance_id, group_communicator )

   use mo_yac_finterface, dummy => yac_fget_groupcomm_instance

   implicit none

   interface

     subroutine yac_get_groupcomm_instance ( yac_instance_id, &
                                             group_communicator ) &
       bind ( c, name='yac_cget_groupcomm_instance_c2f' )

       use, intrinsic :: iso_c_binding, only : c_int
       use mo_yac_finterface, only : YAC_MPI_FINT_KIND

       integer ( kind=c_int ), value      :: yac_instance_id
       integer ( kind=YAC_MPI_FINT_KIND ) :: group_communicator

     end subroutine yac_get_groupcomm_instance

   end interface

   integer, intent (in)  :: yac_instance_id    !< [IN]  YAC instance identifier
   integer, intent (out) :: group_communicator !< [OUT] group MPI communicator

   call yac_get_groupcomm_instance ( yac_instance_id, group_communicator )

end subroutine yac_fget_groupcomm_instance

! ------------------------ def_comp ------------------------------------

subroutine yac_fdef_comp ( comp_name, comp_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_comp

  implicit none

  interface

     subroutine yac_cdef_comp ( comp_name, comp_id ) &
       bind ( c, name='yac_cdef_comp' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char), dimension(*) :: comp_name
       integer   ( kind=c_int )               :: comp_id

     end subroutine yac_cdef_comp

  end interface

  character(len=*), intent(in) :: comp_name !< [IN]  component name
  integer, intent(out)         :: comp_id   !< [OUT] returned handle to the component

  call yac_check_strlength ( comp_name )

  call yac_cdef_comp ( TRIM(comp_name) // c_null_char, comp_id )

end subroutine yac_fdef_comp

subroutine yac_fdef_comp_instance ( yac_instance_id, comp_name, comp_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_comp_instance

  implicit none

  interface

     subroutine yac_cdef_comp_instance ( yac_instance_id, &
                                         comp_name,       &
                                         comp_id ) &
       bind ( c, name='yac_cdef_comp_instance' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer (kind=c_int), value            :: yac_instance_id
       character ( kind=c_char), dimension(*) :: comp_name
       integer   ( kind=c_int )               :: comp_id

     end subroutine yac_cdef_comp_instance

  end interface

  integer, intent(in)          :: yac_instance_id !< [IN]  YAC instance identifier
  character(len=*), intent(in) :: comp_name       !< [IN]  component name
  integer, intent(out)         :: comp_id         !< [OUT] returned handle to the component

  call yac_check_strlength ( comp_name )

  call yac_cdef_comp_instance( yac_instance_id,                &
                               TRIM(comp_name) // c_null_char, &
                               comp_id )

end subroutine yac_fdef_comp_instance

! ------------------------ def_comps ------------------------------------

subroutine yac_fdef_comps ( comp_names, num_comps, comp_ids )

  use, intrinsic :: iso_c_binding, only : c_null_char, c_ptr, c_loc, c_char
  use mo_yac_finterface, dummy => yac_fdef_comps

  implicit none

  interface

     subroutine yac_cdef_comps ( comp_names, num_comps, comp_ids ) &
    &      bind ( c, name='yac_cdef_comps' )

       use, intrinsic :: iso_c_binding, only : c_int, c_ptr

       type ( c_ptr )                :: comp_names(*)
       integer ( kind=c_int ), value :: num_comps
       integer ( kind=c_int )        :: comp_ids(*)

     end subroutine yac_cdef_comps

  end interface

  integer, intent(in)   :: num_comps !< [IN]  number of components
  character(kind=c_char, len=*), intent(in) :: &
                           comp_names(num_comps) !< [IN]  component names
  integer, intent(out)  :: comp_ids(num_comps) !< [OUT] returned handle to the components

  integer :: i, j
  character(kind=c_char), target :: comp_names_cpy(YAC_MAX_CHARLEN+1, num_comps)
  type(c_ptr) :: comp_name_ptrs(num_comps)

  comp_names_cpy = c_null_char

  do i = 1, num_comps
    do j = 1, len_trim(comp_names(i))
      comp_names_cpy(j,i) = comp_names(i)(j:j)
    end do
    comp_name_ptrs(i) = c_loc(comp_names_cpy(1,i));
  end do

  call yac_cdef_comps ( comp_name_ptrs, num_comps, comp_ids )

end subroutine yac_fdef_comps

subroutine yac_fdef_comps_instance ( yac_instance_id, &
                                    comp_names,       &
                                    num_comps,        &
                                    comp_ids )

  use, intrinsic :: iso_c_binding, only : c_null_char, c_ptr, c_loc, c_char
  use mo_yac_finterface, dummy => yac_fdef_comps_instance

  implicit none

  interface

     subroutine yac_cdef_comps_instance ( yac_instance_id, &
                                          comp_names,      &
                                          num_comps,       &
                                          comp_ids ) &
       bind ( c, name='yac_cdef_comps_instance' )

       use, intrinsic :: iso_c_binding, only : c_int, c_ptr

       integer (kind=c_int), value   :: yac_instance_id
       type ( c_ptr )                :: comp_names(*)
       integer ( kind=c_int ), value :: num_comps
       integer ( kind=c_int )        :: comp_ids(*)

     end subroutine yac_cdef_comps_instance

  end interface

  integer, intent(in)   :: yac_instance_id       !< [IN]  YAC instance identifier
  integer, intent(in)   :: num_comps             !< [IN]  number of components
  character(kind=c_char, len=*), intent(in) :: &
                           comp_names(num_comps) !< [IN]  component names
  integer, intent(out)  :: comp_ids(num_comps)   !< [OUT] returned handle to the components

  integer :: i, j
  character(kind=c_char), target :: comp_names_cpy(YAC_MAX_CHARLEN+1, num_comps)
  type(c_ptr) :: comp_name_ptrs(num_comps)

  comp_names_cpy = c_null_char

  do i = 1, num_comps
    do j = 1, len_trim(comp_names(i))
      comp_names_cpy(j,i) = comp_names(i)(j:j)
    end do
    comp_name_ptrs(i) = c_loc(comp_names_cpy(1,i));
  end do

  call yac_cdef_comps_instance ( yac_instance_id, &
                                 comp_name_ptrs,  &
                                 num_comps,       &
                                 comp_ids )

end subroutine yac_fdef_comps_instance


! ------------------------- def_points ----------------------------------

subroutine yac_fdef_points_reg2d_r4 ( grid_id,       &
                                      nbr_points,    &
                                      location,      &
                                      x_points_real, &
                                      y_points_real, &
                                      point_id )

  use mo_yac_finterface, dummy => yac_fdef_points_reg2d_r4

  implicit none

  integer, intent(in)  :: grid_id                      !< [IN]  grid identifier
  integer, intent(in)  :: nbr_points(2)                !< [IN]  number of points in x and y
  integer, intent(in)  :: location                     !< [IN]  location, one of center/edge/vertex
  real, intent(in)     :: x_points_real(nbr_points(1)) !< [IN]  longitudes of points
  real, intent(in)     :: y_points_real(nbr_points(2)) !< [IN]  latitudes of points
  integer, intent(out) :: point_id                     !< [OUT] point identifier

  double precision    ::  x_points(nbr_points(1))
  double precision    ::  y_points(nbr_points(2))

  x_points(:) = dble(x_points_real(:))
  y_points(:) = dble(y_points_real(:))

  call yac_fdef_points_reg2d_r8 ( grid_id,    &
                                  nbr_points, &
                                  location,   &
                                  x_points,   &
                                  y_points,   &
                                  point_id )

end subroutine yac_fdef_points_reg2d_r4

subroutine yac_fdef_points_reg2d_r8 ( grid_id,    &
                                      nbr_points, &
                                      location,   &
                                      x_points,   &
                                      y_points,   &
                                      point_id )

  use mo_yac_finterface, dummy => yac_fdef_points_reg2d_r8

  implicit none

  interface

     subroutine yac_cdef_points_reg2d ( grid_id,      &
                                        nbr_points,   &
                                        location,     &
                                        x_points,     &
                                        y_points,     &
                                        point_id )    &
    &      bind ( c, name='yac_cdef_points_reg2d' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int ), value :: grid_id
       integer ( kind=c_int )        :: nbr_points(2)
       integer ( kind=c_int ), value :: location

       real    ( kind=c_double )     :: x_points(nbr_points(1))
       real    ( kind=c_double )     :: y_points(nbr_points(2))

       integer ( kind=c_int )        :: point_id

     end subroutine yac_cdef_points_reg2d

  end interface

  integer, intent(in)  :: grid_id                         !< [IN]  grid identifier
  integer, intent(in)  :: nbr_points(2)                   !< [IN]  number of points in x and y
  integer, intent(in)  :: location                        !< [IN]  location, one of center/edge/vertex

  double precision, intent(in) :: x_points(nbr_points(1)) !< [IN]  longitudes of points
  double precision, intent(in) :: y_points(nbr_points(2)) !< [IN]  latitudes of points

  integer, intent(out) :: point_id                        !< [OUT] point identifier

  call yac_cdef_points_reg2d ( grid_id,    &
                               nbr_points, &
                               location,   &
                               x_points,   &
                               y_points,   &
                               point_id )

end subroutine yac_fdef_points_reg2d_r8

subroutine yac_fdef_points_curve2d_r4 ( grid_id,       &
                                        nbr_points,    &
                                        location,      &
                                        x_points_real, &
                                        y_points_real, &
                                        point_id )

  use mo_yac_finterface, dummy => yac_fdef_points_curve2d_r4

  implicit none

  integer, intent(in)  :: grid_id                      !< [IN]  grid identifier
  integer, intent(in)  :: nbr_points(2)                !< [IN]  number of points in x and y
  integer, intent(in)  :: location                     !< [IN]  location, one of center/edge/vertex
  real, intent(in)     :: &
    x_points_real(nbr_points(1),nbr_points(2))         !< [IN]  longitudes of points
  real, intent(in)     :: &
    y_points_real(nbr_points(1),nbr_points(2))         !< [IN]  latitudes of points
  integer, intent(out) :: point_id                     !< [OUT] point identifier

  double precision    ::  x_points(nbr_points(1),nbr_points(2))
  double precision    ::  y_points(nbr_points(1),nbr_points(2))

  x_points(:,:) = dble(x_points_real(:,:))
  y_points(:,:) = dble(y_points_real(:,:))

  call yac_fdef_points_curve2d_r8 ( grid_id,    &
                                    nbr_points, &
                                    location,   &
                                    x_points,   &
                                    y_points,   &
                                    point_id )

end subroutine yac_fdef_points_curve2d_r4

subroutine yac_fdef_points_curve2d_r8 ( grid_id,    &
                                        nbr_points, &
                                        location,   &
                                        x_points,   &
                                        y_points,   &
                                        point_id )

  use mo_yac_finterface, dummy => yac_fdef_points_curve2d_r8

  implicit none

  interface

     subroutine yac_cdef_points_curve2d ( grid_id,      &
                                        nbr_points,   &
                                        location,     &
                                        x_points,     &
                                        y_points,     &
                                        point_id )    &
    &      bind ( c, name='yac_cdef_points_curve2d' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int ), value :: grid_id
       integer ( kind=c_int )        :: nbr_points(2)
       integer ( kind=c_int ), value :: location

       real    ( kind=c_double )     :: x_points(nbr_points(1),nbr_points(2))
       real    ( kind=c_double )     :: y_points(nbr_points(1),nbr_points(2))

       integer ( kind=c_int )        :: point_id

     end subroutine yac_cdef_points_curve2d

  end interface

  integer, intent(in)  :: grid_id                         !< [IN]  grid identifier
  integer, intent(in)  :: nbr_points(2)                   !< [IN]  number of points in x and y
  integer, intent(in)  :: location                        !< [IN]  location, one of center/edge/vertex

  double precision, intent(in) :: &
    x_points(nbr_points(1),nbr_points(2))                 !< [IN]  longitudes of points
  double precision, intent(in) :: &
    y_points(nbr_points(1),nbr_points(2))                 !< [IN]  latitudes of points

  integer, intent(out) :: point_id                        !< [OUT] point identifier

  call yac_cdef_points_curve2d ( grid_id,    &
                                 nbr_points, &
                                 location,   &
                                 x_points,   &
                                 y_points,   &
                                 point_id )

end subroutine yac_fdef_points_curve2d_r8

subroutine yac_fdef_points_unstruct_r4 ( grid_id,       &
                                         nbr_points,    &
                                         location,      &
                                         x_points_real, &
                                         y_points_real, &
                                         point_id )

  use mo_yac_finterface, dummy => yac_fdef_points_unstruct_r4

  implicit none

  integer, intent(in)  :: grid_id                   !< [IN]  grid identifier
  integer, intent(in)  :: nbr_points                !< [IN]  number of points
  integer, intent(in)  :: location                  !< [IN]  location, one of center/edge/vertex

  real, intent(in)     :: x_points_real(nbr_points) !< [IN]  longitudes of points
  real, intent(in)     :: y_points_real(nbr_points) !< [IN]  latitudes of points

  integer, intent(out) :: point_id                  !< [OUT] point identifier

  double precision    ::  x_points(nbr_points)
  double precision    ::  y_points(nbr_points)

  x_points(:) = dble(x_points_real(:))
  y_points(:) = dble(y_points_real(:))

  call yac_fdef_points_unstruct_r8 ( grid_id,    &
                                     nbr_points, &
                                     location,   &
                                     x_points,   &
                                     y_points,   &
                                     point_id )

end subroutine yac_fdef_points_unstruct_r4

subroutine yac_fdef_points_unstruct_r8 ( grid_id,    &
                                         nbr_points, &
                                         location,   &
                                         x_points,   &
                                         y_points,   &
                                         point_id )

  use mo_yac_finterface, dummy => yac_fdef_points_unstruct_r8

  implicit none

  interface

     subroutine yac_cdef_points_unstruct ( grid_id,      &
                                           nbr_points,   &
                                           location,     &
                                           x_points,     &
                                           y_points,     &
                                           point_id )    &
    &      bind ( c, name='yac_cdef_points_unstruct' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer (kind=c_int), value :: grid_id
       integer (kind=c_int), value :: nbr_points
       integer (kind=c_int), value :: location

       real    (kind=c_double)     :: x_points(nbr_points)
       real    (kind=c_double)     :: y_points(nbr_points)

       integer (kind=c_int)        :: point_id

     end subroutine yac_cdef_points_unstruct

  end interface

  integer, intent(in)  :: grid_id                      !< [IN]  grid identifier
  integer, intent(in)  :: nbr_points                   !< [IN]  number of points
  integer, intent(in)  :: location                     !< [IN]  location, one of center/edge/vertex

  double precision, intent(in) :: x_points(nbr_points) !< [IN]  longitudes of points
  double precision, intent(in) :: y_points(nbr_points) !< [IN]  latitudes of points

  integer, intent(out) :: point_id                     !< [OUT] grid identifier

  call yac_cdef_points_unstruct  ( grid_id,    &
                                   nbr_points, &
                                   location,   &
                                   x_points,   &
                                   y_points,   &
                                   point_id )

end subroutine yac_fdef_points_unstruct_r8

! ------------------------- def_grid -------------------------------

!>
!! @param[in]  grid_name             grid name
!! @param[in]  nbr_vertices          number of vertices
!! @param[in]  nbr_cells             number of cells
!! @param[in]  nbr_connections       total size of cell_to_vertex
!! @param[in]  nbr_vertices_per_cell number of vertices for each cell
!! @param[in]  x_vertices_real       longitudes of vertices
!! @param[in]  y_vertices_real       latitudes of vertices
!! @param[in]  cell_to_vertex_in     connectivity between vertices and cells\n
!!                                   (the vertex indices per cell have to be
!!                                   in clockwise or counterclockwise ordering)
!! @param[out] grid_id               grid identifier
subroutine yac_fdef_grid_nonuniform_r4 ( grid_name,             &
                                         nbr_vertices,          &
                                         nbr_cells,             &
                                         nbr_connections,       &
                                         nbr_vertices_per_cell, &
                                         x_vertices_real,       &
                                         y_vertices_real,       &
                                         cell_to_vertex_in,     &
                                         grid_id )

  use mo_yac_finterface, dummy => yac_fdef_grid_nonuniform_r4

  implicit none

  character(len=*), intent(in) :: grid_name
  integer, intent(in)  :: nbr_vertices
  integer, intent(in)  :: nbr_cells
  integer, intent(in)  :: nbr_connections
  integer, intent(in)  :: nbr_vertices_per_cell(nbr_cells)

  real, intent(in)     :: x_vertices_real(nbr_vertices)
  real, intent(in)     :: y_vertices_real(nbr_vertices)

  integer, intent(in)  :: cell_to_vertex_in(nbr_connections)

  integer, intent(out) :: grid_id

  double precision     :: x_vertices(nbr_vertices)
  double precision     :: y_vertices(nbr_vertices)

  x_vertices(:) = dble(x_vertices_real(:))
  y_vertices(:) = dble(y_vertices_real(:))

  call yac_fdef_grid_nonuniform_r8 ( grid_name,             &
                                     nbr_vertices,          &
                                     nbr_cells,             &
                                     nbr_connections,       &
                                     nbr_vertices_per_cell, &
                                     x_vertices,            &
                                     y_vertices,            &
                                     cell_to_vertex_in,     &
                                     grid_id )

end subroutine yac_fdef_grid_nonuniform_r4

!>
!! @param[in]  grid_name             grid name
!! @param[in]  nbr_vertices          number of vertices
!! @param[in]  nbr_cells             number of cells
!! @param[in]  nbr_connections       total size of cell_to_vertex
!! @param[in]  nbr_vertices_per_cell number of vertices for each cell
!! @param[in]  x_vertices            longitudes of vertices
!! @param[in]  y_vertices            latitudes of vertices
!! @param[in]  cell_to_vertex_in     connectivity between vertices and cells\n
!!                                   (the vertex indices per cell have to be in
!!                                   clockwise or counterclockwise ordering)
!! @param[out] grid_id               grid identifier
subroutine yac_fdef_grid_nonuniform_r8 ( grid_name,             &
                                         nbr_vertices,          &
                                         nbr_cells,             &
                                         nbr_connections,       &
                                         nbr_vertices_per_cell, &
                                         x_vertices,            &
                                         y_vertices,            &
                                         cell_to_vertex_in,     &
                                         grid_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_grid_nonuniform_r8

  implicit none

  interface

     subroutine yac_cdef_grid_unstruct ( grid_name,             &
                                         nbr_vertices,          &
                                         nbr_cells,             &
                                         nbr_vertices_per_cell, &
                                         x_vertices,            &
                                         y_vertices,            &
                                         cell_to_vertex,        &
                                         grid_id )              &
     &     bind ( c, name='yac_cdef_grid_unstruct' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char, c_double

       character ( kind=c_char ), dimension(*) :: grid_name
       integer ( kind=c_int    ), value        :: nbr_vertices
       integer ( kind=c_int    ), value        :: nbr_cells
       integer ( kind=c_int    )               :: nbr_vertices_per_cell(nbr_cells)
       real    ( kind=c_double )               :: x_vertices(nbr_vertices)
       real    ( kind=c_double )               :: y_vertices(nbr_vertices)
       integer ( kind=c_int    )               :: cell_to_vertex(nbr_cells,nbr_vertices)
       integer ( kind=c_int    )               :: grid_id

     end subroutine yac_cdef_grid_unstruct

  end interface

  character(len=*), intent(in) :: grid_name
  integer, intent(in)          :: nbr_vertices
  integer, intent(in)          :: nbr_cells
  integer, intent(in)          :: nbr_connections
  integer, intent(in)          :: nbr_vertices_per_cell(nbr_cells)

  double precision, intent(in) :: x_vertices(nbr_vertices)
  double precision, intent(in) :: y_vertices(nbr_vertices)

  integer, intent(in)          :: cell_to_vertex_in(nbr_connections)

  integer, intent(out)         :: grid_id


  integer                      :: cell_to_vertex(nbr_connections)

  IF (ANY(cell_to_vertex_in < 1)) &
    CALL yac_abort_message("ERROR(yac_fdef_grid_nonuniform_r8): " // &
      &                    "all entries of cell_to_vertex have to be > 0", &
      &                    __FILE__, __LINE__ )

  cell_to_vertex(:) = cell_to_vertex_in(:) - 1

  call yac_cdef_grid_unstruct  ( TRIM(grid_name) // c_null_char, &
                                 nbr_vertices,                   &
                                 nbr_cells,                      &
                                 nbr_vertices_per_cell,          &
                                 x_vertices,                     &
                                 y_vertices,                     &
                                 cell_to_vertex,                 &
                                 grid_id )

end subroutine yac_fdef_grid_nonuniform_r8

!>
!! @param[in]  grid_name                grid name
!! @param[in]  nbr_vertices             number of vertices
!! @param[in]  nbr_cells                number of cells
!! @param[in]  nbr_vertices_per_cell_in number of vertices for each cell
!! @param[in]  x_vertices_real          longitudes of vertices
!! @param[in]  y_vertices_real          latitudes of vertices
!! @param[in]  cell_to_vertex_in        connectivity between vertices and cells\n
!!                                      (the vertex indices per cell have to be
!!                                      in clockwise or counterclockwise ordering)
!! @param[out] grid_id                  grid identifier
subroutine yac_fdef_grid_unstruct_r4 ( grid_name,                &
                                       nbr_vertices,             &
                                       nbr_cells,                &
                                       nbr_vertices_per_cell_in, &
                                       x_vertices_real,          &
                                       y_vertices_real,          &
                                       cell_to_vertex_in,        &
                                       grid_id )

  use mo_yac_finterface, dummy => yac_fdef_grid_unstruct_r4

  implicit none

  character(len=*), intent(in) :: grid_name
  integer, intent(in)  :: nbr_vertices
  integer, intent(in)  :: nbr_cells
  integer, intent(in)  :: nbr_vertices_per_cell_in

  real, intent(in)     :: x_vertices_real(nbr_vertices)
  real, intent(in)     :: y_vertices_real(nbr_vertices)

  integer, intent(in)  :: cell_to_vertex_in(nbr_vertices_per_cell_in,nbr_cells)

  integer, intent(out)   :: grid_id

  double precision       :: x_vertices(nbr_vertices)
  double precision       :: y_vertices(nbr_vertices)

  x_vertices(:) = dble(x_vertices_real(:))
  y_vertices(:) = dble(y_vertices_real(:))

  call yac_fdef_grid_unstruct_r8 ( grid_name,                &
                                   nbr_vertices,             &
                                   nbr_cells,                &
                                   nbr_vertices_per_cell_in, &
                                   x_vertices,               &
                                   y_vertices,               &
                                   cell_to_vertex_in,        &
                                   grid_id )

end subroutine yac_fdef_grid_unstruct_r4

subroutine yac_fdef_grid_unstruct_r8 ( grid_name,                &
                                       nbr_vertices,             &
                                       nbr_cells,                &
                                       nbr_vertices_per_cell_in, &
                                       x_vertices,               &
                                       y_vertices,               &
                                       cell_to_vertex_in,        &
                                       grid_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_grid_unstruct_r8

  implicit none

  interface

     subroutine yac_cdef_grid_unstruct ( grid_name,             &
                                         nbr_vertices,          &
                                         nbr_cells,             &
                                         nbr_vertices_per_cell, &
                                         x_vertices,            &
                                         y_vertices,            &
                                         cell_to_vertex,        &
                                         grid_id )              &
     &     bind ( c, name='yac_cdef_grid_unstruct' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char, c_double

       character ( kind=c_char ), dimension(*) :: grid_name
       integer ( kind=c_int    ), value        :: nbr_vertices
       integer ( kind=c_int    ), value        :: nbr_cells
       integer ( kind=c_int    )               :: nbr_vertices_per_cell(nbr_cells)
       real    ( kind=c_double )               :: x_vertices(nbr_vertices)
       real    ( kind=c_double )               :: y_vertices(nbr_vertices)
       integer ( kind=c_int    )               :: cell_to_vertex(nbr_cells,nbr_vertices)
       integer ( kind=c_int    )               :: grid_id

     end subroutine yac_cdef_grid_unstruct

  end interface

  character(len=*), intent(in) :: grid_name                !< [IN]  grid name
  integer, intent(in)          :: nbr_vertices             !< [IN]  number of vertices
  integer, intent(in)          :: nbr_cells                !< [IN]  number of cells
  integer, intent(in)          :: nbr_vertices_per_cell_in !< [IN]  number of vertices for each cell

  double precision, intent(in) :: x_vertices(nbr_vertices) !< [IN]  longitudes of vertices
  double precision, intent(in) :: y_vertices(nbr_vertices) !< [IN]  latitudes of vertices

  integer, intent(in)  :: cell_to_vertex_in( &
                            nbr_vertices_per_cell_in, &
                            nbr_cells)                     !< [IN] connectivity between vertices and cells\n
                                                           !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)

  integer, intent(out)         :: grid_id                  !< [OUT] grid identifier

  integer                      :: nbr_vertices_per_cell(nbr_cells)
  integer                      :: cell_to_vertex(nbr_vertices_per_cell_in,nbr_cells)

  nbr_vertices_per_cell(:) = nbr_vertices_per_cell_in

  IF (ANY(cell_to_vertex_in < 1)) &
    CALL yac_abort_message("ERROR(yac_fdef_grid_unstruct_r8): " // &
      &                    "all entries of cell_to_vertex have to be > 0", &
      &                    __FILE__, __LINE__ )

  cell_to_vertex(:,:) = cell_to_vertex_in(:,:) - 1

  call yac_cdef_grid_unstruct  ( TRIM(grid_name) // c_null_char, &
                                 nbr_vertices,                   &
                                 nbr_cells,                      &
                                 nbr_vertices_per_cell,          &
                                 x_vertices,                     &
                                 y_vertices,                     &
                                 cell_to_vertex,                 &
                                 grid_id )

end subroutine yac_fdef_grid_unstruct_r8

subroutine yac_fdef_grid_curve2d_r4 ( grid_name,       &
                                      nbr_vertices,    &
                                      cyclic,          &
                                      x_vertices_real, &
                                      y_vertices_real, &
                                      grid_id )

  use mo_yac_finterface, dummy => yac_fdef_grid_curve2d_r4

  implicit none

  character(len=*), intent(in) :: grid_name          !< [IN]  grid name
  integer, intent(in)  :: nbr_vertices(2)            !< [IN]  number of cells in each dimension
  integer, intent(in)  :: cyclic(2)                  !< [IN]  cyclic behavior of cells in each dimension

  real, intent(in)     :: &
    x_vertices_real(nbr_vertices(1),nbr_vertices(2)) !< [IN]  longitudes of vertices
  real, intent(in)     :: &
    y_vertices_real(nbr_vertices(1),nbr_vertices(2)) !< [IN]  latitudes of vertices

  integer, intent(out) :: grid_id                    !< [OUT] grid identifier

  double precision     ::  x_vertices(nbr_vertices(1),nbr_vertices(2))
  double precision     ::  y_vertices(nbr_vertices(1),nbr_vertices(2))

  x_vertices(:,:) = dble(x_vertices_real(:,:))
  y_vertices(:,:) = dble(y_vertices_real(:,:))

  call yac_fdef_grid_curve2d_r8 ( grid_name,    &
                                  nbr_vertices, &
                                  cyclic,       &
                                  x_vertices,   &
                                  y_vertices,   &
                                  grid_id )

end subroutine yac_fdef_grid_curve2d_r4

subroutine yac_fdef_grid_curve2d_r8 ( grid_name,    &
                                    nbr_vertices, &
                                    cyclic,       &
                                    x_vertices,   &
                                    y_vertices,   &
                                    grid_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_grid_curve2d_r8

  implicit none

  interface

     subroutine yac_cdef_grid_curve2d ( grid_name,    &
                                        nbr_vertices, &
                                        cyclic,       &
                                        x_vertices,   &
                                        y_vertices,   &
                                        grid_id )     &
    &      bind ( c, name='yac_cdef_grid_curve2d' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char, c_double

       character ( kind=c_char ), dimension(*) :: grid_name
       integer ( kind=c_int    )               :: nbr_vertices(2)
       integer ( kind=c_int    )               :: cyclic(2)
       real    ( kind=c_double )               :: x_vertices(nbr_vertices(1),nbr_vertices(2))
       real    ( kind=c_double )               :: y_vertices(nbr_vertices(1),nbr_vertices(2))
       integer ( kind=c_int    )               :: grid_id

     end subroutine yac_cdef_grid_curve2d

  end interface

  character(len=*), intent(in) :: grid_name                   !< [IN]  grid name
  integer, intent(in)          :: nbr_vertices(2)             !< [IN]  number of cells in each dimension
  integer, intent(in)          :: cyclic(2)                   !< [IN]  cyclic behavior of cells in each dimension

  double precision, intent(in) :: &
    x_vertices(nbr_vertices(1),nbr_vertices(2))               !< [IN]  longitudes of vertices
  double precision, intent(in) :: &
    y_vertices(nbr_vertices(1),nbr_vertices(2))               !< [IN]  latitudes of vertices

  integer, intent(out)         :: grid_id                     !< [OUT] grid identifier

  call yac_cdef_grid_curve2d ( TRIM(grid_name) // c_null_char, &
                               nbr_vertices,                   &
                               cyclic,                         &
                               x_vertices,                     &
                               y_vertices,                     &
                               grid_id )

end subroutine yac_fdef_grid_curve2d_r8

subroutine yac_fdef_grid_reg2d_r4 ( grid_name,       &
                                    nbr_vertices,    &
                                    cyclic,          &
                                    x_vertices_real, &
                                    y_vertices_real, &
                                    grid_id )

  use mo_yac_finterface, dummy => yac_fdef_grid_reg2d_r4

  implicit none

  character(len=*), intent(in) :: grid_name                !< [IN]  grid name
  integer, intent(in)  :: nbr_vertices(2)                  !< [IN]  number of cells in each dimension
  integer, intent(in)  :: cyclic(2)                        !< [IN]  cyclic behavior of cells in each dimension

  real, intent(in)     :: x_vertices_real(nbr_vertices(1)) !< [IN]  longitudes of vertices
  real, intent(in)     :: y_vertices_real(nbr_vertices(2)) !< [IN]  latitudes of vertices

  integer, intent(out) :: grid_id                          !< [OUT] grid identifier

  double precision     ::  x_vertices(nbr_vertices(1))
  double precision     ::  y_vertices(nbr_vertices(2))

  x_vertices(:) = dble(x_vertices_real(:))
  y_vertices(:) = dble(y_vertices_real(:))

  call yac_fdef_grid_reg2d_r8 ( grid_name,    &
                                nbr_vertices, &
                                cyclic,       &
                                x_vertices,   &
                                y_vertices,   &
                                grid_id )

end subroutine yac_fdef_grid_reg2d_r4

subroutine yac_fdef_grid_reg2d_r8 ( grid_name,    &
                                    nbr_vertices, &
                                    cyclic,       &
                                    x_vertices,   &
                                    y_vertices,   &
                                    grid_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_grid_reg2d_r8

  implicit none

  interface

     subroutine yac_cdef_grid_reg2d ( grid_name,    &
                                      nbr_vertices, &
                                      cyclic,       &
                                      x_vertices,   &
                                      y_vertices,   &
                                      grid_id )     &
    &      bind ( c, name='yac_cdef_grid_reg2d' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char, c_double

       character ( kind=c_char ), dimension(*) :: grid_name
       integer ( kind=c_int    )               :: nbr_vertices(2)
       integer ( kind=c_int    )               :: cyclic(2)
       real    ( kind=c_double )               :: x_vertices(nbr_vertices(1))
       real    ( kind=c_double )               :: y_vertices(nbr_vertices(2))
       integer ( kind=c_int    )               :: grid_id

     end subroutine yac_cdef_grid_reg2d

  end interface

  character(len=*), intent(in) :: grid_name                   !< [IN]  grid name
  integer, intent(in)          :: nbr_vertices(2)             !< [IN]  number of cells in each dimension
  integer, intent(in)          :: cyclic(2)                   !< [IN]  cyclic behavior of cells in each dimension

  double precision, intent(in) :: x_vertices(nbr_vertices(1)) !< [IN]  longitudes of vertices
  double precision, intent(in) :: y_vertices(nbr_vertices(2)) !< [IN]  latitudes of vertices

  integer, intent(out)         :: grid_id                     !< [OUT] grid identifier

  call yac_cdef_grid_reg2d ( TRIM(grid_name) // c_null_char, &
                             nbr_vertices,                   &
                             cyclic,                         &
                             x_vertices,                     &
                             y_vertices,                     &
                             grid_id )

end subroutine yac_fdef_grid_reg2d_r8

! ------------------------- set global_index ------------------------------

subroutine yac_fset_global_index ( global_index, &
                                   location,     &
                                   grid_id )

  use mo_yac_finterface, dummy => yac_fset_global_index

  implicit none

  interface

     subroutine yac_cset_global_index ( global_index, &
                                        location,     &
                                        grid_id )     &
    &      bind ( c, name='yac_cset_global_index' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int )        :: global_index(*)
       integer ( kind=c_int ), value :: location
       integer ( kind=c_int ), value :: grid_id

     end subroutine yac_cset_global_index

  end interface

  integer, intent(in)  :: global_index(*) !< [IN] global index
  integer, intent(in)  :: location        !< [IN] location
  integer, intent(in)  :: grid_id         !< [IN] point set indentifier

  call yac_cset_global_index ( global_index, &
                               location,     &
                               grid_id )

end subroutine yac_fset_global_index

! ------------------------- set core_mask ------------------------------

subroutine yac_fset_core_lmask ( is_core,    &
                                 location, &
                                 grid_id )

  use mo_yac_finterface, dummy => yac_fset_core_lmask
  use, intrinsic :: iso_c_binding, only : c_size_t

  implicit none

  logical, intent(in)  :: is_core(*) !< [IN] core flag
                                     !< false, cell/vertex/edge is halo
                                     !< true, cell/vertex/edge is core
  integer, intent(in)  :: location   !< [IN] location
  integer, intent(in)  :: grid_id    !< [IN] point set indentifier

  integer (kind=c_size_t) :: i, count
  integer, allocatable :: int_is_core(:)

  interface

     function yac_get_grid_size ( location, &
                                  grid_id ) &
           result ( grid_size )               &
           bind ( c, name='yac_get_grid_size' )

       use, intrinsic :: iso_c_binding, only : c_int, c_size_t

       integer ( kind=c_int ), value :: location
       integer ( kind=c_int ), value :: grid_id
       integer ( kind=c_size_t) :: grid_size

     end function yac_get_grid_size

  end interface

  count = yac_get_grid_size(location, grid_id)
  allocate(int_is_core(count))

  do i = 1, count
     if ( is_core(i) ) then
        int_is_core(i) = 1
     else
        int_is_core(i) = 0
     endif
  enddo

  call yac_fset_core_imask ( int_is_core, &
                             location,    &
                             grid_id )

end subroutine yac_fset_core_lmask

subroutine yac_fset_core_imask ( is_core,    &
                                 location, &
                                 grid_id )

  use mo_yac_finterface, dummy => yac_fset_core_imask

  implicit none

  interface

     subroutine yac_cset_core_mask ( mask,        &
                                     location,    &
                                     grid_id )    &
    &      bind ( c, name='yac_cset_core_mask' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int )        :: mask(*)
       integer ( kind=c_int ), value :: location
       integer ( kind=c_int ), value :: grid_id

     end subroutine yac_cset_core_mask

  end interface


  integer, intent(in)  :: is_core(*) !< [IN] core flag
                                     !< 0, cell/vertex/edge is halo
                                     !< 1, cell/vertex/edge is core
  integer, intent(in)  :: location   !< [IN] location
  integer, intent(in)  :: grid_id    !< [IN] point set indentifier

  call yac_cset_core_mask ( is_core,  &
                            location, &
                            grid_id )

end subroutine yac_fset_core_imask

! ------------------------- set_mask ------------------------------

subroutine yac_fset_lmask ( is_valid,     &
                            points_id )

  use mo_yac_finterface, dummy => yac_fset_lmask
  use, intrinsic :: iso_c_binding, only : c_size_t

  implicit none

  logical, intent(in)  :: is_valid(*)  !< [IN] logical mask
                                       !< false, point is masked out
                                       !< true, point is valid
  integer, intent(in)  :: points_id    !< [IN] point set indentifier

  integer ( kind=c_size_t) :: i, count
  integer, allocatable :: int_is_valid(:)

  interface

     function yac_get_points_size_f ( points_id ) &
           result ( points_size )                 &
           bind ( c, name='yac_get_points_size' )

       use, intrinsic :: iso_c_binding, only : c_int, c_size_t

       integer ( kind=c_int ), value :: points_id
       integer ( kind=c_size_t) :: points_size

     end function yac_get_points_size_f

  end interface

  count = yac_get_points_size_f(points_id)
  allocate(int_is_valid(count))

  do i = 1, count
    int_is_valid(i) = MERGE(1,0,is_valid(i))
  enddo

  call yac_fset_imask ( int_is_valid, points_id )

end subroutine yac_fset_lmask

subroutine yac_fset_imask ( is_valid, &
                            points_id )

  use mo_yac_finterface, dummy => yac_fset_imask

  implicit none

  interface

     subroutine yac_cset_mask ( is_valid,     &
                                points_id )   &
    &      bind ( c, name='yac_cset_mask' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int )        :: is_valid(*)
       integer ( kind=c_int ), value :: points_id

     end subroutine yac_cset_mask

  end interface

  integer, intent(in)  :: is_valid(*)  !< [IN] integer mask
                                       !< 0, point is masked out
                                       !< 1, point is valid
  integer, intent(in)  :: points_id    !< [IN] point set indentifier

  call yac_cset_mask ( is_valid, points_id )

end subroutine yac_fset_imask

! ------------------------- def_mask ------------------------------

subroutine yac_fdef_lmask ( grid_id,    &
                            nbr_points, &
                            location,   &
                            is_valid,   &
                            mask_id )

  use mo_yac_finterface, dummy => yac_fdef_lmask

  implicit none

  integer, intent(in)  :: grid_id     !< [IN] grid identifier
  integer, intent(in)  :: nbr_points  !< [IN] number of points
  integer, intent(in)  :: location    !< [IN] location, one of center/edge/vertex
  logical, intent(in)  :: is_valid(*) !< [IN] logical mask
                                      !< false, point is masked out
                                      !< true, point is valid
  integer, intent(out) :: mask_id     !< [OUT] mask identifier

  integer :: i
  integer, allocatable :: int_is_valid(:)

  allocate(int_is_valid(nbr_points))

  do i = 1, nbr_points
     int_is_valid(i) = MERGE(1,0,is_valid(i))
  enddo

  call yac_fdef_imask ( grid_id,      &
                        nbr_points,   &
                        location,     &
                        int_is_valid, &
                        mask_id )

end subroutine yac_fdef_lmask

subroutine yac_fdef_imask ( grid_id,    &
                            nbr_points, &
                            location,   &
                            is_valid,   &
                            mask_id )

  use mo_yac_finterface, dummy => yac_fdef_imask

  implicit none

  interface

     subroutine yac_cdef_mask ( grid_id,    &
                                nbr_points, &
                                location,   &
                                is_valid,   &
                                mask_id )   &
    &      bind ( c, name='yac_cdef_mask' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: grid_id
       integer ( kind=c_int ), value :: nbr_points
       integer ( kind=c_int ), value :: location
       integer ( kind=c_int )        :: is_valid(*)
       integer ( kind=c_int )        :: mask_id

     end subroutine yac_cdef_mask

  end interface

  integer, intent(in)  :: grid_id     !< [IN] grid identifier
  integer, intent(in)  :: nbr_points  !< [IN] number of points
  integer, intent(in)  :: location    !< [IN] location, one of center/edge/vertex
  integer, intent(in)  :: is_valid(*) !< [IN] logical mask
                                      !< false, point is masked out
                                      !< true, point is valid
  integer, intent(out) :: mask_id     !< [OUT] mask identifier

  call yac_cdef_mask ( grid_id,    &
                       nbr_points, &
                       location,   &
                       is_valid,   &
                       mask_id )

end subroutine yac_fdef_imask

! ----------------------------- def_field -------------------------------

subroutine yac_fdef_field ( field_name,    &
                            component_id,  &
                            point_ids,     &
                            num_pointsets, &
                            field_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_field

  implicit none

  interface

     subroutine yac_cdef_field ( field_name,    &
                                 component_id,  &
                                 point_ids,     &
                                 num_pointsets, &
                                 field_id )     &
    &      bind ( c, name='yac_cdef_field' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char ), dimension(*) :: field_name !< [IN] short name of the field
       integer ( kind=c_int ), value :: component_id         !< [IN] component identifier
       integer ( kind=c_int )        :: point_ids(*)         !< [IN] point identifier
       integer ( kind=c_int ), value :: num_pointsets        !< [IN] number of pointsets per grid
       integer ( kind=c_int )        :: field_id             !< [OUT] returned field handle

     end subroutine yac_cdef_field

  end interface

  !
  ! Definition of coupling fields
  !
  character(len=*), intent (in) :: field_name    !< [IN] short name of the field
  integer, intent (in)          :: component_id  !< [IN] component identifier
  integer, intent (in)          :: point_ids(*)  !< [IN] point identifier
  integer, intent (in)          :: num_pointsets !< [IN] number of pointsets per grid
  integer, intent (out)         :: field_id      !< [OUT] returned field handle

  call yac_check_strlength ( field_name )

  call yac_cdef_field ( TRIM(field_name) // c_null_char, &
                        component_id,                    &
                        point_ids,                       &
                        num_pointsets,                   &
                        field_id )

end subroutine yac_fdef_field

! ----------------------------- def_field_mask---------------------------

subroutine yac_fdef_field_mask ( field_name,    &
                                 component_id,  &
                                 point_ids,     &
                                 mask_ids,      &
                                 num_pointsets, &
                                 field_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_field_mask

  implicit none

  interface

     subroutine yac_cdef_field_mask ( field_name,    &
                                      component_id,  &
                                      point_ids,     &
                                      mask_ids,      &
                                      num_pointsets, &
                                      field_id )     &
    &      bind ( c, name='yac_cdef_field_mask' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char ), dimension(*) :: field_name !< [IN] short name of the field
       integer ( kind=c_int ), value :: component_id         !< [IN] component identifier
       integer ( kind=c_int )        :: point_ids(*)         !< [IN] point identifier
       integer ( kind=c_int )        :: mask_ids(*)          !< [IN] mask identifier
       integer ( kind=c_int ), value :: num_pointsets        !< [IN] number of pointsets per grid
       integer ( kind=c_int )        :: field_id             !< [OUT] returned field handle

     end subroutine yac_cdef_field_mask

  end interface


  !
  ! Definition of coupling fields
  !
  character(len=*), intent (in) :: field_name    !< [IN] short name of the field
  integer, intent (in)          :: component_id  !< [IN] component identifier
  integer, intent (in)          :: point_ids(*)  !< [IN] point identifier
  integer, intent (in)          :: mask_ids(*)   !< [IN] mask identifier
  integer, intent (in)          :: num_pointsets !< [IN] number of pointsets per grid
  integer, intent (out)         :: field_id      !< [OUT] returned field handle

  call yac_check_strlength ( field_name )

  call yac_cdef_field_mask ( TRIM(field_name) // c_null_char, &
                             component_id,                    &
                             point_ids,                       &
                             mask_ids,                        &
                             num_pointsets,                   &
                             field_id )

end subroutine yac_fdef_field_mask

! ---------------------------------- put --------------------------------

!>
!! @param[in]  field_id        field identifier
!! @param[in]  nbr_hor_points  number of horizontal points
!! @param[in]  nbr_pointsets   number of point sets
!! @param[in]  collection_size number of vertical level or bundles
!! @param[in]  send_field      returned field
!! @param[out] info            returned info
!! @param[out] ierror          returned error
subroutine yac_fput_r4 ( field_id,         &
     &                   nbr_hor_points,   &
     &                   nbr_pointsets,    &
     &                   collection_size,  &
     &                   send_field,       &
     &                   info,             &
     &                   ierror )

  use mo_yac_finterface, dummy => yac_fput_r4

  implicit none

  integer, intent (in)  :: field_id
  integer, intent (in)  :: nbr_hor_points
  integer, intent (in)  :: nbr_pointsets
  integer, intent (in)  :: collection_size
  real, intent (in)     :: send_field(nbr_hor_points, &
                                      nbr_pointsets, &
                                      collection_size)
  integer, intent (out) :: info
  integer, intent (out) :: ierror

  double precision :: send_buffer(nbr_hor_points, nbr_pointsets, collection_size)

  send_buffer = DBLE(send_field)

  call yac_fput_r8 ( field_id,         &
                     nbr_hor_points,   &
                     nbr_pointsets,    &
                     collection_size,  &
                     send_buffer,      &
                     info,             &
                     ierror )

end subroutine yac_fput_r4

!>
!! @param[in]  field_id        field identifier
!! @param[in]  nbr_pointsets   number of point sets
!! @param[in]  collection_size number of vertical level or bundles
!! @param[in]  send_field      returned field
!! @param[out] info            returned info
!! @param[out] ierror          returned error
subroutine yac_fput_r4_ptr ( field_id,         &
     &                       nbr_pointsets,    &
     &                       collection_size,  &
     &                       send_field,       &
     &                       info,             &
     &                       ierror )

  use mo_yac_finterface, dummy => yac_fput_r4_ptr

  implicit none

  integer, intent (in)     :: field_id
  integer, intent (in)     :: nbr_pointsets
  integer, intent (in)     :: collection_size
  type(yac_r4_ptr), intent (in) :: send_field(nbr_pointsets, &
                                              collection_size)
  integer, intent (out)    :: info
  integer, intent (out)    :: ierror

  integer :: i, j
  type(yac_r8_ptr) :: send_field_r8(nbr_pointsets, collection_size)

  do i = 1, collection_size
    do j = 1, nbr_pointsets
      allocate(send_field_r8(j, i)%p(SIZE(send_field(j, i)%p)))
      send_field_r8(j, i)%p = dble(send_field(j, i)%p)
    end do
  end do

  call yac_fput_r8_ptr ( field_id,         &
     &                   nbr_pointsets,    &
     &                   collection_size,  &
     &                   send_field_r8,    &
     &                   info,             &
     &                   ierror )

  do i = 1, collection_size
    do j = 1, nbr_pointsets
      deallocate(send_field_r8(j, i)%p)
    end do
  end do

end subroutine yac_fput_r4_ptr

!>
!! @param[in]  field_id        field identifier
!! @param[in]  nbr_hor_points  number of horizontal points
!! @param[in]  collection_size number of vertical level or bundles
!! @param[in]  send_field      returned field
!! @param[out] info            returned info
!! @param[out] ierror          returned error
subroutine yac_fput_single_pointset_r4 ( field_id,         &
     &                                   nbr_hor_points,   &
     &                                   collection_size,  &
     &                                   send_field,       &
     &                                   info,             &
     &                                   ierror )

  use mo_yac_finterface, dummy => yac_fput_single_pointset_r4

  implicit none

  integer, intent (in)          :: field_id
  integer, intent (in)          :: nbr_hor_points
  integer, intent (in)          :: collection_size
  real, intent (in)             :: send_field(nbr_hor_points, &
                                              collection_size)
  integer, intent (out)         :: info
  integer, intent (out)         :: ierror

  double precision :: send_buffer(nbr_hor_points, collection_size)

  send_buffer = DBLE(send_field)

  call yac_fput_single_pointset_r8 ( field_id,         &
                                     nbr_hor_points,   &
                                     collection_size,  &
                                     send_buffer,      &
                                     info,             &
                                     ierror )

end subroutine yac_fput_single_pointset_r4

!>
!! @param[in]  field_id        field identifier
!! @param[in]  nbr_hor_points  number of horizontal points
!! @param[in]  collection_size number of vertical level or bundles
!! @param[in]  nbr_pointsets   number of point sets
!! @param[in]  send_field      returned field
!! @param[out] info            returned info
!! @param[out] ierror          returned error
subroutine yac_fput_r8 ( field_id,         &
     &                   nbr_hor_points,   &
     &                   collection_size,  &
     &                   nbr_pointsets,    &
     &                   send_field,       &
     &                   info,             &
     &                   ierror )

  use mo_yac_finterface, dummy => yac_fput_r8

  implicit none

  interface

     subroutine yac_cput_ ( field_id,         &
                            collection_size,  &
                            send_field,       &
                            info,             &
                            ierror ) bind ( c, name='yac_cput_' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int ), value :: field_id
       integer ( kind=c_int ), value :: collection_size
       real    ( kind=c_double )     :: send_field(*)
       integer ( kind=c_int )        :: info
       integer ( kind=c_int )        :: ierror

     end subroutine yac_cput_

  end interface

  integer, intent (in)          :: field_id
  integer, intent (in)          :: nbr_hor_points
  integer, intent (in)          :: nbr_pointsets
  integer, intent (in)          :: collection_size
  double precision, intent (in) :: send_field(nbr_hor_points, &
                                              nbr_pointsets, &
                                              collection_size)
  integer, intent (out)         :: info
  integer, intent (out)         :: ierror

  call yac_cput_ ( field_id,         &
                   collection_size,  &
                   send_field,       &
                   info,             &
                   ierror )

end subroutine yac_fput_r8

!>
!! @param[in]  field_id        field identifier
!! @param[in]  nbr_pointsets   number of point sets
!! @param[in]  collection_size number of vertical level or bundles
!! @param[in]  send_field      returned field
!! @param[out] info            returned info
!! @param[out] ierror          returned error
subroutine yac_fput_r8_ptr ( field_id,         &
     &                       nbr_pointsets,    &
     &                       collection_size,  &
     &                       send_field,       &
     &                       info,             &
     &                       ierror )

  use mo_yac_finterface, dummy => yac_fput_r8_ptr
  use iso_c_binding, only: c_ptr, c_loc

  implicit none

  interface

     subroutine yac_cput_ptr_ ( field_id,         &
                                collection_size,  &
                                send_field,       &
                                info,             &
                                ierror ) bind ( c, name='yac_cput_ptr_' )

       use, intrinsic :: iso_c_binding, only : c_int, c_ptr

       integer ( kind=c_int ), value :: field_id
       integer ( kind=c_int ), value :: collection_size
       type(c_ptr)                   :: send_field(*)
       integer ( kind=c_int )        :: info
       integer ( kind=c_int )        :: ierror

     end subroutine yac_cput_ptr_

  end interface

  integer, intent (in)     :: field_id
  integer, intent (in)     :: nbr_pointsets
  integer, intent (in)     :: collection_size
  type(yac_r8_ptr), intent (in) :: send_field(nbr_pointsets, collection_size)
  integer, intent (out)    :: info
  integer, intent (out)    :: ierror

  integer :: i, j
  type(c_ptr) :: send_field_(nbr_pointsets, collection_size)

  do i = 1, collection_size
    do j = 1, nbr_pointsets
      send_field_(j, i) = c_loc(send_field(j, i)%p(1))
    end do
  end do

  call yac_cput_ptr_ ( field_id,         &
     &                 collection_size,  &
     &                 send_field_,      &
     &                 info,             &
     &                 ierror )

end subroutine yac_fput_r8_ptr

!>
!! @param[in]  field_id        field identifier
!! @param[in]  nbr_hor_points  number of horizontal points
!! @param[in]  collection_size number of vertical level or bundles
!! @param[in]  send_field      returned field
!! @param[out] info            returned info
!! @param[out] ierror          returned error
subroutine yac_fput_single_pointset_r8 ( field_id,         &
     &                                   nbr_hor_points,   &
     &                                   collection_size,  &
     &                                   send_field,       &
     &                                   info,             &
     &                                   ierror )

  use mo_yac_finterface, dummy => yac_fput_single_pointset_r8

  implicit none

  interface

     subroutine yac_cput_ ( field_id,         &
                            collection_size,  &
                            send_field,       &
                            info,             &
                            ierror ) bind ( c, name='yac_cput_' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int ), value :: field_id
       integer ( kind=c_int ), value :: collection_size
       real    ( kind=c_double )     :: send_field(*)
       integer ( kind=c_int )        :: info
       integer ( kind=c_int )        :: ierror

     end subroutine yac_cput_

  end interface

  integer, intent (in)          :: field_id
  integer, intent (in)          :: nbr_hor_points
  integer, intent (in)          :: collection_size
  double precision, intent (in) :: send_field(nbr_hor_points, &
                                              collection_size)
  integer, intent (out)         :: info
  integer, intent (out)         :: ierror

  call yac_cput_ ( field_id,         &
     &             collection_size,  &
     &             send_field,       &
     &             info,             &
     &             ierror )

end subroutine yac_fput_single_pointset_r8

! ---------------------------------- get -------------------------------

subroutine yac_fget_r4 ( field_id,         &
     &                   nbr_hor_points,   &
     &                   collection_size,  &
     &                   recv_field,       &
     &                   info,             &
     &                   ierror )

  use mo_yac_finterface, dummy => yac_fget_r4

  implicit none

  integer, intent (in)          :: field_id        !< [IN] field identifier
  integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
  real, intent (inout)          :: recv_field(nbr_hor_points, collection_size)
                                                   !< [INOUT] returned field
  integer, intent (out)         :: info            !< [OUT] returned info
  integer, intent (out)         :: ierror          !< [OUT] returned error

  double precision :: recv_buffer(nbr_hor_points, collection_size)

  recv_buffer = DBLE(recv_field)

  call yac_fget_r8 ( field_id,        &
                     nbr_hor_points,  &
                     collection_size, &
                     recv_buffer,     &
                     info,            &
                     ierror )

  recv_field = REAL(recv_buffer)

end subroutine yac_fget_r4

subroutine yac_fget_r4_ptr ( field_id,         &
  &                          collection_size,  &
  &                          recv_field,       &
  &                          info,             &
  &                          ierror )

  use mo_yac_finterface, dummy => yac_fget_r4_ptr

  implicit none

  integer, intent (in)  :: field_id        !< [IN] field identifier
  integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
  type(yac_r4_ptr)      :: recv_field(collection_size) !< [INOUT] returned field
  integer, intent (out) :: info            !< [OUT] returned info
  integer, intent (out) :: ierror          !< [OUT] returned error handler

  integer :: i
  type(yac_r8_ptr) :: recv_field_r8(collection_size)

  do i = 1, collection_size
    allocate(recv_field_r8(i)%p(SIZE(recv_field(i)%p)))
    recv_field_r8(i)%p = DBLE(recv_field(i)%p)
  end do

  call yac_fget_r8_ptr ( field_id,         &
     &                   collection_size,  &
     &                   recv_field_r8,    &
     &                   info,             &
     &                   ierror )

  do i = 1, collection_size
    recv_field(i)%p = REAL(recv_field_r8(i)%p)
    deallocate(recv_field_r8(i)%p)
  end do

end subroutine yac_fget_r4_ptr

subroutine yac_fget_r8 ( field_id,         &
     &                   nbr_hor_points,   &
     &                   collection_size,  &
     &                   recv_field,       &
     &                   info,             &
     &                   ierror )

  use mo_yac_finterface, dummy => yac_fget_r8

  implicit none

  interface

     subroutine yac_cget_ ( field_id,         &
                            collection_size,  &
                            recv_field,       &
                            info,             &
                            ierror ) bind ( c, name='yac_cget_' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int ), value :: field_id            !< [IN] point identifier
       integer ( kind=c_int ), value :: collection_size     !< [IN] collection size
       real    ( kind=c_double )     :: recv_field(*)       !< [OUT] returned field
       integer ( kind=c_int )        :: info                !< [OUT] returned info
       integer ( kind=c_int )        :: ierror              !< [OUT] returned error handler

     end subroutine yac_cget_

  end interface

  integer, intent (in)          :: field_id        !< [IN] field identifier
  integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
  double precision, intent (inout):: recv_field(nbr_hor_points, collection_size)
                                                  !< [INOUT] returned field
  integer, intent (out)         :: info           !< [OUT] returned info
  integer, intent (out)         :: ierror         !< [OUT] returned error

  call yac_cget_ ( field_id,         &
                   collection_size,  &
                   recv_field,       &
                   info,             &
                   ierror )

end subroutine yac_fget_r8

subroutine yac_fget_r8_ptr ( field_id,         &
                             collection_size,  &
                             recv_field,       &
                             info,             &
                             ierror )

  use mo_yac_finterface, dummy => yac_fget_r8_ptr
  use iso_c_binding, only: c_ptr, c_loc

  implicit none

  interface

     subroutine yac_cget ( field_id,         &
                           collection_size,  &
                           recv_field,       &
                           info,             &
                           ierror ) bind ( c, name='yac_cget' )

       use, intrinsic :: iso_c_binding, only : c_int, c_ptr

       integer ( kind=c_int ), value :: field_id            !< [IN] point identifier
       integer ( kind=c_int ), value :: collection_size     !< [IN] collection size
       type(c_ptr)                   :: recv_field(*)       !< [OUT] returned field
       integer ( kind=c_int )        :: info                !< [OUT] returned info
       integer ( kind=c_int )        :: ierror              !< [OUT] returned error handler

     end subroutine yac_cget

  end interface

  integer, intent (in)  :: field_id        !< [IN] field identifier
  integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
  type(yac_r8_ptr)      :: recv_field(collection_size) !< [OUT] returned field
  integer, intent (out) :: info            !< [OUT] returned info
  integer, intent (out) :: ierror          !< [OUT] returned error handler

  integer :: i
  type(c_ptr) :: recv_field_(collection_size)

  do i = 1, collection_size
    recv_field_(i) = c_loc(recv_field(i)%p(1))
  end do

  call yac_cget ( field_id,         &
                  collection_size,  &
                  recv_field_,      &
                  info,             &
                  ierror )

end subroutine yac_fget_r8_ptr

! ---------------------------------- exchange --------------------------------

subroutine yac_fexchange_r4 ( send_field_id,       &
     &                        recv_field_id,       &
     &                        send_nbr_hor_points, &
     &                        send_nbr_pointsets,  &
     &                        recv_nbr_hor_points, &
     &                        collection_size,     &
     &                        send_field,          &
     &                        recv_field,          &
     &                        send_info,           &
     &                        recv_info,           &
     &                        ierror )

  use mo_yac_finterface, dummy => yac_fexchange_r4

  implicit none

  integer, intent (in)  :: send_field_id       !< [IN] field identifier
  integer, intent (in)  :: recv_field_id       !< [IN] field identifier
  integer, intent (in)  :: send_nbr_hor_points !< [IN] number of horizontal send points
  integer, intent (in)  :: send_nbr_pointsets  !< [IN] number of send point sets
  integer, intent (in)  :: recv_nbr_hor_points !< [IN] number of horizontal recv points
  integer, intent (in)  :: collection_size     !< [IN] number of vertical level or bundles
  real, intent (in)     :: send_field(send_nbr_hor_points, &
                                      send_nbr_pointsets,  &
                                      collection_size)
                                               !< [IN] returned send field
  real, intent (inout)  :: recv_field(recv_nbr_hor_points, &
                                      collection_size)
                                               !< [INOUT] returned recv field
  integer, intent (out) :: send_info           !< [OUT] returned send info
  integer, intent (out) :: recv_info           !< [OUT] returned recv info
  integer, intent (out) :: ierror              !< [OUT] returned error

  double precision :: send_buffer(send_nbr_hor_points, &
                                  send_nbr_pointsets,  &
                                  collection_size)
  double precision :: recv_buffer(recv_nbr_hor_points, &
                                  collection_size)

  send_buffer = DBLE(send_field)
  recv_buffer = DBLE(recv_field)

  call yac_fexchange_r8 ( send_field_id,       &
                          recv_field_id,       &
                          send_nbr_hor_points, &
                          send_nbr_pointsets,  &
                          recv_nbr_hor_points, &
                          collection_size,     &
                          send_buffer,         &
                          recv_buffer,         &
                          send_info,           &
                          recv_info,           &
                          ierror )

  recv_field = REAL(recv_buffer)

end subroutine yac_fexchange_r4

subroutine yac_fexchange_r4_ptr ( send_field_id,      &
     &                            recv_field_id,      &
     &                            send_nbr_pointsets, &
     &                            collection_size,    &
     &                            send_field,         &
     &                            recv_field,         &
     &                            send_info,          &
     &                            recv_info,          &
     &                            ierror )

  use mo_yac_finterface, dummy => yac_fexchange_r4_ptr

  implicit none

  integer, intent (in)  :: send_field_id       !< [IN] send field identifier
  integer, intent (in)  :: recv_field_id       !< [IN] recv field identifier
  integer, intent (in)  :: send_nbr_pointsets  !< [IN] number of send point sets
  integer, intent (in)  :: collection_size     !< [IN] number of vertical level or bundles
  type(yac_r4_ptr), intent (in) ::                        &
                           send_field(send_nbr_pointsets, &
                                      collection_size)
                                               !< [IN] returned send field
  type(yac_r4_ptr)      :: recv_field(collection_size)
                                               !< [INOUT] returned recv field
  integer, intent (out) :: send_info           !< [OUT] returned send info
  integer, intent (out) :: recv_info           !< [OUT] returned recv info
  integer, intent (out) :: ierror              !< [OUT] returned error

  integer :: i, j
  type(yac_r8_ptr) :: send_field_r8(send_nbr_pointsets, collection_size)
  type(yac_r8_ptr) :: recv_field_r8(collection_size)

  do i = 1, collection_size
    do j = 1, send_nbr_pointsets
      allocate(send_field_r8(j, i)%p(SIZE(send_field(j, i)%p)))
      send_field_r8(j, i)%p = dble(send_field(j, i)%p)
    end do
  end do
  do i = 1, collection_size
    allocate(recv_field_r8(i)%p(SIZE(recv_field(i)%p)))
    recv_field_r8(i)%p = dble(recv_field(i)%p)
  end do

  call yac_fexchange_r8_ptr ( send_field_id,      &
     &                        recv_field_id,      &
     &                        send_nbr_pointsets, &
     &                        collection_size,    &
     &                        send_field_r8,      &
     &                        recv_field_r8,      &
     &                        send_info,          &
     &                        recv_info,          &
     &                        ierror )

  do i = 1, collection_size
    recv_field(i)%p = real(recv_field_r8(i)%p)
    deallocate(recv_field_r8(i)%p)
  end do
  do i = 1, collection_size
    do j = 1, send_nbr_pointsets
      deallocate(send_field_r8(j, i)%p)
    end do
  end do

end subroutine yac_fexchange_r4_ptr

subroutine yac_fexchange_single_pointset_r4 ( send_field_id,       &
     &                                        recv_field_id,       &
     &                                        send_nbr_hor_points, &
     &                                        recv_nbr_hor_points, &
     &                                        collection_size,     &
     &                                        send_field,          &
     &                                        recv_field,          &
     &                                        send_info,           &
     &                                        recv_info,           &
     &                                        ierror )

  use mo_yac_finterface, dummy => yac_fexchange_single_pointset_r4

  implicit none

  integer, intent (in)  :: send_field_id        !< [IN] send field identifier
  integer, intent (in)  :: recv_field_id        !< [IN] recv field identifier
  integer, intent (in)  :: send_nbr_hor_points  !< [IN] number of horizontal send points
  integer, intent (in)  :: recv_nbr_hor_points  !< [IN] number of horizontal recv points
  integer, intent (in)  :: collection_size      !< [IN] number of vertical level or bundles
  real, intent (in)     :: send_field(send_nbr_hor_points, &
                                      collection_size)
                                                !< [IN] returned send field
  real, intent (inout)  :: recv_field(recv_nbr_hor_points, &
                                      collection_size)
                                                !< [INOUT] returned recv field
  integer, intent (out) :: send_info            !< [OUT] returned send info
  integer, intent (out) :: recv_info            !< [OUT] returned recv info
  integer, intent (out) :: ierror               !< [OUT] returned error

  double precision :: send_buffer(send_nbr_hor_points, collection_size)
  double precision :: recv_buffer(recv_nbr_hor_points, collection_size)

  send_buffer = DBLE(send_field)
  recv_buffer = DBLE(recv_field)

  call yac_fexchange_single_pointset_r8 ( send_field_id,       &
                                          recv_field_id,       &
                                          send_nbr_hor_points, &
                                          recv_nbr_hor_points, &
                                          collection_size,     &
                                          send_buffer,         &
                                          recv_buffer,         &
                                          send_info,           &
                                          recv_info,           &
                                          ierror )

  recv_field = REAL(recv_buffer)

end subroutine yac_fexchange_single_pointset_r4

!>
!! @param[in]    send_field_id       send field identifier
!! @param[in]    recv_field_id       recv field identifier
!! @param[in]    send_nbr_hor_points number of horizontal send points
!! @param[in]    send_nbr_pointsets  number of send point sets
!! @param[in]    recv_nbr_hor_points number of horizontal recv points
!! @param[in]    collection_size     number of vertical level or bundles
!! @param[in]    send_field          returned send field
!! @param[inout] recv_field          returned recv field
!! @param[out]   send_info           returned send info
!! @param[out]   recv_info           returned recv info
!! @param[out]   ierror              returned error
subroutine yac_fexchange_r8 ( send_field_id,       &
     &                        recv_field_id,       &
     &                        send_nbr_hor_points, &
     &                        send_nbr_pointsets,  &
     &                        recv_nbr_hor_points, &
     &                        collection_size,     &
     &                        send_field,          &
     &                        recv_field,          &
     &                        send_info,           &
     &                        recv_info,           &
     &                        ierror )

  use mo_yac_finterface, dummy => yac_fexchange_r8

  implicit none

  interface

     subroutine yac_cexchange_ ( send_field_id,   &
                                 recv_field_id,   &
                                 collection_size, &
                                 send_field,      &
                                 recv_field,      &
                                 send_info,       &
                                 recv_info,       &
                                 ierror )         &
                                 bind ( c, name='yac_cexchange_' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int ), value :: send_field_id
       integer ( kind=c_int ), value :: recv_field_id
       integer ( kind=c_int ), value :: collection_size
       real    ( kind=c_double )     :: send_field(*)
       real    ( kind=c_double )     :: recv_field(*)
       integer ( kind=c_int )        :: send_info
       integer ( kind=c_int )        :: recv_info
       integer ( kind=c_int )        :: ierror

     end subroutine yac_cexchange_

  end interface

  integer, intent (in)  :: send_field_id
  integer, intent (in)  :: recv_field_id
  integer, intent (in)  :: send_nbr_hor_points
  integer, intent (in)  :: send_nbr_pointsets
  integer, intent (in)  :: recv_nbr_hor_points
  integer, intent (in)  :: collection_size
  double precision, intent (in) ::               &
                           send_field(           &
                            send_nbr_hor_points, &
                            send_nbr_pointsets,  &
                            collection_size)
  double precision, intent (inout)::             &
                           recv_field(           &
                            recv_nbr_hor_points, &
                            collection_size)
  integer, intent (out) :: send_info
  integer, intent (out) :: recv_info
  integer, intent (out) :: ierror

  call yac_cexchange_ ( send_field_id,   &
                        recv_field_id,   &
                        collection_size, &
                        send_field,      &
                        recv_field,      &
                        send_info,       &
                        recv_info,       &
                        ierror )

end subroutine yac_fexchange_r8

!>
!! @param[in]    send_field_id       send field identifier
!! @param[in]    recv_field_id       recv field identifier
!! @param[in]    send_nbr_pointsets  number of send point sets
!! @param[in]    collection_size     number of vertical level or bundles
!! @param[in]    send_field          returned send field handle
!! @param[inout] recv_field          returned recv field handle
!! @param[out]   send_info           returned send info
!! @param[out]   recv_info           returned recv info
!! @param[out]   ierror              returned error
subroutine yac_fexchange_r8_ptr ( send_field_id,      &
     &                            recv_field_id,      &
     &                            send_nbr_pointsets, &
     &                            collection_size,    &
     &                            send_field,         &
     &                            recv_field,         &
     &                            send_info,          &
     &                            recv_info,          &
     &                            ierror )

  use mo_yac_finterface, dummy => yac_fexchange_r8_ptr
  use iso_c_binding, only: c_ptr, c_loc

  implicit none

  interface

     subroutine yac_cexchange_ptr_ ( send_field_id,   &
                                     recv_field_id,   &
                                     collection_size, &
                                     send_field,      &
                                     recv_field,      &
                                     send_info,       &
                                     recv_info,       &
                                     ierror )         &
                                     bind ( c, name='yac_cexchange_ptr_' )

       use, intrinsic :: iso_c_binding, only : c_int, c_ptr

       integer ( kind=c_int ), value :: send_field_id
       integer ( kind=c_int ), value :: recv_field_id
       integer ( kind=c_int ), value :: collection_size
       type(c_ptr)                   :: send_field(*)
       type(c_ptr)                   :: recv_field(*)
       integer ( kind=c_int )        :: send_info
       integer ( kind=c_int )        :: recv_info
       integer ( kind=c_int )        :: ierror

     end subroutine yac_cexchange_ptr_

  end interface

  integer, intent (in)  :: send_field_id
  integer, intent (in)  :: recv_field_id
  integer, intent (in)  :: send_nbr_pointsets
  integer, intent (in)  :: collection_size
  type(yac_r8_ptr), intent (in) ::                        &
                           send_field(send_nbr_pointsets, &
                                      collection_size)
  type(yac_r8_ptr)      :: recv_field(collection_size)
  integer, intent (out) :: send_info
  integer, intent (out) :: recv_info
  integer, intent (out) :: ierror

  integer :: i, j
  type(c_ptr) :: send_field_(send_nbr_pointsets, collection_size)
  type(c_ptr) :: recv_field_(collection_size)

  do i = 1, collection_size
    do j = 1, send_nbr_pointsets
      send_field_(j, i) = c_loc(send_field(j, i)%p(1))
    end do
  end do
  do i = 1, collection_size
    recv_field_(i) = c_loc(recv_field(i)%p(1))
  end do

  call yac_cexchange_ptr_ ( send_field_id,   &
                            recv_field_id,   &
                            collection_size, &
                            send_field_,     &
                            recv_field_,     &
                            send_info,       &
                            recv_info,       &
                            ierror )

end subroutine yac_fexchange_r8_ptr

!>
!! @param[in]    send_field_id       send field identifier
!! @param[in]    recv_field_id       recv field identifier
!! @param[in]    send_nbr_hor_points number of horizontal send points
!! @param[in]    recv_nbr_hor_points number of horizontal recv points
!! @param[in]    collection_size     number of vertical level or bundles
!! @param[in]    send_field          returned send field
!! @param[inout] recv_field          returned recv field
!! @param[out]   send_info           returned send info
!! @param[out]   recv_info           returned recv info
!! @param[out]   ierror              returned error
subroutine yac_fexchange_single_pointset_r8 ( send_field_id,       &
     &                                        recv_field_id,       &
     &                                        send_nbr_hor_points, &
     &                                        recv_nbr_hor_points, &
     &                                        collection_size,     &
     &                                        send_field,          &
     &                                        recv_field,          &
     &                                        send_info,           &
     &                                        recv_info,           &
     &                                        ierror )

  use mo_yac_finterface, dummy => yac_fexchange_single_pointset_r8

  implicit none

  interface

     subroutine yac_cexchange_ ( send_field_id,   &
                                 recv_field_id,   &
                                 collection_size, &
                                 send_field,      &
                                 recv_field,      &
                                 send_info,       &
                                 recv_info,       &
                                 ierror )         &
                                 bind ( c, name='yac_cexchange_' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int ), value :: send_field_id
       integer ( kind=c_int ), value :: recv_field_id
       integer ( kind=c_int ), value :: collection_size
       real    ( kind=c_double )     :: send_field(*)
       real    ( kind=c_double )     :: recv_field(*)
       integer ( kind=c_int )        :: send_info
       integer ( kind=c_int )        :: recv_info
       integer ( kind=c_int )        :: ierror

     end subroutine yac_cexchange_

  end interface

  integer, intent (in)  :: send_field_id
  integer, intent (in)  :: recv_field_id
  integer, intent (in)  :: send_nbr_hor_points
  integer, intent (in)  :: recv_nbr_hor_points
  integer, intent (in)  :: collection_size
  double precision, intent (in) ::                         &
                           send_field(send_nbr_hor_points, &
                                      collection_size)
  double precision, intent (inout)::                       &
                           recv_field(recv_nbr_hor_points, &
                                      collection_size)
  integer, intent (out) :: send_info
  integer, intent (out) :: recv_info
  integer, intent (out) :: ierror

  call yac_cexchange_ ( send_field_id,   &
                        recv_field_id,   &
                        collection_size, &
                        send_field,      &
                        recv_field,      &
                        send_info,       &
                        recv_info,       &
                        ierror )

end subroutine yac_fexchange_single_pointset_r8

! ----------------------------------------------------------------------

subroutine yac_fget_localcomm ( local_communicator, comp_id )

   use mo_yac_finterface, dummy => yac_fget_localcomm

   implicit none

   interface

     subroutine yac_get_localcomm ( local_communicator, comp_id ) &
       bind ( c, name='yac_get_localcomm_f2c' )

       use, intrinsic :: iso_c_binding, only : c_int
       use mo_yac_finterface, only : YAC_MPI_FINT_KIND

       integer ( kind=YAC_MPI_FINT_KIND ) :: local_communicator
       integer ( kind=c_int ), value      :: comp_id

     end subroutine yac_get_localcomm

   end interface

   integer, intent (out) :: local_communicator !< [OUT] component MPI communicator
   integer, intent (in)  :: comp_id            !< [IN] component identifier

   call yac_get_localcomm ( local_communicator, comp_id )

end subroutine yac_fget_localcomm

! ----------------------------------------------------------------------

subroutine yac_fget_comp_io_proc_step ( comp_id, io_proc_step )

   use mo_yac_finterface, dummy => yac_fget_comp_io_proc_step

   implicit none

   interface

     subroutine yac_cget_comp_io_proc_step ( comp_id, io_proc_step ) &
       bind ( c, name='yac_cget_comp_io_proc_step' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: comp_id
       integer ( kind=c_int ) :: io_proc_step

     end subroutine yac_cget_comp_io_proc_step

   end interface

   integer, intent (in)  :: comp_id            !< [IN] component identifier
   integer, intent (out) :: io_proc_step       !< [OUT] stride between ranks of io processes

   call yac_cget_comp_io_proc_step ( comp_id, io_proc_step )

end subroutine yac_fget_comp_io_proc_step

! ----------------------------------------------------------------------

subroutine yac_fset_comp_io_proc_step ( comp_id, io_proc_step )

   use mo_yac_finterface, dummy => yac_fset_comp_io_proc_step

   implicit none

   interface

     subroutine yac_cset_comp_io_proc_step ( comp_id, io_proc_step ) &
       bind ( c, name='yac_cset_comp_io_proc_step' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: comp_id
       integer ( kind=c_int ), value :: io_proc_step

     end subroutine yac_cset_comp_io_proc_step

   end interface

   integer, intent (in) :: comp_id            !< [IN] component identifier
   integer, intent (in) :: io_proc_step       !< [IN] stride between ranks of io processes

   call yac_cset_comp_io_proc_step ( comp_id, io_proc_step )

end subroutine yac_fset_comp_io_proc_step

! ----------------------------------------------------------------------

subroutine yac_fget_pair_rootcomm ( local_communicator, comp_name1, comp_name2 )

   use, intrinsic :: iso_c_binding, only : c_null_char
   use mo_yac_finterface, dummy => yac_fget_pair_rootcomm

   implicit none

   interface

     subroutine yac_cget_pair_rootcomm ( local_communicator, &
                                         comp_name1,         &
                                         comp_name2 )        &
    &      bind ( c, name='yac_cget_pair_rootcomm_f2c' )

       use, intrinsic :: iso_c_binding, only : c_char
       use mo_yac_finterface, only : YAC_MPI_FINT_KIND

       integer ( kind=YAC_MPI_FINT_KIND )     :: local_communicator
       character ( kind=c_char), dimension(*) :: comp_name1
       character ( kind=c_char), dimension(*) :: comp_name2

     end subroutine yac_cget_pair_rootcomm

   end interface

   integer, intent (out) :: local_communicator !< [OUT] component MPI communicator
   character(len=*), intent(in) :: comp_name1  !< [IN]  first component name
   character(len=*), intent(in) :: comp_name2  !< [IN]  second component name

   call yac_cget_pair_rootcomm ( local_communicator,              &
                                 TRIM(comp_name1) // c_null_char, &
                                 TRIM(comp_name2) // c_null_char  )

end subroutine yac_fget_pair_rootcomm

subroutine yac_fget_pair_rootcomm_instance ( yac_instance_id,    &
                                             local_communicator, &
                                             comp_name1, comp_name2 )

   use, intrinsic :: iso_c_binding, only : c_null_char
   use mo_yac_finterface, dummy => yac_fget_pair_rootcomm_instance

   implicit none

   interface

     subroutine yac_cget_pair_rootcomm_instance ( yac_instance_id,    &
                                                  local_communicator, &
                                                  comp_name1,         &
                                                  comp_name2 )        &
    &      bind ( c, name='yac_cget_pair_rootcomm_instance_f2c' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char
       use mo_yac_finterface, only : YAC_MPI_FINT_KIND

       integer   ( kind=c_int ), value        :: yac_instance_id
       integer ( kind=YAC_MPI_FINT_KIND )     :: local_communicator
       character ( kind=c_char), dimension(*) :: comp_name1
       character ( kind=c_char), dimension(*) :: comp_name2

     end subroutine yac_cget_pair_rootcomm_instance

   end interface

   integer, intent(in)          :: yac_instance_id    !< [IN]  YAC instance identifier
   integer, intent(out)         :: local_communicator !< [OUT] component MPI communicator
   character(len=*), intent(in) :: comp_name1         !< [IN]  first component name
   character(len=*), intent(in) :: comp_name2         !< [IN]  second component name

   call yac_cget_pair_rootcomm_instance ( yac_instance_id,                 &
                                          local_communicator,              &
                                          TRIM(comp_name1) // c_null_char, &
                                          TRIM(comp_name2) // c_null_char  )

end subroutine yac_fget_pair_rootcomm_instance

! ------------------- search/end of definition -------------------------

subroutine yac_fsearch ( ierror )

   use mo_yac_finterface, dummy => yac_fsearch

   implicit none

   interface

     subroutine yac_csearch ( ierror ) bind ( c, name='yac_csearch' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ) :: ierror !< [OUT] returned error handler

     end subroutine yac_csearch

   end interface

   integer, intent (out) :: ierror !< [OUT] returned error

   call yac_csearch ( ierror )

end subroutine yac_fsearch

subroutine yac_fsearch_instance ( yac_instance_id, ierror )

   use mo_yac_finterface, dummy => yac_fsearch_instance

   implicit none

   interface

     subroutine yac_csearch_instance ( yac_instance_id, ierror ) &
       bind ( c, name='yac_csearch_instance' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: yac_instance_id
       integer ( kind=c_int )        :: ierror

     end subroutine yac_csearch_instance

   end interface

   integer, intent(in)  :: yac_instance_id !< [IN]  YAC instance identifier
   integer, intent(out) :: ierror          !< [OUT] returned error

   call yac_csearch_instance ( yac_instance_id, ierror )

end subroutine yac_fsearch_instance

! ------------------------ query routines -----------------------------

subroutine yac_fget_nbr_fields ( nbr_fields )

   use mo_yac_finterface, dummy => yac_fget_nbr_fields

   implicit none

   interface

     subroutine yac_cget_nbr_fields ( nbr_fields ) &
           bind ( c, name='yac_cget_nbr_fields' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ) :: nbr_fields !< [OUT]  number of components

     end subroutine yac_cget_nbr_fields

   end interface

   integer, intent (out) :: nbr_fields !< [OUT] number of coupling fields

   call yac_cget_nbr_fields ( nbr_fields )

end subroutine yac_fget_nbr_fields

subroutine yac_fget_nbr_fields_instance ( yac_instance_id, nbr_fields )

   use mo_yac_finterface, dummy => yac_fget_nbr_fields_instance

   implicit none

   interface

     subroutine yac_cget_nbr_fields_instance ( yac_instance_id, nbr_fields ) &
           bind ( c, name='yac_cget_nbr_fields_instance' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: yac_instance_id
       integer ( kind=c_int )        :: nbr_fields

     end subroutine yac_cget_nbr_fields_instance

   end interface

   integer, intent(in)   :: yac_instance_id !< [IN]  YAC instance identifier
   integer, intent (out) :: nbr_fields      !< [OUT] number of coupling fields

   call yac_cget_nbr_fields_instance ( yac_instance_id, nbr_fields )

end subroutine yac_fget_nbr_fields_instance

! ---------------------------------------------------------------------

subroutine yac_fget_field_ids ( nbr_fields, field_ids )

   use mo_yac_finterface, dummy => yac_fget_field_ids

   implicit none

   interface

     subroutine yac_cget_field_ids ( nbr_fields, field_ids ) &
       bind ( c, name='yac_cget_field_ids' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: nbr_fields
       integer ( kind=c_int )        :: field_ids(nbr_fields)

     end subroutine yac_cget_field_ids

   end interface

   integer, intent (in)  :: nbr_fields            !< [IN] number of coupling fields
   integer, intent (out) :: field_ids(nbr_fields) !< [OUT] list of field IDs

   call yac_cget_field_ids ( nbr_fields, field_ids )

end subroutine yac_fget_field_ids

subroutine yac_fget_field_ids_instance ( yac_instance_id, &
                                         nbr_fields,      &
                                         field_ids )

   use mo_yac_finterface, dummy => yac_fget_field_ids_instance

   implicit none

   interface

     subroutine yac_cget_field_ids_instance ( yac_instance_id, &
                                              nbr_fields,      &
                                              field_ids )      &
       bind ( c, name='yac_cget_field_ids_instance' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: yac_instance_id
       integer ( kind=c_int ), value :: nbr_fields
       integer ( kind=c_int )        :: field_ids(nbr_fields)

     end subroutine yac_cget_field_ids_instance

   end interface

   integer, intent(in)   :: yac_instance_id       !< [IN]  YAC instance identifier
   integer, intent (in)  :: nbr_fields            !< [IN] number of coupling fields
   integer, intent (out) :: field_ids(nbr_fields) !< [OUT] list of field IDs

   call yac_cget_field_ids_instance ( yac_instance_id, &
                                      nbr_fields,      &
                                      field_ids )

end subroutine yac_fget_field_ids_instance

! ---------------------------------------------------------------------

subroutine yac_fget_role_of_field_by_id ( field_id, role )

   use mo_yac_finterface, dummy => yac_fget_role_of_field_by_id

   implicit none

   interface

     subroutine yac_cget_role_of_field_by_id ( field_id, role ) &
       bind ( c, name='yac_cget_role_of_field_by_id' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value  :: field_id
       integer ( kind=c_int)          :: role

    end subroutine yac_cget_role_of_field_by_id

   end interface

   integer, intent (in)  :: field_id !< [IN]  field identifier
   integer, intent (out) :: role     !< [OUT] role for this field\n
                                     !!       NOTHING (0) | SOURCE (1) | TARGET (2)

   call yac_cget_role_of_field_by_id ( field_id, role )

end subroutine yac_fget_role_of_field_by_id

! ---------------------------------------------------------------------

subroutine yac_fget_role_of_field_by_name ( field_name, role )

   use mo_yac_finterface, dummy => yac_fget_role_of_field_by_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

    subroutine yac_cget_role_of_field_by_name ( field_name, role ) &
      bind ( c, name='yac_cget_role_of_field_by_name' )

      use, intrinsic :: iso_c_binding, only : c_int, c_char

      character ( kind=c_char), dimension(*) :: field_name
      integer ( kind=c_int)                  :: role

    end subroutine yac_cget_role_of_field_by_name

   end interface

   character (len=*), intent (in)    :: field_name !< [IN] field name
   integer, intent (out)             :: role       !< [OUT] role for this field\n
                                                   !!       NOTHING (0) | SOURCE (1) | TARGET (2)

   call yac_check_strlength ( field_name )

   call yac_cget_role_of_field_by_name ( TRIM(field_name) // c_null_char, &
                                         role )

end subroutine yac_fget_role_of_field_by_name

subroutine yac_fget_role_of_field_by_name_instance ( yac_instance_id, &
                                                     field_name,      &
                                                     role )

   use mo_yac_finterface, dummy => yac_fget_role_of_field_by_name_instance
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

    subroutine yac_cget_role_of_field_by_name_instance ( yac_instance_id, &
                                                         field_name,      &
                                                         role )           &
      bind ( c, name='yac_cget_role_of_field_by_name_instance' )

      use, intrinsic :: iso_c_binding, only : c_int, c_char

      integer ( kind=c_int ), value          :: yac_instance_id
      character ( kind=c_char), dimension(*) :: field_name
      integer ( kind=c_int)                  :: role

    end subroutine yac_cget_role_of_field_by_name_instance

   end interface

   integer, intent(in)            :: yac_instance_id !< [IN]  YAC instance identifier
   character (len=*), intent (in) :: field_name      !< [IN]  field name
   integer, intent (out)          :: role            !< [OUT] role for this field\n
                                                     !!       NOTHING (0) | SOURCE (1) | TARGET (2)

   call yac_check_strlength ( field_name )

   call yac_cget_role_of_field_by_name_instance ( &
      yac_instance_id,                            &
      TRIM(field_name) // c_null_char,            &
      role )

end subroutine yac_fget_role_of_field_by_name_instance

! ---------------------------------------------------------------------

subroutine yac_fget_role_of_field_by_field_and_grid_name ( &
   field_name, grid_name, role )

   use mo_yac_finterface, &
      dummy => yac_fget_role_of_field_by_field_and_grid_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

    subroutine yac_cget_role_of_field_by_field_and_grid_name ( &
      field_name, grid_name, role )                            &
      bind ( c, name='yac_cget_role_of_field_by_field_and_grid_name' )

      use, intrinsic :: iso_c_binding, only : c_int, c_char

      character ( kind=c_char), dimension(*) :: field_name
      character ( kind=c_char), dimension(*) :: grid_name
      integer ( kind=c_int)                  :: role

    end subroutine yac_cget_role_of_field_by_field_and_grid_name

   end interface

   character (len=*), intent (in)    :: field_name !< [IN] field name
   character (len=*), intent (in)    :: grid_name  !< [IN] grid name
   integer, intent (out)             :: role       !< [OUT] role for this field\n
                                                   !!       NOTHING (0) | SOURCE (1) | TARGET (2)

   call yac_check_strlength ( field_name )
   call yac_check_strlength ( grid_name )

   call yac_cget_role_of_field_by_field_and_grid_name ( &
      TRIM(field_name) // c_null_char, &
      TRIM(grid_name) // c_null_char, role )

end subroutine yac_fget_role_of_field_by_field_and_grid_name

subroutine yac_fget_role_of_field_by_field_and_grid_name_instance ( &
   yac_instance_id, field_name, grid_name, role )

   use mo_yac_finterface, &
      dummy => yac_fget_role_of_field_by_field_and_grid_name_instance
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

    subroutine yac_cget_role_of_field_by_field_and_grid_name_instance ( &
      yac_instance_id, field_name, grid_name, role )                    &
      bind ( c, name='yac_cget_role_of_field_by_field_and_grid_name_instance' )

      use, intrinsic :: iso_c_binding, only : c_int, c_char

      integer ( kind=c_int ), value          :: yac_instance_id
      character ( kind=c_char), dimension(*) :: field_name
      character ( kind=c_char), dimension(*) :: grid_name
      integer ( kind=c_int)                  :: role

    end subroutine yac_cget_role_of_field_by_field_and_grid_name_instance

   end interface

   integer, intent(in)            :: yac_instance_id !< [IN]  YAC instance identifier
   character (len=*), intent (in) :: field_name      !< [IN]  field name
   character (len=*), intent (in) :: grid_name       !< [IN]  grid name
   integer, intent (out)          :: role            !< [OUT] role for this field\n
                                                     !!       NOTHING (0) | SOURCE (1) | TARGET (2)

   call yac_check_strlength ( field_name )
   call yac_check_strlength ( grid_name )

   call yac_cget_role_of_field_by_field_and_grid_name_instance ( &
      yac_instance_id,                                           &
      TRIM(field_name) // c_null_char,                           &
      TRIM(grid_name) // c_null_char,                            &
      role )

end subroutine yac_fget_role_of_field_by_field_and_grid_name_instance

! ---------------------------------------------------------------------

subroutine yac_fget_field_name_from_id ( field_id, field_name )

   use mo_yac_finterface, dummy => yac_fget_field_name_from_id
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

     subroutine yac_cget_field_name_from_id ( field_id, field_name ) &
       bind ( c, name='yac_cget_field_name_from_id' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer ( kind=c_int ), value          :: field_id   !< [IN]  field ID
       character ( kind=c_char), dimension(*) :: field_name !< [OUT] field name

     end subroutine yac_cget_field_name_from_id

   end interface

   integer, intent (in)            :: field_id   !< [IN]  field identifier
   character (len=*), intent (out) :: field_name !< [OUT] field name

   integer :: i

   call yac_check_strlength ( field_name )

   call yac_cget_field_name_from_id ( field_id, field_name )

   do i = 1, len(field_name)
      if (field_name(i:i) == c_null_char) exit
   end do

   field_name(i:len(field_name)) = ' '

end subroutine yac_fget_field_name_from_id

! ---------------------------------------------------------------------

subroutine yac_fget_field_id_from_name ( field_name, field_id )

   use mo_yac_finterface, dummy => yac_fget_field_id_from_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

     subroutine yac_cget_field_id_from_name ( field_name, field_id ) &
       bind ( c, name='yac_cget_field_id_from_name' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char), dimension(*) :: field_name
       integer ( kind=c_int )                 :: field_id

     end subroutine yac_cget_field_id_from_name

   end interface

   character (len=*), intent (in) :: field_name !< [IN] field name
   integer, intent (out)          :: field_id   !< [OUT]  field identifier

   call yac_check_strlength ( field_name )

   call yac_cget_field_id_from_name ( TRIM(field_name) // c_null_char, &
                                      field_id)

end subroutine yac_fget_field_id_from_name

subroutine yac_fget_field_id_from_name_instance ( yac_instance_id, &
                                                  field_name,      &
                                                  field_id )

   use mo_yac_finterface, dummy => yac_fget_field_id_from_name_instance
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

     subroutine yac_cget_field_id_from_name_instance ( yac_instance_id, &
                                                       field_name,      &
                                                       field_id )       &
       bind ( c, name='yac_cget_field_id_from_name_instance' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer ( kind=c_int ), value          :: yac_instance_id
       character ( kind=c_char), dimension(*) :: field_name
       integer ( kind=c_int )                 :: field_id

     end subroutine yac_cget_field_id_from_name_instance

   end interface

   integer, intent(in)            :: yac_instance_id !< [IN]  YAC instance identifier
   character (len=*), intent (in) :: field_name      !< [IN]  field name
   integer, intent (out)          :: field_id        !< [OUT] field identifier

   call yac_check_strlength ( field_name )

   call yac_cget_field_id_from_name_instance ( &
      yac_instance_id,                         &
      TRIM(field_name) // c_null_char,         &
      field_id)

end subroutine yac_fget_field_id_from_name_instance

! ---------------------------------------------------------------------

subroutine yac_fget_field_id_from_field_and_grid_name ( &
   field_name, grid_name, field_id )

   use mo_yac_finterface, &
      dummy => yac_fget_field_id_from_field_and_grid_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

     subroutine yac_cget_field_id_from_field_and_grid_name ( &
       field_name, grid_name, field_id ) &
       bind ( c, name='yac_cget_field_id_from_field_and_grid_name' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char), dimension(*) :: field_name
       character ( kind=c_char), dimension(*) :: grid_name
       integer ( kind=c_int )                 :: field_id

     end subroutine yac_cget_field_id_from_field_and_grid_name

   end interface

   character (len=*), intent (in) :: field_name !< [IN] field name
   character (len=*), intent (in) :: grid_name  !< [IN] grid name
   integer, intent (out)          :: field_id   !< [OUT]  field identifier

   call yac_check_strlength ( field_name )
   call yac_check_strlength ( grid_name )

   call yac_cget_field_id_from_field_and_grid_name ( &
      TRIM(field_name) // c_null_char, &
      TRIM(grid_name) // c_null_char, field_id)

end subroutine yac_fget_field_id_from_field_and_grid_name

subroutine yac_fget_field_id_from_field_and_grid_name_instance ( &
   yac_instance_id, field_name, grid_name, field_id )

   use mo_yac_finterface, &
      dummy => yac_fget_field_id_from_field_and_grid_name_instance
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

     subroutine yac_cget_field_id_from_field_and_grid_name_instance ( &
       yac_instance_id, field_name, grid_name, field_id )             &
       bind ( c, name='yac_cget_field_id_from_field_and_grid_name_instance' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer ( kind=c_int ), value          :: yac_instance_id
       character ( kind=c_char), dimension(*) :: field_name
       character ( kind=c_char), dimension(*) :: grid_name
       integer ( kind=c_int )                 :: field_id

     end subroutine yac_cget_field_id_from_field_and_grid_name_instance

   end interface

   integer, intent(in)            :: yac_instance_id !< [IN]  YAC instance identifier
   character (len=*), intent (in) :: field_name      !< [IN]  field name
   character (len=*), intent (in) :: grid_name       !< [IN]  grid name
   integer, intent (out)          :: field_id        !< [OUT] field identifier

   call yac_check_strlength ( field_name )
   call yac_check_strlength ( grid_name )

   call yac_cget_field_id_from_field_and_grid_name_instance ( &
      yac_instance_id,                         &
      TRIM(field_name) // c_null_char,         &
      TRIM(grid_name) // c_null_char,         &
      field_id)

end subroutine yac_fget_field_id_from_field_and_grid_name_instance

! ---------------------------------------------------------------------

subroutine yac_fget_model_timestep_id ( field_id, string )

   use mo_yac_finterface, dummy => yac_fget_model_timestep_id
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

     subroutine yac_cget_model_timestep_id ( field_id, string ) &
       bind ( c, name='yac_cget_model_timestep_id' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer ( kind=c_int ), value          :: field_id
       character ( kind=c_char), dimension(*) :: string

     end subroutine yac_cget_model_timestep_id

    end interface

   integer, intent (in)            :: field_id !< [IN]  field identifier
   character (len=*), intent (out) :: string   !< [OUT] timestep in iso format

   integer :: i

   call yac_cget_model_timestep_id ( field_id, string )

   do i = 1, len(string)
      if (string(i:i) == c_null_char) exit
   end do

   string(i:len(string)) = ' '

end subroutine yac_fget_model_timestep_id

! ---------------------------------------------------------------------

subroutine yac_fget_model_timestep_name ( field_name, string )

   use mo_yac_finterface, dummy => yac_fget_model_timestep_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

     subroutine yac_cget_model_timestep_name ( field_name, string ) &
    &      bind ( c, name='yac_cget_model_timestep_name' )

       use, intrinsic :: iso_c_binding, only : c_char

       character ( kind=c_char), dimension(*) :: field_name
       character ( kind=c_char), dimension(*) :: string

     end subroutine yac_cget_model_timestep_name

   end interface

   character (len=*), intent (in)  :: field_name !< [IN] field name
   character (len=*), intent (out) :: string     !< [OUT] timestep in iso format

   integer :: i

   call yac_check_strlength ( field_name )

   call yac_cget_model_timestep_name ( TRIM(field_name) // c_null_char, &
                                       string )

   do i = 1, len(string)
      if (string(i:i) == c_null_char) exit
   end do

   string(i:len(string)) = ' '

end subroutine yac_fget_model_timestep_name

subroutine yac_fget_model_timestep_name_instance ( yac_instance_id, &
                                                   field_name,      &
                                                   string )

   use mo_yac_finterface, dummy => yac_fget_model_timestep_name_instance
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

     subroutine yac_cget_model_timestep_name_instance ( yac_instance_id, &
                                                        field_name,      &
                                                        string ) &
       bind ( c, name='yac_cget_model_timestep_name_instance' )

       use, intrinsic :: iso_c_binding, only : c_char, c_int

       integer ( kind=c_int ), value          :: yac_instance_id
       character ( kind=c_char), dimension(*) :: field_name
       character ( kind=c_char), dimension(*) :: string

     end subroutine yac_cget_model_timestep_name_instance

   end interface

   integer, intent(in)             :: yac_instance_id !< [IN]  YAC instance identifier
   character (len=*), intent (in)  :: field_name      !< [IN]  field name
   character (len=*), intent (out) :: string          !< [OUT] timestep in iso format

   integer :: i

   call yac_check_strlength ( field_name )

   call yac_cget_model_timestep_name_instance ( &
      yac_instance_id,                          &
      TRIM(field_name) // c_null_char,          &
      string )

   do i = 1, len(string)
      if (string(i:i) == c_null_char) exit
   end do

   string(i:len(string)) = ' '

end subroutine yac_fget_model_timestep_name_instance

! ---------------------------------------------------------------------

subroutine yac_fget_coupling_timestep_id ( field_id, string )

   use mo_yac_finterface, dummy => yac_fget_coupling_timestep_id
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

     subroutine yac_cget_coupling_timestep_id ( field_id, string ) &
    &      bind ( c, name='yac_cget_coupling_timestep_id' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer ( kind=c_int ), value          :: field_id
       character ( kind=c_char), dimension(*) :: string

     end subroutine yac_cget_coupling_timestep_id

   end interface

   integer, intent (in)            :: field_id !< [IN]  field identifier
   character (len=*), intent (out) :: string   !< [OUT] timestep in iso format

   integer :: i

   call yac_cget_coupling_timestep_id ( field_id, string )

   do i = 1, len(string)
      if (string(i:i) == c_null_char) exit
   end do

   string(i:len(string)) = ' '

end subroutine yac_fget_coupling_timestep_id

! ---------------------------------------------------------------------

subroutine yac_fget_coupling_timestep_name ( field_name, string )

   use mo_yac_finterface, dummy => yac_fget_coupling_timestep_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

     subroutine yac_cget_coupling_timestep_name ( field_name, string ) &
    &      bind ( c, name='yac_cget_coupling_timestep_name' )

       use, intrinsic :: iso_c_binding, only : c_char

       character ( kind=c_char), dimension(*) :: field_name
       character ( kind=c_char), dimension(*) :: string

     end subroutine yac_cget_coupling_timestep_name

   end interface

   character (len=*), intent (in)  :: field_name !< [IN] field name
   character (len=*), intent (out) :: string     !< [OUT] timestep in iso format

   integer :: i

   call yac_check_strlength ( field_name )

   call yac_cget_coupling_timestep_name ( TRIM(field_name) // c_null_char, &
                                          string )

   do i = 1, len(string)
      if (string(i:i) == c_null_char) exit
   end do

   string(i:len(string)) = ' '

end subroutine yac_fget_coupling_timestep_name

subroutine yac_fget_coupling_timestep_name_instance ( yac_instance_id, &
                                                      field_name,      &
                                                      string )

   use mo_yac_finterface, dummy => yac_fget_coupling_timestep_name_instance
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

     subroutine yac_cget_coupling_timestep_name_instance ( yac_instance_id, &
                                                           field_name,      &
                                                           string ) &
       bind ( c, name='yac_cget_coupling_timestep_name_instance' )

       use, intrinsic :: iso_c_binding, only : c_char, c_int

       integer ( kind=c_int ), value          :: yac_instance_id
       character ( kind=c_char), dimension(*) :: field_name
       character ( kind=c_char), dimension(*) :: string

     end subroutine yac_cget_coupling_timestep_name_instance

   end interface

   integer, intent(in)             :: yac_instance_id !< [IN]  YAC instance identifier
   character (len=*), intent (in)  :: field_name      !< [IN]  field name
   character (len=*), intent (out) :: string          !< [OUT] timestep in iso format

   integer :: i

   call yac_check_strlength ( field_name )

   call yac_cget_coupling_timestep_name_instance ( &
      yac_instance_id,                             &
      TRIM(field_name) // c_null_char,             &
      string )

   do i = 1, len(string)
      if (string(i:i) == c_null_char) exit
   end do

   string(i:len(string)) = ' '

end subroutine yac_fget_coupling_timestep_name_instance

! ---------------------------------------------------------------------

subroutine yac_fget_collection_size_id ( field_id, collection_size )

   use mo_yac_finterface, dummy => yac_fget_collection_size_id

   implicit none

   interface

     subroutine yac_cget_collection_size_id ( field_id, collection_size ) &
    &      bind ( c, name='yac_cget_collection_size_id' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: field_id
       integer ( kind=c_int )        :: collection_size

     end subroutine yac_cget_collection_size_id

   end interface

   integer, intent (in)  :: field_id        !< [IN]  field identifier
   integer, intent (out) :: collection_size !< [OUT] collection size

   call yac_cget_collection_size_id ( field_id, collection_size )

end subroutine yac_fget_collection_size_id

! ---------------------------------------------------------------------

subroutine yac_fget_collection_size_name ( field_name, collection_size )

   use mo_yac_finterface, dummy => yac_fget_collection_size_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

     subroutine yac_cget_collection_size_name ( field_name, collection_size ) &
    &      bind ( c, name='yac_cget_collection_size_name' )

       use, intrinsic :: iso_c_binding, only : c_char, c_int

       character ( kind=c_char), dimension(*) :: field_name
       integer ( kind=c_int)                  :: collection_size

     end subroutine yac_cget_collection_size_name

   end interface

   character (len=*), intent (in) :: field_name      !< [IN] field name
   integer, intent (out)          :: collection_size !< [OUT] collection size

   call yac_check_strlength ( field_name )

   call yac_cget_collection_size_name ( TRIM(field_name) // c_null_char, &
                                        collection_size )

end subroutine yac_fget_collection_size_name

subroutine yac_fget_collection_size_name_instance ( yac_instance_id, &
                                                    field_name,      &
                                                    collection_size )

   use mo_yac_finterface, dummy => yac_fget_collection_size_name_instance
   use, intrinsic :: iso_c_binding, only: c_null_char

   implicit none

   interface

     subroutine yac_cget_collection_size_name_instance ( yac_instance_id,  &
                                                         field_name,       &
                                                         collection_size ) &
       bind ( c, name='yac_cget_collection_size_name_instance' )

       use, intrinsic :: iso_c_binding, only : c_char, c_int

       integer ( kind=c_int ), value          :: yac_instance_id
       character ( kind=c_char), dimension(*) :: field_name
       integer ( kind=c_int )                 :: collection_size

     end subroutine yac_cget_collection_size_name_instance

   end interface

   integer, intent(in)            :: yac_instance_id !< [IN]  YAC instance identifier
   character (len=*), intent (in) :: field_name      !< [IN]  field name
   integer, intent (out)          :: collection_size !< [OUT] collection size

   call yac_cget_collection_size_name_instance ( &
      yac_instance_id,                           &
      TRIM(field_name) // c_null_char,           &
      collection_size )

end subroutine yac_fget_collection_size_name_instance

! ---------------------------------------------------------------------

function yac_fget_start_datetime () result (start_datetime_string)

  use, intrinsic :: iso_c_binding, only : c_ptr, c_char, c_null_char, c_f_pointer
  use mo_yac_finterface, dummy => yac_fget_start_datetime

  implicit none

  interface
     function yac_cget_start_datetime () &
        bind ( c, name='yac_cget_start_datetime' )

     use, intrinsic :: iso_c_binding, only : c_ptr
     type(c_ptr) :: yac_cget_start_datetime

     end function yac_cget_start_datetime
  end interface

  type (c_ptr)                                           :: c_string_ptr
  character (len=1, kind=c_char), dimension (:), pointer :: start_datetime_chars => null ()
  character (len=YAC_MAX_CHARLEN)                        :: start_datetime_string
  integer :: i

  c_string_ptr = yac_cget_start_datetime()

  CALL c_f_pointer ( c_string_ptr, start_datetime_chars, [ YAC_MAX_CHARLEN ])

  start_datetime_string = " "

  do i=1, YAC_MAX_CHARLEN
    if (start_datetime_chars(i) == c_null_char) exit
    start_datetime_string(i:i) = start_datetime_chars(i)
  end do

end function yac_fget_start_datetime

function yac_fget_start_datetime_instance (yac_instance_id) &
  result (start_datetime_string)

  use, intrinsic :: iso_c_binding, only : c_ptr, c_char, c_null_char, c_f_pointer
  use mo_yac_finterface, dummy => yac_fget_start_datetime_instance

  implicit none

  interface
     function yac_cget_start_datetime_instance (yac_instance_id) &
        bind ( c, name='yac_cget_start_datetime_instance' )

     use, intrinsic :: iso_c_binding, only : c_ptr, c_int
     integer ( kind=c_int ), value :: yac_instance_id
     type(c_ptr)                   :: yac_cget_start_datetime_instance

     end function yac_cget_start_datetime_instance
  end interface

  integer, intent(in) :: yac_instance_id !< [IN]  YAC instance identifier

  type (c_ptr)                                           :: c_string_ptr
  character (len=1, kind=c_char), dimension (:), pointer :: start_datetime_chars => null ()
  character (len=YAC_MAX_CHARLEN)                        :: start_datetime_string
  integer :: i

  c_string_ptr = yac_cget_start_datetime_instance(yac_instance_id)

  CALL c_f_pointer ( c_string_ptr, start_datetime_chars, [ YAC_MAX_CHARLEN ])

  start_datetime_string = " "

  do i=1, YAC_MAX_CHARLEN
    if (start_datetime_chars(i) == c_null_char) exit
    start_datetime_string(i:i) = start_datetime_chars(i)
  end do

end function yac_fget_start_datetime_instance

! ---------------------------------------------------------------------

function yac_fget_end_datetime () result (end_datetime_string)

  use, intrinsic :: iso_c_binding, only : c_ptr, c_char, c_null_char, c_f_pointer
  use mo_yac_finterface, dummy => yac_fget_end_datetime

  implicit none

  interface
     function yac_cget_end_datetime () &
        bind ( c, name='yac_cget_end_datetime' )

     use, intrinsic :: iso_c_binding, only : c_ptr
     type(c_ptr) :: yac_cget_end_datetime

     end function yac_cget_end_datetime
  end interface

  type (c_ptr)                                           :: c_string_ptr
  character (len=1, kind=c_char), dimension (:), pointer :: end_datetime_chars => null ()
  character (len=YAC_MAX_CHARLEN)                        :: end_datetime_string
  integer :: i

  c_string_ptr = yac_cget_end_datetime()

  CALL c_f_pointer ( c_string_ptr, end_datetime_chars, [ YAC_MAX_CHARLEN ])

  end_datetime_string = " "

  do i=1, YAC_MAX_CHARLEN
    if (end_datetime_chars(i) == c_null_char) exit
    end_datetime_string(i:i) = end_datetime_chars(i)
  end do

end function yac_fget_end_datetime

function yac_fget_end_datetime_instance (yac_instance_id) &
  result (end_datetime_string)

  use, intrinsic :: iso_c_binding, only : c_ptr, c_char, c_null_char, c_f_pointer
  use mo_yac_finterface, dummy => yac_fget_end_datetime_instance

  implicit none

  interface
     function yac_cget_end_datetime_instance (yac_instance_id) &
        bind ( c, name='yac_cget_end_datetime_instance' )

     use, intrinsic :: iso_c_binding, only : c_ptr, c_int
     integer ( kind=c_int ), value :: yac_instance_id
     type(c_ptr)                   :: yac_cget_end_datetime_instance

     end function yac_cget_end_datetime_instance
  end interface

  integer, intent(in) :: yac_instance_id !< [IN]  YAC instance identifier

  type (c_ptr)                                           :: c_string_ptr
  character (len=1, kind=c_char), dimension (:), pointer :: end_datetime_chars => null ()
  character (len=YAC_MAX_CHARLEN)                        :: end_datetime_string
  integer :: i

  c_string_ptr = yac_cget_end_datetime_instance(yac_instance_id)

  CALL c_f_pointer ( c_string_ptr, end_datetime_chars, [ YAC_MAX_CHARLEN ])

  end_datetime_string = " "

  do i=1, YAC_MAX_CHARLEN
    if (end_datetime_chars(i) == c_null_char) exit
    end_datetime_string(i:i) = end_datetime_chars(i)
  end do

end function yac_fget_end_datetime_instance

! ---------------------------------------------------------------------

subroutine yac_check_strlength ( in_string )

  use mo_yac_finterface

  implicit none

  character (len=*), intent (in)  :: in_string
  character (len=YAC_MAX_CHARLEN) :: message

  if ( LEN_TRIM(in_string) > YAC_MAX_CHARLEN ) then
     message = "string too long"
     print *, "string ", TRIM(in_string), "exceeds length of ", YAC_MAX_CHARLEN
     call yac_abort_message ( message , __FILE__, __LINE__ )
  endif

end subroutine

! ---------------------------------------------------------------------

subroutine yac_redirstdout_f ( filestem, parallel, my_pe, npes )

  use mo_yac_finterface, dummy => yac_redirstdout_f
  use, intrinsic :: iso_c_binding, only : c_null_char

  implicit none

  interface
     subroutine yac_redirstdout_c ( filestem, parallel, my_pe, npes ) &
    &      bind ( c, name='yac_redirstdout' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char ), dimension(*) :: filestem
       integer ( kind=c_int ), value           :: parallel
       integer ( kind=c_int ), value           :: my_pe
       integer ( kind=c_int ), value           :: npes

     end subroutine yac_redirstdout_c
  end interface

  character(len=*), intent(in) :: filestem
  integer, intent(in) :: parallel
  integer, intent(in) :: my_pe
  integer, intent(in) :: npes

  call yac_check_strlength ( filestem )

  call yac_redirstdout_c ( &
     TRIM(filestem) // c_null_char, parallel, my_pe, npes)

end subroutine yac_redirstdout_f

! ---------------------------------------------------------------------

function yac_fget_transients_count( comp_id ) result (transients_count)

  implicit none

  integer, intent(in) :: comp_id
  integer             :: transients_count

  interface

    function yac_get_transients_count_f ( comp_id ) &
      result ( transients_count )                   &
      bind ( c, name='yac_get_transients_count' )

      use, intrinsic :: iso_c_binding, only : c_int

      implicit none

      integer ( kind=c_int ), value :: comp_id
      integer ( kind=c_int )        :: transients_count

    end function yac_get_transients_count_f

  end interface

  transients_count = yac_get_transients_count_f(comp_id)

end function yac_fget_transients_count

subroutine yac_fget_transient_ids( comp_id, transient_ids )

  implicit none

  integer, intent(in)  :: comp_id
  integer, intent(out) :: transient_ids(*)

  interface

    subroutine yac_get_transient_ids_f ( comp_id, transient_ids ) &
      bind ( c, name='yac_get_transient_ids' )

      use, intrinsic :: iso_c_binding, only : c_int

      integer ( kind=c_int ), value :: comp_id
      integer ( kind=c_int )        :: transient_ids(*)

    end subroutine yac_get_transient_ids_f

  end interface

  call yac_get_transient_ids_f ( comp_id, transient_ids )

end subroutine yac_fget_transient_ids

function yac_fis_transient_source( transient_id ) result (is_source)

  implicit none

  integer, intent(in) :: transient_id
  logical             :: is_source

  interface

    function yac_is_transient_source_f ( transient_id ) &
      result ( is_source )                   &
      bind ( c, name='yac_is_transient_source' )

      use, intrinsic :: iso_c_binding, only : c_int

      implicit none

      integer ( kind=c_int ), value :: transient_id
      integer ( kind=c_int )        :: is_source

    end function yac_is_transient_source_f

  end interface

  is_source = yac_is_transient_source_f(transient_id) /= 0

end function yac_fis_transient_source

function yac_fis_transient_target( transient_id ) result (is_target)

  implicit none

  integer, intent(in) :: transient_id
  logical             :: is_target

  interface

    function yac_is_transient_target_f ( transient_id ) &
      result ( is_target )                   &
      bind ( c, name='yac_is_transient_target' )

      use, intrinsic :: iso_c_binding, only : c_int

      implicit none

      integer ( kind=c_int ), value :: transient_id
      integer ( kind=c_int )        :: is_target

    end function yac_is_transient_target_f

  end interface

  is_target = yac_is_transient_target_f(transient_id) /= 0

end function yac_fis_transient_target

function yac_fget_transient_collection_size( transient_id ) &
  result (collection_size)

  implicit none

  integer, intent(in) :: transient_id
  integer             :: collection_size

  interface

    function yac_get_transient_collection_size_f ( transient_id ) &
      result ( collection_size )                                  &
      bind ( c, name='yac_get_transient_collection_size' )

      use, intrinsic :: iso_c_binding, only : c_int

      implicit none

      integer ( kind=c_int ), value :: transient_id
      integer ( kind=c_int )        :: collection_size

    end function yac_get_transient_collection_size_f

  end interface

  collection_size = yac_get_transient_collection_size_f(transient_id)

end function yac_fget_transient_collection_size

function yac_fget_transient_coupling_period ( transient_id ) &
  result (coupling_period)

  use, intrinsic :: iso_c_binding, only : c_ptr, c_char, c_null_char, c_f_pointer
  use mo_yac_finterface, dummy => yac_fget_transient_coupling_period

  implicit none

  integer, intent(in)              :: transient_id
  type (c_ptr)                     :: c_string_ptr
  character (kind=c_char), pointer :: coupling_period_chars(:)
  character (len=YAC_MAX_CHARLEN)  :: coupling_period
  integer :: i

  interface

    function yac_get_transient_coupling_period_f ( transient_id ) &
      result ( coupling_period )                                  &
      bind ( c, name='yac_get_transient_coupling_period' )

      use, intrinsic :: iso_c_binding, only : c_int, c_ptr
      integer ( kind=c_int ), value :: transient_id
      type(c_ptr)                   :: coupling_period
    end function yac_get_transient_coupling_period_f

  end interface

  c_string_ptr = yac_get_transient_coupling_period_f(transient_id)

  CALL c_f_pointer ( c_string_ptr, coupling_period_chars, (/YAC_MAX_CHARLEN/))

  coupling_period = " "

  do i=1, YAC_MAX_CHARLEN
    if (coupling_period_chars(i) == c_null_char) exit
    coupling_period(i:i) = coupling_period_chars(i)
  end do

end function yac_fget_transient_coupling_period

function yac_fget_transient_source_grid_name ( transient_id ) &
  result (source_grid_name)

  use, intrinsic :: iso_c_binding, only : c_ptr, c_char, c_null_char, c_f_pointer
  use mo_yac_finterface, dummy => yac_fget_transient_source_grid_name

  implicit none

  integer, intent(in)              :: transient_id
  type (c_ptr)                     :: c_string_ptr
  character (kind=c_char), pointer :: source_grid_name_chars(:)
  character (len=YAC_MAX_CHARLEN)  :: source_grid_name
  integer :: i

  interface

    function yac_get_transient_source_grid_name_f ( transient_id ) &
      result ( source_grid_name )                                  &
      bind ( c, name='yac_get_transient_source_grid_name' )

      use, intrinsic :: iso_c_binding, only : c_int, c_ptr
      integer ( kind=c_int ), value :: transient_id
      type(c_ptr)                   :: source_grid_name
    end function yac_get_transient_source_grid_name_f

  end interface

  c_string_ptr = yac_get_transient_source_grid_name_f(transient_id)

  CALL c_f_pointer ( c_string_ptr, source_grid_name_chars, (/YAC_MAX_CHARLEN/))

  source_grid_name = " "

  do i=1, YAC_MAX_CHARLEN
    if (source_grid_name_chars(i) == c_null_char) exit
    source_grid_name(i:i) = source_grid_name_chars(i)
  end do

end function yac_fget_transient_source_grid_name

function yac_fget_transient_target_grid_name ( transient_id ) &
  result (target_grid_name)

  use, intrinsic :: iso_c_binding, only : c_ptr, c_char, c_null_char, c_f_pointer
  use mo_yac_finterface, dummy => yac_fget_transient_target_grid_name

  implicit none

  integer, intent(in)              :: transient_id
  type (c_ptr)                     :: c_string_ptr
  character (kind=c_char), pointer :: target_grid_name_chars(:)
  character (len=YAC_MAX_CHARLEN)  :: target_grid_name
  integer :: i

  interface

    function yac_get_transient_target_grid_name_f ( transient_id ) &
      result ( target_grid_name )                                  &
      bind ( c, name='yac_get_transient_target_grid_name' )

      use, intrinsic :: iso_c_binding, only : c_int, c_ptr
      integer ( kind=c_int ), value :: transient_id
      type(c_ptr)                   :: target_grid_name
    end function yac_get_transient_target_grid_name_f

  end interface

  c_string_ptr = yac_get_transient_target_grid_name_f(transient_id)

  CALL c_f_pointer ( c_string_ptr, target_grid_name_chars, (/YAC_MAX_CHARLEN/))

  target_grid_name = " "

  do i=1, YAC_MAX_CHARLEN
    if (target_grid_name_chars(i) == c_null_char) exit
    target_grid_name(i:i) = target_grid_name_chars(i)
  end do

end function yac_fget_transient_target_grid_name

function yac_fget_transient_name ( transient_id ) result (transient_name)

  use, intrinsic :: iso_c_binding, only : c_ptr, c_char, c_null_char, c_f_pointer
  use mo_yac_finterface, dummy => yac_fget_transient_name

  implicit none

  integer, intent(in)              :: transient_id
  type (c_ptr)                     :: c_string_ptr
  character (kind=c_char), pointer :: transient_name_chars(:)
  character (len=YAC_MAX_CHARLEN)  :: transient_name
  integer :: i

  interface

    function yac_get_transient_name_f ( transient_id ) &
      result ( transient_name )                                  &
      bind ( c, name='yac_get_transient_name' )

      use, intrinsic :: iso_c_binding, only : c_int, c_ptr
      integer ( kind=c_int ), value :: transient_id
      type(c_ptr)                   :: transient_name
    end function yac_get_transient_name_f

  end interface

  c_string_ptr = yac_get_transient_name_f(transient_id)

  CALL c_f_pointer ( c_string_ptr, transient_name_chars, (/YAC_MAX_CHARLEN/))

  transient_name = " "

  do i=1, YAC_MAX_CHARLEN
    if (transient_name_chars(i) == c_null_char) exit
    transient_name(i:i) = transient_name_chars(i)
  end do

end function yac_fget_transient_name

