/**
 * @file test_dummy_coupling7_c.c
 *
 * @copyright Copyright  (C)  2021 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "yac_config.h"

#define EXACT

#include <mpi.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "yac_interface.h"
#include "read_icon_grid.h"
#include "read_mpiom_grid.h"
#include "generate_cubed_sphere.h"
#include "test_function.h"
#include "utils.h"

#include "tests.h"

#define DUMMY_VALUE (-1337.0)

static void master_component();
static void slave_component();

int main (void) {
#ifdef YAC_NETCDF_ENABLED

  MPI_Init(NULL, NULL);
  xt_initialize(MPI_COMM_WORLD);

  int global_rank, global_size;
  MPI_Comm_rank(MPI_COMM_WORLD, &global_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &global_size);

  if (global_size != 5) {
    fprintf(stderr, "Wrong number of processes (should be 5)\n");
    exit(EXIT_FAILURE);
  }
  
  switch (global_rank) {
    case (0):
    case (1):
    case (2):
      master_component();
      break;
    case (3):
    case (4):
      slave_component();
      break;
  };

  xt_finalize();
  MPI_Finalize();

  return TEST_EXIT_CODE;

#else

  return EXIT_SKIP_TEST;

#endif
}

static void register_icon_grid(
  MPI_Comm comm, char const * grid_name,
  int * corner_point_id, int * cell_point_id, double ** dummy_field_data) {

  int comm_rank, comm_size;
  MPI_Comm_rank(comm, &comm_rank);
  MPI_Comm_size(comm, &comm_size);

  int nbr_vertices;
  int nbr_cells;
  int * num_vertices_per_cell;
  int * cell_to_vertex;
  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;

  int * cell_mask;
  int * global_cell_id;
  int * cell_core_mask;
  int * global_corner_id;
  int * corner_core_mask;

  read_part_icon_grid_information(
    grid_name, &nbr_vertices, &nbr_cells, &num_vertices_per_cell,
    &cell_to_vertex, &x_vertices, &y_vertices, &x_cells, &y_cells,
    &global_cell_id, &cell_mask,
    &cell_core_mask, &global_corner_id, &corner_core_mask,
    comm_rank, comm_size);

  int grid_id;
  yac_cdef_grid_unstruct(
    "R2B02", nbr_vertices, nbr_cells, num_vertices_per_cell,
    x_vertices, y_vertices, cell_to_vertex, &grid_id);

  yac_cset_global_index(global_corner_id, YAC_LOCATION_CORNER, grid_id);
  yac_cset_core_mask(corner_core_mask, YAC_LOCATION_CORNER, grid_id);
  yac_cset_global_index(global_cell_id, YAC_LOCATION_CELL, grid_id);
  yac_cset_core_mask(cell_core_mask, YAC_LOCATION_CELL, grid_id);

  yac_cdef_points_unstruct(
    grid_id, nbr_vertices, YAC_LOCATION_CORNER, x_vertices, y_vertices,
    corner_point_id);
  yac_cdef_points_unstruct(
    grid_id, nbr_cells, YAC_LOCATION_CELL, x_cells, y_cells,
    cell_point_id);

  for (size_t i = 0; i < nbr_cells; ++i) cell_mask[i] = cell_mask[i] == 0;

  yac_cset_mask(cell_mask, *cell_point_id);

  *dummy_field_data = xmalloc(nbr_cells * sizeof(**dummy_field_data));
  for (int i = 0; i < nbr_cells; ++i)
    (*dummy_field_data)[i] =
      (cell_core_mask[i])?
        (test_harmonic(x_cells[i]*YAC_RAD, y_cells[i]*YAC_RAD)):DUMMY_VALUE;

  delete_icon_grid_data(&cell_mask,
                        &global_cell_id,
                        &cell_core_mask,
                        &num_vertices_per_cell,
                        &global_corner_id,
                        &corner_core_mask,
                        &cell_to_vertex,
                        &x_cells,
                        &y_cells,
                        &x_vertices,
                        &y_vertices);
}

static void master_component() {

  int global_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &global_rank);
  int couple_config_id;

  // generate coupling configuration
  if (global_rank == 0) {

    // get basic coupling configuration from file
    yac_get_couple_config_xml(
      "coupling_test7.xml", "coupling.xsd", &couple_config_id);

    // add slave component
    yac_add_couple_config_component(couple_config_id, "out_comp");

    // grids on slave components
    yac_add_couple_config_grid(couple_config_id, "cube_N20");
    yac_add_couple_config_grid(couple_config_id, "mpiom_GR30_lsm");

    // set the collection size for the transients that are already in the
    // coupling configuration
    yac_set_couple_config_transient_collection_size(
      couple_config_id, "default_field_a", 1);
    yac_set_couple_config_transient_collection_size(
      couple_config_id, "default_field_b", 1);
    yac_set_couple_config_transient_collection_size(
      couple_config_id, "default_field_c", 1);

    // add two additional transients that are not already in the
    // default configuration
    yac_add_couple_config_transient(couple_config_id, "additional_field_a", 1);
    yac_add_couple_config_transient(couple_config_id, "additional_field_b", 1);

    // define a n-nearest-neighbour interpolation stack
    int nnn_interp_stack_id;
    yac_get_interp_stack_config(&nnn_interp_stack_id);
    yac_add_interp_stack_config_nnn(
      nnn_interp_stack_id, YAC_NNN_DIST, 16, -1.0);
    yac_add_interp_stack_config_fixed(
      nnn_interp_stack_id, DUMMY_VALUE);

    // define a second order conservative stack
    int conserv_interp_stack_id;
    yac_get_interp_stack_config(&conserv_interp_stack_id);
    yac_add_interp_stack_config_conservative(
      conserv_interp_stack_id, 2, 0, 0, YAC_CONSERV_DESTAREA);
    yac_add_interp_stack_config_fixed(
      conserv_interp_stack_id, DUMMY_VALUE);    

    // fields that are to be sent to the slave component
    yac_add_couple_config_transient_couple(
      couple_config_id,
      "default_field_b",       // field name
      "ICON",                  // source component
      "R2B02",                 // source grid
      "1",                     // source timestep: every second
      0,                       // source lag: 0
      "out_comp",              // target component
      "cube_N20",              // target grid
      "1",                     // target timestep: every second
      0,                       // target lat: 0
      1,                       // mapping is done on source
      nnn_interp_stack_id,     // n-nearest-neighbour interpolation stack
      YAC_REDUCTION_INSTANT,   // coupling period operation: none
      "1",                     // coupling period: every second
      0,                       // write weight file: no
      "weights_default_b.nc"); // weight file name
    yac_add_couple_config_transient_couple(
      couple_config_id,
      "default_field_c",        // field name
      "ICON",                   // source component
      "R2B02",                  // source grid
      "1",                      // source timestep: every second
      0,                        // source lag: 0
      "out_comp",               // target component
      "mpiom_GR30_lsm",         // target grid
      "10",                     // target timestep: every ten second
      0,                        // target lat: 0
      1,                        // mapping is done on source
      conserv_interp_stack_id,  // 2nd order conservative interpolation stack
      YAC_REDUCTION_ACCUMULATE, // coupling period operation: accumulate source field
      "10",                     // coupling period: every second
      0,                        // write weight file: no
      "weights_default_c.nc");  // weight file name
    yac_add_couple_config_transient_couple(
      couple_config_id,
      "additional_field_a",       // field name
      "ICON",                     // source component
      "R2B02",                    // source grid
      "1",                        // source timestep: every second
      0,                          // source lag: 0
      "out_comp",                 // target component
      "cube_N20",                 // target grid
      "10",                       // target timestep: every ten seconds
      0,                          // target lat: 0
      1,                          // mapping is done on source
      nnn_interp_stack_id,        // n-nearest-neighbour interpolation stack
      YAC_REDUCTION_TIME_AVERAGE, // coupling period operation: average source field
      "10",                       // coupling period: every second
      0,                          // write weight file: no
      "weights_additiona_a.nc");  // weight file name
    yac_add_couple_config_transient_couple(
      couple_config_id,
      "additional_field_b",      // field
      "ICON",                    // source component
      "R2B02",                   // source grid
      "1",                       // source timestep: every second
      0,                         // source lag: 0
      "out_comp",                // target component
      "mpiom_GR30_lsm",          // target grid
      "1",                       // target timestep: every second
      0,                         // target lat: 0
      1,                         // mapping is done on source
      nnn_interp_stack_id,       // n-nearest-neighbour interpolation stack
      YAC_REDUCTION_INSTANT,     // coupling period operation: accumulate source field
      "1",                       // coupling period: every second
      1,                         // write weight file: yes
      "weights_additiona_b.nc"); // weight file name

    yac_free_interp_stack_config(conserv_interp_stack_id);
    yac_free_interp_stack_config(nnn_interp_stack_id);
  } else {
    couple_config_id = COUPLE_CONFIG_DUMMY_ID;
  }

  // init YAC with coupling configuration from rank 0
  yac_cinit_couple_config(couple_config_id);

  // register "ICON" component
  int comp_id;
  yac_cdef_comp("ICON", &comp_id);

  // get MPI communicator for the local component
  MPI_Comm comp_comm;
  yac_cget_localcomm(&comp_comm, comp_id);

  // read the ICON grid from file and register it
  int corner_point_id, cell_point_id;
  double * dummy_field_data;
  register_icon_grid(
    comp_comm, "icon_grid_R02B02.nc",
    &corner_point_id, &cell_point_id, &dummy_field_data);

  MPI_Comm_free(&comp_comm);

  // register fields
  int field_ids[4];
  yac_cdef_field(
    "default_field_b", comp_id, &cell_point_id, 1, &(field_ids[0]));
  yac_cdef_field(
    "default_field_c", comp_id, &cell_point_id, 1, &(field_ids[1]));
  yac_cdef_field(
    "additional_field_a", comp_id, &cell_point_id, 1, &(field_ids[2]));
  yac_cdef_field(
    "additional_field_b", comp_id, &cell_point_id, 1, &(field_ids[3]));

  // compute interpolation weights
  int ierror;
  yac_csearch(&ierror);

  // exchange number of timesteps
  MPI_Comm comp_pair_comm;
  yac_cget_pair_rootcomm(&comp_pair_comm, "ICON", "out_comp");
  int num_timesteps = 100;
  MPI_Bcast(&num_timesteps, 1, MPI_INT, 0, comp_pair_comm);
  MPI_Comm_free(&comp_pair_comm);

  // do some time steps
  for (int t = 0; t < num_timesteps; ++t) {

    {
      int info, err;
      double *point_set_data[1];
      double **collection_data[1] = {point_set_data};
      point_set_data[0] = dummy_field_data;

      yac_cput(field_ids[0], 1, collection_data, &info, &err);
    }

    {
      int info, err;
      double *point_set_data[1];
      double **collection_data[1] = {point_set_data};
      point_set_data[0] = dummy_field_data;

      yac_cput(field_ids[1], 1, collection_data, &info, &err);
    }

    {
      int info, err;
      double *point_set_data[1];
      double **collection_data[1] = {point_set_data};
      point_set_data[0] = dummy_field_data;

      yac_cput(field_ids[2], 1, collection_data, &info, &err);
    }

    {
      int info, err;
      double *point_set_data[1];
      double **collection_data[1] = {point_set_data};
      point_set_data[0] = dummy_field_data;

      yac_cput(field_ids[3], 1, collection_data, &info, &err);
    }
  }

  free(dummy_field_data);

  yac_cfinalize();
}

struct field_config {
  char const * name;
  int id;
  char const * grid_name;
  int collection_size;
  int coupling_period;
  double * data;
  size_t size;
};

static void get_field_configuration(
  int comp_id, int * field_count, struct field_config ** fields) {

  // gets the number of active transients for a component
  *field_count = yac_get_transients_count(comp_id);

  *fields = xmalloc((size_t)*field_count * sizeof(**fields));

  int * transient_ids = xmalloc((size_t)*field_count * sizeof(*transient_ids));

  // gets ids for all active transients
  yac_get_transient_ids(comp_id, transient_ids);

  for (int i = 0; i < *field_count; ++i) {

    int transient_id = transient_ids[i];
    int is_source = yac_is_transient_source(transient_id);
    int is_target = yac_is_transient_target(transient_id);
    char const * transient_name = yac_get_transient_name(transient_id);

    if (is_source) {
      fprintf(
        stderr, "ERROR: IOSERVER is set as source for transient \"%s\"\n",
        transient_name);
      exit(EXIT_FAILURE);

    } else if (is_target) {

      (*fields)[i].name = strdup(transient_name);
      (*fields)[i].grid_name =
        strdup(yac_get_transient_target_grid_name(transient_id));
      (*fields)[i].collection_size =
        yac_get_transient_collection_size(transient_id);
      (*fields)[i].coupling_period =
        atoi(yac_get_transient_coupling_period(transient_id));

    } else if ((!is_source) && (!is_target)) {
      fprintf(
        stderr, "ERROR: out_comp is neither source nor target for \"%s\", "
                "this should not have happend\n", transient_name);
      exit(EXIT_FAILURE);
    }
  }
  free(transient_ids);
}

static void register_cube_grid(
  MPI_Comm comm, int comp_id, char const * grid_name,
  int * corner_point_id, int * cell_point_id,
  size_t * nbr_vertices_out, size_t * nbr_cells_out) {

  int n;
  if (sscanf(grid_name, "cube_N%d", &n) != 1) {
    fprintf(
      stderr, "ERROR(register_cube_grid): invalid cube grid name \"%s\"",
      grid_name);
    exit(EXIT_FAILURE);
  }

  int comm_rank, comm_size;
  MPI_Comm_rank(comm, &comm_rank);
  MPI_Comm_size(comm, &comm_size);

  unsigned nbr_vertices;
  unsigned nbr_cells;
  unsigned * num_vertices_per_cell;
  unsigned * cell_to_vertex;
  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;

  int * cell_core_mask;
  int * corner_core_mask;
  int * global_cell_id;
  int * global_corner_id;

  generate_part_cube_grid_information(n, &nbr_vertices, &nbr_cells,
                                      &num_vertices_per_cell, &cell_to_vertex,
                                      &x_vertices, &y_vertices, &x_cells,
                                      &y_cells, &global_cell_id,
                                      &cell_core_mask, &global_corner_id,
                                      &corner_core_mask, comm_rank, comm_size);

  int grid_id;
  yac_cdef_grid_unstruct(
    grid_name, nbr_vertices, nbr_cells, (int*)num_vertices_per_cell,
    x_vertices, y_vertices, (int*)cell_to_vertex, &grid_id);

  yac_cset_global_index(global_corner_id, YAC_LOCATION_CORNER, grid_id);
  yac_cset_core_mask(corner_core_mask, YAC_LOCATION_CORNER, grid_id);
  yac_cset_global_index(global_cell_id, YAC_LOCATION_CELL, grid_id);
  yac_cset_core_mask(cell_core_mask, YAC_LOCATION_CELL, grid_id);

  yac_cdef_points_unstruct(
    grid_id, nbr_vertices, YAC_LOCATION_CORNER, x_vertices, y_vertices,
    corner_point_id);
  yac_cdef_points_unstruct(
    grid_id, nbr_cells, YAC_LOCATION_CELL, x_cells, y_cells,
    cell_point_id);
  *nbr_vertices_out = (size_t)nbr_vertices;
  *nbr_cells_out = (size_t)nbr_cells;

  free(cell_core_mask);
  free(corner_core_mask);
  free(global_cell_id);
  free(global_corner_id);
  free(x_vertices);
  free(y_vertices);
  free(x_cells);
  free(y_cells);
  free(num_vertices_per_cell);
  free(cell_to_vertex);
}

static void register_mpiom_grid(
  MPI_Comm comm, int comp_id, char const * grid_name,
  int * corner_point_id, int * cell_point_id,
  size_t * nbr_vertices_out, size_t * nbr_cells_out) {

  char grid_file_name[1024];
  if (sscanf(grid_name, "mpiom_%s", grid_file_name) != 1) {
    fprintf(
      stderr, "ERROR(register_mpiom_grid): invalid mpiom grid name \"%s\"",
      grid_name);
    exit(EXIT_FAILURE);
  }
  strcat(grid_file_name, ".nc");

  int comm_rank, comm_size;
  MPI_Comm_rank(comm, &comm_rank);
  MPI_Comm_size(comm, &comm_size);

  int nbr_vertices;
  int nbr_cells;
  int * num_vertices_per_cell;
  int * cell_to_vertex;
  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;

  int * cell_mask;
  int * cell_core_mask;
  int * corner_core_mask;
  int * global_cell_id;
  int * global_corner_id;

  read_part_mpiom_grid_information(
    grid_file_name, &nbr_vertices, &nbr_cells,
    &num_vertices_per_cell, &cell_to_vertex,
    &x_vertices, &y_vertices, &x_cells, &y_cells,
    &global_cell_id, &cell_mask, &cell_core_mask,
    &global_corner_id, &corner_core_mask, comm_rank, comm_size);

  for (int i = 0; i < nbr_cells; ++i) cell_mask[i] = cell_mask[i] == 0;

  int grid_id;
  yac_cdef_grid_unstruct(
    grid_name, nbr_vertices, nbr_cells, num_vertices_per_cell,
    x_vertices, y_vertices, cell_to_vertex, &grid_id);

  yac_cset_global_index(global_corner_id, YAC_LOCATION_CORNER, grid_id);
  yac_cset_core_mask(corner_core_mask, YAC_LOCATION_CORNER, grid_id);
  yac_cset_global_index(global_cell_id, YAC_LOCATION_CELL, grid_id);
  yac_cset_core_mask(cell_core_mask, YAC_LOCATION_CELL, grid_id);

  yac_cdef_points_unstruct(
    grid_id, nbr_vertices, YAC_LOCATION_CORNER, x_vertices, y_vertices,
    corner_point_id);
  yac_cdef_points_unstruct(
    grid_id, nbr_cells, YAC_LOCATION_CELL, x_cells, y_cells,
    cell_point_id);
  yac_cset_mask(cell_mask, *cell_point_id);
  *nbr_vertices_out = (size_t)nbr_vertices;
  *nbr_cells_out = (size_t)nbr_cells;

  free(cell_mask);
  free(cell_core_mask);
  free(corner_core_mask);
  free(global_cell_id);
  free(global_corner_id);
  free(x_vertices);
  free(y_vertices);
  free(x_cells);
  free(y_cells);
  free(num_vertices_per_cell);
  free(cell_to_vertex);
}

static int compare_fields_grid_name(void const * a, void const * b) {

  return
    strcmp(
      ((struct field_config*)a)->grid_name,
      ((struct field_config*)b)->grid_name);
}

static void register_fields(
  MPI_Comm comm, int comp_id, int field_count, struct field_config * fields) {

  // sort field by grid names
  qsort(
    fields, (size_t)field_count, sizeof(*fields), compare_fields_grid_name);

  char const * prev_grid_name = "not_a_grid_name";
  int corner_point_id = -1, cell_point_id = -1;
  size_t nbr_vertices = 0;
  size_t nbr_cells;

  for (int i = 0; i < field_count; ++i) {

    // register grids on demand
    if (strcmp(prev_grid_name, fields[i].grid_name)) {

      prev_grid_name = fields[i].grid_name;

      if (!strcmp(prev_grid_name, "cube_N20")) {

        register_cube_grid(
          comm, comp_id, prev_grid_name,
          &corner_point_id, &cell_point_id, &nbr_vertices, &nbr_cells);

      } else if (!strcmp(prev_grid_name, "mpiom_GR30_lsm")) {

        register_mpiom_grid(
          comm, comp_id, prev_grid_name,
          &corner_point_id, &cell_point_id, &nbr_vertices, &nbr_cells);

      } else {
        fprintf(stderr, "ERROR(out_comp): unsupported grid name: \"%s\"\n",
                prev_grid_name);
        exit(EXIT_FAILURE);
      }
    }

    // register field
    yac_cdef_field(
      fields[i].name, comp_id, &cell_point_id, 1, &(fields[i].id));
    fields[i].data = xmalloc(nbr_cells * sizeof(*(fields[i].data)));
    fields[i].size = nbr_cells;
  }
}

static void slave_component() {

  // init YAC with coupling configuration from rank 0
  yac_cinit_couple_config(COUPLE_CONFIG_DUMMY_ID);

  // register "out_comp" component
  int comp_id;
  yac_cdef_comp("out_comp", &comp_id);

  // get MPI communicator for the local component
  MPI_Comm comp_comm;
  yac_cget_localcomm(&comp_comm, comp_id);

  // extract information about the coupling
  int field_count;
  struct field_config * fields;
  get_field_configuration(comp_id, &field_count, &fields);

  // register grids required by the field configuration
  register_fields(comp_comm, comp_id, field_count, fields);

  MPI_Comm_free(&comp_comm);

  // compute interpolation weights
  int ierror;
  yac_csearch(&ierror);

  // exchange number of timesteps
  MPI_Comm comp_pair_comm;
  yac_cget_pair_rootcomm(&comp_pair_comm, "ICON", "out_comp");
  int num_timesteps;
  MPI_Bcast(&num_timesteps, 1, MPI_INT, 0, comp_pair_comm);
  MPI_Comm_free(&comp_pair_comm);

  // do some time steps
  for (int t = 0; t < num_timesteps; ++t) {

    if (!(t % fields[0].coupling_period)) {
      int info, err;
      double *collection_data[1] = {fields[0].data};
      yac_cget(fields[0].id, 1, collection_data, &info, &err);
    }
    if (!(t % fields[1].coupling_period)) {
      int info, err;
      double *collection_data[1] = {fields[1].data};
      yac_cget(fields[1].id, 1, collection_data, &info, &err);
    }
    if (!(t % fields[2].coupling_period)) {
      int info, err;
      double *collection_data[1] = {fields[2].data};
      yac_cget(fields[2].id, 1, collection_data, &info, &err);
    }
    if (!(t % fields[3].coupling_period)) {
      int info, err;
      double *collection_data[1] = {fields[3].data};
      yac_cget(fields[3].id, 1, collection_data, &info, &err);
    }
  }

  for (int i = 0; i < field_count; ++i) {
    free((void*)(fields[i].name));
    free((void*)(fields[i].grid_name));
    free((void*)(fields[i].data));
  }
  free(fields);

  yac_cfinalize();
}
